package fr.unistra.ibmc.paradise.core;

import fr.unistra.ibmc.paradise.ArchitectureException;

import java.io.File;
import java.util.ArrayList;
import java.util.List;


public abstract class AbstractMolecule extends AbstractDocumentable implements Molecule {

    transient private List<Annotation> annotations;
    private List<Annotation> backupAnnotations;
    protected List<ParadiseFeature> selectedFeatures;
    protected String name = "?";
    protected StringBuffer sequence;
    protected ParadiseID paradiseID;
    transient protected MoleculeFactory moleculeFactory;
    protected Taxon taxon;
    protected String[] phylogeny;
    protected Source source;
    protected File associatedFile;
    protected Organism organism;

    protected AbstractMolecule(ParadiseID paradiseID, MoleculeFactory moleculeFactory, String name, Source source) {
        this.moleculeFactory = moleculeFactory;
        this.paradiseID = paradiseID;
        this.setName(name);
        this.source = source;
        this.annotations = new ArrayList<Annotation>();
        this.backupAnnotations = new ArrayList<Annotation>();
        this.selectedFeatures = new ArrayList<ParadiseFeature>();
        this.sequence = new StringBuffer();
        this.comments = new ArrayList<Comment>();
    }

    public void addModifiedResidue(String modifiedResidue, String unmodifiedResidue) {

    }


    public Source getSource() {
        return this.source;
    }

    public void setSource(Source source) {
        this.source = source;
    }

    public void setAssociatedFile(File file) {
        this.associatedFile = file;
    }

    public File getAssociatedFile() {
        return this.associatedFile;
    }

    public Organism getOrganism() {
        return this.organism;
    }

    public void setOrganism(Organism organism) {
        this.organism = organism;
    }

    public Molecule searchCopy(List<Molecule> molecules) {
        for (Molecule _m : molecules) {
            if (this.getParadiseID().equals(_m.getParadiseID()))
                return _m;
        }
        for (Molecule _m : molecules) {
            if (this.printSequence().equals(_m.printSequence()))
                return _m;
        }
        return null;
    }

    public void makeAllAnnotationsPersistent() {
        this.backupAnnotations.addAll(this.annotations);
    }

    public void makeAllAnnotationsNotPersistent() {
        this.backupAnnotations.clear();
    }

    public void recoverBackupAnnotations() {
        this.annotations = new ArrayList<Annotation>();
        this.annotations.addAll(this.backupAnnotations);
    }


    public void addFeature(ParadiseFeature feature, Location l) throws ArchitectureException {
        if (l.getEnd() > this.getLength())
            throw new fr.unistra.ibmc.paradise.ArchitectureException("Location to add to a feature outside of a molecule boundaries " + l.getEnd() + " > " + this.getLength(), this, l, feature);
        //TODO throws an Exception to avoid a null Location and/or a null ParadiseFeature
        //since only one Annotation per Feature per Molecule, if this feature is already registered, its Location is extended
        if (this.getAnnotation(feature) != null) {
            this.getAnnotation(feature).addToLocation(l);
            return;
        }
        //otherwise, a new Annotation is produced and registered
        AnnotationImpl a = new AnnotationImpl(feature, l);
        a.molecule = this;
        if (this.annotations == null)
            this.annotations = new ArrayList<Annotation>();
        this.annotations.add(a);
        try {
            feature.addAnnotation(a);
        }
        //if problem to add this new Annotation, we unregister the annotation
        catch (fr.unistra.ibmc.paradise.ArchitectureException e) {
            a.molecule = null;
            this.annotations.remove(a);
            //we forward the exception to explain what's the problem to the calling method
            throw e;
        }
    }

    public void moveFeature(ParadiseFeature feature, Location l) throws fr.unistra.ibmc.paradise.ArchitectureException {
        if (l.getEnd() > this.getLength())
            throw new fr.unistra.ibmc.paradise.ArchitectureException("Location to add to a feature outside of a molecule boundaries " + l.getEnd() + " > " + this.getLength(), this, l, feature);
        Location previousLocation = feature.getFullLocation(this);
        //[fjossinet]] we are not manipulating the location object directly. We are using the addToLocation and removeFromLocation to warn the parent and sub-features that they need to update themselves too
        if (previousLocation != null) {
            //[fjossinet] at this moment, the location can be strange for a given feature (for example a base-base interaction annotating more than two positions) => can it be a problem in a near future ?
            feature.addToLocation(l, this);
            //we remove only the non-overlapping region between the old and the new location
            feature.removeFromLocation(previousLocation.differenceOf(l), this);
        }
    }

    public void removeFeature(ParadiseFeature feature) {
        if (this.annotations == null)
            return;
        AnnotationImpl hit = null;
        for (Annotation a : this.annotations)
            if (a.getFeature().equals(feature)) {
                hit = (AnnotationImpl) a;
                break;
            }
        if (hit != null) {
            this.annotations.remove(hit);
            feature.removeAnnotation(hit);
            //set the Molecule reference to null at the end otherwise problem in the feature.removeAnnotation(hit) call
            hit.molecule = null;
        }
    }

    public boolean isSelected(ParadiseFeature feature) {
        return this.selectedFeatures.contains(feature);
    }

    public void addSelectedFeature(ParadiseFeature feature) {
        this.selectedFeatures.add(feature);
    }

    public void removeSelectedFeature(ParadiseFeature feature) {
        this.selectedFeatures.remove(feature);
    }

    public void removeAllSelectedFeatures() {
        this.selectedFeatures.clear();
    }

    public List<ParadiseFeature> getSelectedFeatures() {
        return new ArrayList<ParadiseFeature>(this.selectedFeatures);
    }

    public <T extends ParadiseFeature> List<T> getSelectedFeatures(Class<T> c) {
        List<T> features = new ArrayList<T>();
        for (ParadiseFeature f : this.selectedFeatures)
            if (c.isInstance(f))
                features.add((T) f);
        return features;
    }

    public List<Annotation> getAnnotations() {
        if (this.annotations == null)
            return new ArrayList<Annotation>();
        return new ArrayList<Annotation>(this.annotations);
    }

    public List<Annotation> getAnnotations(Class c, Location l) {
        List<Annotation> hits = new ArrayList<Annotation>();
        for (Annotation a : this.getAnnotations(c))
            if (a.getLocation().isIncludedIn(l))
                hits.add(a);
        return hits;
    }

    public List<Annotation> getAnnotations(Location l) {
        List<Annotation> hits = new ArrayList<Annotation>();
        if (this.annotations == null)
            return hits;
        for (Annotation a : this.annotations)
            if (a.getLocation().isIncludedIn(l))
                hits.add(a);
        return hits;
    }

    public Annotation getAnnotation(ParadiseFeature feature) {
        if (this.annotations == null)
            return null;
        for (Annotation a : this.annotations) {
            if (a.getFeature().equals(feature)) {
                return a;
            }
        }
        return null;
    }

    public List<Annotation> getAnnotations(Class c) {
        List<Annotation> hits = new ArrayList<Annotation>();
        for (Annotation a : this.getAnnotations())
            if (c.isInstance(a.getFeature()))
                hits.add(a);
        return hits;
    }

    public void setName(String name) {
        //tludwig if the name contains special characters, it can mess up with temporary directory creation in some analysis
        this.name = name.replace("/","_").replace("\\","_");
    }

    public String getName() {
        return name;
    }

    public int getLength() {
        return this.sequence.length();
    }

    public String printSequence(Location l) {
        return this.sequence.substring(l.getStart() - 1, l.getEnd());
    }

    public String getResidueAt(int position) throws fr.unistra.ibmc.paradise.ArchitectureException {
        if (position <= 0 || position > this.getLength())
            throw new fr.unistra.ibmc.paradise.ArchitectureException("The position asked for is outside the molecule's boundaries", this, new Location(position));
        else
            return Character.toString(this.sequence.charAt(position - 1));
    }

    public Taxon getTaxon() {
        return this.taxon;
    }

    public void setTaxon(Taxon taxon) {
        this.taxon = taxon;
    }

    public String printSequence() {
        return this.sequence.toString();
    }

    public ParadiseID getParadiseID() {
        return this.paradiseID;
    }

    public String[] getPhylogeny() {
        return this.phylogeny.clone();
    }

    /**
     * Two molecules are equals if they have the same ParadiseID
     *
     * @param o
     * @return
     */
    public boolean equals(Object o) {
        if (!Molecule.class.isInstance(o))
            return false;
        return this.getParadiseID().equals(((Molecule) o).getParadiseID());
    }

    public class AnnotationImpl extends AbstractDocumentable implements Annotation {
        protected Location location;
        protected ParadiseFeature feature;
        protected Molecule molecule;


        private AnnotationImpl(ParadiseFeature feature, Location location) {
            this.feature = feature;
            this.location = new Location(location);
        }

        public Molecule getMolecule() {
            return this.molecule;
        }

        public ParadiseFeature getFeature() {
            return this.feature;
        }

        public Location getLocation() {
            return this.location;
        }

        public void addToLocation(Location l) throws fr.unistra.ibmc.paradise.ArchitectureException {
            this.feature.addToLocation(l, this.molecule);
        }

        public void removeFromLocation(Location l) {
            this.feature.removeFromLocation(l, this.molecule);
        }
    }

    public String toString() {
        return (Protein.class.isInstance(this) ? "Protein ":"RNA ")+this.getName()+" ("+this.getLength()+(Protein.class.isInstance(this) ? "aas":"nts")+") from "+this.getSource().getId();
    }
}
