package fr.unistra.ibmc.paradise.core;

import fr.unistra.ibmc.paradise.BiologicalSymbolException;

import java.util.List;

public interface Molecule extends Documentable, ParadiseProjectItem {

    /**
     * Add a new residue at the end of the molecule
     *
     * @param residue
     * @throws fr.unistra.ibmc.paradise.BiologicalSymbolException if the new residue is unknown
     */
    public void addResidue(String residue) throws BiologicalSymbolException;

    public void removeFeature(ParadiseFeature feature);

    public void addFeature(ParadiseFeature feature, Location l) throws fr.unistra.ibmc.paradise.ArchitectureException;

    public ParadiseID getParadiseID();

    public void makeAllAnnotationsPersistent();

    public void makeAllAnnotationsNotPersistent();

    public void recoverBackupAnnotations();

    public void addModifiedResidue(String modifiedResidue, String unmodifiedResidue);

    /**
     * Return an empty String array if no phylogeny available for this molecule
     *
     * @return
     */
    public String[] getPhylogeny();

    public Organism getOrganism();

    public void setOrganism(Organism organism);

    /**
     * Set a new location for the feature given as argument. Do nothing if the feature doesn't annotate this molecule
     *
     * @param feature the feature to relocate
     * @param l       the new location
     * @throws fr.unistra.ibmc.paradise.ArchitectureException if the location is out of the molecule's boundaries
     */
    public void moveFeature(ParadiseFeature feature, Location l) throws fr.unistra.ibmc.paradise.ArchitectureException;

    public List<Annotation> getAnnotations();

    public boolean isSelected(ParadiseFeature feature);

    /**
     * Return all the Annotation objects wrapping a feature object which is an instance of the class given as argument
     * and whose Location is included in the Location given as argument
     *
     * @param c
     * @param l
     * @return
     */
    public List<Annotation> getAnnotations(Class c, Location l);

    /**
     * Return all the Annotation objects whose Location is included in the Location gicen as argument
     *
     * @param l
     * @return
     */
    public List<Annotation> getAnnotations(Location l);

    /**
     * Return the Annotation object wrapping the feature given as argument
     *
     * @param feature
     * @return null if no Annotation object found
     */
    public Annotation getAnnotation(ParadiseFeature feature);

    /**
     * Return all the Annotation objects wrapping a feature object which is an instance of the class given as argument
     *
     * @param c
     * @return
     */
    public List<Annotation> getAnnotations(Class c);

    public void addSelectedFeature(ParadiseFeature feature);

    public void removeSelectedFeature(ParadiseFeature feature);

    public void removeAllSelectedFeatures();

    public List<ParadiseFeature> getSelectedFeatures();

    public <T extends ParadiseFeature> List<T> getSelectedFeatures(Class<T> c);

    public void setName(String name);

    public String getName();

    public int getLength();

    /**
     * Return the subsequence as a String for the Location given as argument
     *
     * @param l
     * @return
     */
    public String printSequence(Location l);

    /**
     * Return the sequence of the full molecule as a String
     *
     * @return
     */
    public String printSequence();

    /**
     * Search a molecule in the list given as argument which is identical to the current molecule
     * The first criteria of identity is the same ParadiseID.
     * The second criteria of identity is the same sequence.
     *
     * @return
     */
    public Molecule searchCopy(List<Molecule> molecules);

    /**
     * Return the residue symbol stored at the position given as argument. If this position is annotated with a ModifiedResidue,
     * the symbol of the unmodified residue is returned instead
     *
     * @param position has to be between 1 and molecule length
     * @return
     * @throws fr.unistra.ibmc.paradise.ArchitectureException if the position is outside of the Molecule's boundaries
     */
    public String getResidueAt(int position) throws fr.unistra.ibmc.paradise.ArchitectureException;

    public Taxon getTaxon();

    public void setTaxon(Taxon taxon);

    public boolean equals(Object o);

    public interface Annotation extends Documentable {

        public Molecule getMolecule();

        public ParadiseFeature getFeature();

        /**
         * Return the location of this feature only (not the full location with the subfeatures)
         * Any modification from the calling method will modify the original location. This is really different than the getFullLocation() method provided by the ParadiseFeature concept.
         * @return
         */
        public Location getLocation();

        /**
         * @param l
         * @throws fr.unistra.ibmc.paradise.ArchitectureException if any problem with the new Location (out of the Molecule boundaries for exemple)
         */
        public void addToLocation(Location l) throws fr.unistra.ibmc.paradise.ArchitectureException;

        /**
         * @param l
         */
        public void removeFromLocation(Location l);

    }
}
