package fr.unistra.ibmc.paradise.core.features;

import fr.unistra.ibmc.paradise.ArchitectureException;
import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.utils.Residue;

import java.io.File;
import java.util.*;


public abstract class AbstractParadiseFeature extends AbstractDocumentable implements ParadiseFeature {

    protected List<ParadiseFeature> subfeatures, parentfeatures;
    protected Map<Molecule, Molecule.Annotation> annotations;
    protected ParadiseFeatureFactory factory;
    protected Source source;
    protected File associatedFile;
    protected ParadiseID paradiseID;

    protected AbstractParadiseFeature(ParadiseFeatureFactory factory, int subFeaturesNb) {
        this.factory = factory;
        this.subfeatures = new ArrayList<ParadiseFeature>(subFeaturesNb);
        //[fjossinet] to have the good size at start (necessary for example to order the ResidueIdentity objects inside a multiple alignment.
        for (int i = 0; i < subFeaturesNb; i++)
            this.subfeatures.add(null);
        this.parentfeatures = new ArrayList<ParadiseFeature>();
        this.annotations = new HashMap<Molecule,Molecule.Annotation>();
        this.paradiseID = new ParadiseID();
    }

    public ParadiseID getParadiseID() {
        return paradiseID;
    }

    protected AbstractParadiseFeature(ParadiseFeatureFactory factory) {
        this(factory, 0);
    }

    public Source getSource() {
        return this.source;
    }

    public void setSource(Source source) {
        this.source = source;
    }

    public void setAssociatedFile(File file) {
        this.associatedFile = file;
    }

    public File getAssociatedFile() {
        return this.associatedFile;
    }

    public ParadiseFeatureFactory getFactory() {
        return factory;
    }

    public List<ParadiseFeature> getSubFeatures() {
        return new ArrayList<ParadiseFeature>(this.subfeatures);
    }

    public <T extends ParadiseFeature> List<T> getSubFeatures(Class<T> c) {
        return CollectionFilter.filter(c, this.getSubFeatures());
    }

    public <T extends ParadiseFeature> List<T> getSubFeatures(Class<T> c, Molecule molecule) {
        List<T> list = new ArrayList<T>();
        for (T feature : getSubFeatures(c))
            if (feature.annotates(molecule))
                list.add(feature);
        return list;
    }

    public <T extends ParadiseFeature> List<T> getSubFeatures(Class<T> c, Molecule molecule, Location location) {
        List<T> list = new ArrayList<T>();
        for (T feature : getSubFeatures(c))
            if (feature.annotates(molecule) && feature.getFullLocation(molecule).matchWith(location))
                list.add(feature);
        return list;
    }

    public <T extends ParadiseFeature> List<T> getSubFeaturesRecursively(Class<T> c) {
        List<T> ret = new ArrayList<T>();
        for (ParadiseFeature subFeature : this.getSubFeatures()) {
            if (c.isInstance(subFeature))
                ret.add(c.cast(subFeature));
            else
                ret.addAll(subFeature.getSubFeaturesRecursively(c));
        }
        return ret;
    }

    public List<ParadiseFeature> getSubFeaturesRecursively() {
        List<ParadiseFeature> ret = new ArrayList<ParadiseFeature>();
        for (ParadiseFeature subFeature : this.getSubFeatures()) {
            ret.add(subFeature);
            ret.addAll(subFeature.getSubFeaturesRecursively());
        }
        return ret;
    }

    public List<ParadiseFeature> getParentFeatures() {
        return new ArrayList<ParadiseFeature>(this.parentfeatures);
    }

    public <T extends ParadiseFeature> List<T> getParentFeatures(Class<T> c) {
        return CollectionFilter.filter(c, this.getParentFeatures());
    }

    public List<Molecule.Annotation> getAnnotations() {
        return new ArrayList<Molecule.Annotation>(this.annotations.values());
    }

    public boolean annotates(Molecule molecule) {
        for (Molecule.Annotation annotation : this.getAnnotations())
            if (annotation.getMolecule()!= null && annotation.getMolecule().getParadiseID().equals(molecule.getParadiseID()))
                return true;
        return false;
    }

    /**
     * Checks recursively if a feature is, at some level, a subfeature of this feature
     *
     * @param subFeature - the subfeature to search
     * @return - true, if it is
     */
    public boolean hasSubFeature(ParadiseFeature subFeature) {
        for (ParadiseFeature feature : this.getSubFeatures()) {
            if (feature == subFeature || feature.hasSubFeature(subFeature))
                return true;
        }
        return false;
    }

    protected void addSubFeature(int index, ParadiseFeature feature) throws ArchitectureException {
        if (feature.getAnnotations().size() == 0)
            throw new ArchitectureException("A feature has to annotate at least one Molecule before to be added as a subfeature", this, feature);
        if (!this.subfeatures.contains(feature)) {
            this.subfeatures.add(index, feature);
            ((AbstractParadiseFeature) feature).parentfeatures.add(this);
            for (Molecule.Annotation a : feature.getAnnotations())
                a.getMolecule().addFeature(this, a.getLocation()); //if this feature already annotates the a.getMolecule(), its Location will be extended (see the addFeature method)
        }
    }

    protected void addSubFeature(ParadiseFeature feature) throws ArchitectureException {
        this.addSubFeature(this.subfeatures.size(), feature);
    }

    public void removeSubFeature(ParadiseFeature feature) {
        if (this.subfeatures.contains(feature)) {
            this.subfeatures.remove(feature);
            ((AbstractParadiseFeature) feature).parentfeatures.remove(this);
        }
    }

    /**
     * Return the location of the annotation for this ParadiseFeature and for the molecule given as argument
     * Return the union of its own location and the location for the molecule given as argument for each sub-features
     * WARNING: returned a a new Location, so if modified by the calling method, the original location is not modified
     * @param molecule
     * @return an empty location if this feature doesn't annotate the molecule given as argument.
     */
    public Location getFullLocation(Molecule molecule) {
        Location l = new Location();
        Molecule.Annotation a = this.annotations.get(molecule);
        if (a != null)
            l.add(a.getLocation());
        for (ParadiseFeature f:this.subfeatures)
            if (!MainParadiseFeature.class.isInstance(f))
                l.add(f.getFullLocation(molecule));
        return l;
    }

    /**
     * This method should not be specialized or we can have a problem with the relocateTo method of Molecule.
     *
     * @param a
     * @throws ArchitectureException
     */
    public void addAnnotation(Molecule.Annotation a) throws fr.unistra.ibmc.paradise.ArchitectureException {
        if (a.getFeature().equals(this) && !this.annotations.containsValue(a)) {
            //we pass this test only if we come directly from the addFeature method provided by Molecule. So the test between the location and the molecule boundaries has been done
            this.annotations.put(a.getMolecule(),a);
            /*
            for (ParadiseFeature f : this.getParentFeatures())
                f.addToLocation(a.getLocation(), a.getMolecule());*/
            //same result if we do a.getMolecule().addFeature(f,a.getLocation());
        }
    }


    /**
     * This method can be specialized to do some additional cleaning steps.
     * Example: if an helix has been deleted, and if this helix annotated a pseudoknot, this pseudoknot cannot exist anymore and have to be deleted too
     *
     * @param a
     */
    public void removeAnnotation(Molecule.Annotation a) {
        if (this.annotations.containsValue(a)) {
            this.annotations.remove(a.getMolecule());
            if (a.getMolecule() != null) /*that's why the molecule reference for this Annotation has to be setted to null once this method realized)*/ {
                a.getMolecule().removeFeature(this);
                /* we have decided not to do that
                for (ParadiseFeature f:this.getParentFeatures())
                    f.removeFromLocation(a.getLocation(),a.getMolecule());*/
            }
            if (this.annotations.size() == 0) {
                //this feature is useless
                for (ParadiseFeature f : this.parentfeatures)
                    ((AbstractParadiseFeature) f).subfeatures.remove(this);
                this.parentfeatures.clear();
                for (ParadiseFeature f : this.subfeatures)
                    ((AbstractParadiseFeature) f).parentfeatures.remove(this);
                this.subfeatures.clear();
            }
        }
    }

    public void addToLocation(Location l, Molecule m) throws ArchitectureException {
        if (l.getEnd() > m.getLength())
            throw new fr.unistra.ibmc.paradise.ArchitectureException("Location to add to a feature outside of a molecule boundaries", m, l, this);
        Molecule.Annotation a = m.getAnnotation(this);
        if (a != null)
            a.getLocation().add(l);
        //if this feature has never annotated this molecule before => new Annotation for this Molecule and this feature
        else
            m.addFeature(this, l);
    }

    public void removeFromLocation(Location l, Molecule m) {
        Molecule.Annotation a = m.getAnnotation(this);
        if (a != null) {
            //cannot remove more than its own location
            l = a.getLocation().intersectionOf(l);
            //we cannot remove a Location overlapping the Location of a registered subfeatures
            for (ParadiseFeature f : this.subfeatures)
                for (Molecule.Annotation _a : f.getAnnotations())
                    if (m == _a.getMolecule()) {
                        l = l.differenceOf(_a.getLocation());
                    }
            a.getLocation().remove(l);
            for (ParadiseFeature f : this.getParentFeatures())
                f.removeFromLocation(l, m);
            if (a.getLocation().isEmpty())
                m.removeFeature(this);
        }
    }

    /**
     * Return all the molecules annotated by this ParadiseFeature
     *
     * @return a Collection keeping the ordered according to their paradise lsid
     */
    public Collection<Molecule> getMolecules() {
        Set<Molecule> molecules = new TreeSet<Molecule>(new Comparator() {

            public int compare(Object o1, Object o2) {
                return ((Molecule) o1).getParadiseID().toString().compareTo(((Molecule) o2).getParadiseID().toString());
            }
        });
        molecules.addAll(this.annotations.keySet());
        for (ParadiseFeature f:this.subfeatures) {
            if (!MainParadiseFeature.class.isInstance(f)) //we don't want to gather the molecules of a sub-mainfeature 
                molecules.addAll(f.getMolecules());
        }
        return molecules;
    }

    public boolean contains(Residue r) {
        Location l;
        for (Molecule m : this.getMolecules()) {
            if (m.getParadiseID().equals(r.getMoleculeId())) {
                l = this.getFullLocation(m);
                if (l.hasPosition(r.getAbsolutePosition()))
                    return true;
            }
        }
        return false;
    }

    public boolean hasNoSubFeatures() {
        return this.subfeatures.size() == 0;
    }

    public boolean hasNoParentFeatures() {
        return this.parentfeatures.size() == 0;
    }

    public boolean annotatesNothing() {
        return this.annotations.size() == 0;
    }

    public Residues getResidues() {
        Residues residues = new Residues();
        int[] boundaries;
        for (Molecule m : this.getMolecules()) {
            boundaries = this.getFullLocation(m).getBoundaries();
            for (int i = 0; i < boundaries.length; i += 2)
                for (int j = boundaries[i]; j <= boundaries[i + 1]; j++)
                    residues.add(new Residue(j, m));
        }
        return residues;
    }

    public boolean equals(Object o) {
        if (!ParadiseFeature.class.isInstance(o))
            return false;
        return this.getParadiseID().equals(((ParadiseFeature) o).getParadiseID());
    }


}
