package fr.unistra.ibmc.paradise.core.features;

import fr.unistra.ibmc.paradise.core.Location;
import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.utils.Residue;
import fr.unistra.ibmc.paradise.ArchitectureException;

import java.util.Iterator;
import java.util.List;

public class Helix extends StructuralDomain {

    private boolean isPseudoKnot;

    protected Helix(StructuralFeatureFactory factory, String name) {
        super(factory, name);
    }

    public boolean isPseudoKnot() {
        return this.isPseudoKnot;
    }

    public void isPseudoKnot(boolean isPseudoKnot) {
        this.isPseudoKnot = isPseudoKnot;
    }

    public void removeBaseBaseInteraction(BaseBaseInteraction interaction) {
        interaction.removeAtomAtomInteractions();
        for (Molecule m : interaction.getMolecules())
            m.removeFeature(interaction);
        this.removeSubFeature(interaction);
    }

    public List<BaseBaseInteraction> getBaseBaseInteractions() throws ArchitectureException {
        return this.getSubFeatures(BaseBaseInteraction.class);
    }

    public Residue get5PrimeEnd(Residue _3primeEnd) {
        Residue[] _5PrimeEnds = this.get5PrimeEnds();
        if (_5PrimeEnds[0].getMoleculeId().equals(_3primeEnd.getMoleculeId()) && (_5PrimeEnds[0].getAbsolutePosition() + this.getLength() - 1) == _3primeEnd.getAbsolutePosition())
            return _5PrimeEnds[0];
        else
        if (_5PrimeEnds[1].getMoleculeId().equals(_3primeEnd.getMoleculeId()) && (_5PrimeEnds[1].getAbsolutePosition() + this.getLength() - 1) == _3primeEnd.getAbsolutePosition())
            return _5PrimeEnds[1];
        else
            return null;
    }

    public Residue get3PrimeEnd(Residue _5primeEnd) {
        Residue[] _3PrimeEnds = this.get3PrimeEnds();
        if (_3PrimeEnds[0].getMoleculeId().equals(_5primeEnd.getMoleculeId()) && _3PrimeEnds[0].getAbsolutePosition() == (_5primeEnd.getAbsolutePosition() + this.getLength() - 1))
            return _3PrimeEnds[0];
        else
        if (_3PrimeEnds[1].getMoleculeId().equals(_5primeEnd.getMoleculeId()) && _3PrimeEnds[1].getAbsolutePosition() == (_5primeEnd.getAbsolutePosition() + this.getLength() - 1))
            return _3PrimeEnds[1];
        else
            return null;
    }

    public Residue getPairedResidue(Residue r) {
        for (BaseBaseInteraction interaction : this.getBaseBaseInteractions())
            if (interaction.getResidue().equals(r))
                return interaction.getPartnerResidue();
            else if (interaction.getPartnerResidue().equals(r))
                return interaction.getResidue();
        return null;
    }

    /**
     * Return the 5 prime ends of the helix. This method doesn't guarantee to return the ends in the same order as the get3PrimeEnds method (meaning that the 5-prime end Residue returned at position 0 by this method doesn't necessary belong to the same strand of the 3-prime end Residue returned at position 0 by the get3PrimeEnds method).
     * To have a perfect correspondance, used the get5PrimeEnd and get3PrimeEnd methods:
     * <i>
     * Residue2D[] _5PrimeEnds = this.get5PrimeEnds();
     * Residue2D _3PrimeEnd1 = this.get3PrimeEnd(_5PrimeEnds[0]),_3PrimeEnd2 = this.get3PrimeEnd(_5PrimeEnds[1]);
     * </i>
     * Between two consecutive calls of this method, the order of the Residue objects returned can change too.
     *
     * @return
     */
    public Residue[] get5PrimeEnds() {
        Residue[] residues = new Residue[2];
        if (this.getMolecules().size() == 1) {
            Molecule m = this.getMolecules().iterator().next();
            Location l = this.getFullLocation(m);
            //to fix the problem of an helix with contiguous strands
            if (l.getBoundaries().length != 4) {
                residues[0] = new Residue(l.getStart(), m);
                residues[1] = new Residue(l.getEnd() - l.getLength() / 2 + 1, m);
            } else {
                residues[0] = new Residue(l.getBoundaries()[0], m);
                residues[1] = new Residue(l.getBoundaries()[2], m);
            }
        } else if (this.getMolecules().size() == 2) {
            Iterator<Molecule> it = this.getMolecules().iterator();
            Molecule m1 = it.next(), m2 = it.next();
            Location l1 = this.getFullLocation(m1), l2 = this.getFullLocation(m2);
            residues[0] = new Residue(l1.getStart(), m1);
            residues[1] = new Residue(l2.getStart(), m2);
        }
        return residues;
    }

    public Residue[] get3PrimeEnds() {
        Residue[] residues = new Residue[2];
        if (this.getMolecules().size() == 1) {
            Molecule m = this.getMolecules().iterator().next();
            Location l = this.getFullLocation(m);
            //to fix the problem of an helix with contiguous strands
            if (l.getBoundaries().length != 4) {
                residues[0] = new Residue(l.getStart() + l.getLength() / 2 - 1, m);
                residues[1] = new Residue(l.getEnd(), m);
            } else {
                residues[0] = new Residue(l.getBoundaries()[1], m);
                residues[1] = new Residue(l.getBoundaries()[3], m);
            }
        } else if (this.getMolecules().size() == 2) {
            Iterator<Molecule> it = this.getMolecules().iterator();
            Molecule m1 = it.next(), m2 = it.next();
            Location l1 = this.getFullLocation(m1), l2 = this.getFullLocation(m2);
            residues[0] = new Residue(l1.getEnd(), m1);
            residues[1] = new Residue(l2.getEnd(), m2);
        }
        return residues;
    }

    /**
     * The length of an Helix is the half of its Location
     *
     * @return
     */
    public int getLength() {
        Iterator<Molecule> it = this.getMolecules().iterator();
        int length = this.getFullLocation(it.next()).getLength();
        //if interMolecular helix
        if (it.hasNext())
            length += this.getFullLocation(it.next()).getLength();
        return length / 2;
    }

}
