package fr.unistra.ibmc.paradise.core.features;

import fr.unistra.ibmc.paradise.BiologicalSymbolException;
import fr.unistra.ibmc.paradise.ArchitectureException;
import fr.unistra.ibmc.paradise.core.Molecule;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

/**
 * User: fjossinet
 * Date: 6 janv. 2007
 */
public abstract class Residue3D extends AbstractParadiseFeature {

    protected List<Atom> atoms;
    protected List<TorsionAngle> torsionAngles;
    protected String name;
    protected boolean isModified;

    protected Residue3D(StructuralFeatureFactory factory, String name) {
        super(factory);
        this.atoms = new ArrayList<Atom>();
        this.torsionAngles = new ArrayList<TorsionAngle>();
        this.name = name;
    }

    public StructuralFeatureFactory getFactory() {
        return (StructuralFeatureFactory) super.getFactory();
    }

    public boolean isModified() {
        return this.isModified;
    }

    public Atom setAtomCoordinates(String atomName, float x, float y, float z) {
        Atom a = this.getAtom(atomName);
        if (a != null)
            a.setCoordinates(x, y, z);
        else {
            a = new Residue3D.Atom(atomName);
            this.atoms.add(a);
            a.setCoordinates(x, y, z);
            this.isModified = true;
        }
        return a;
    }

    public List<TorsionAngle> getTorsionAngle() {
        return new ArrayList<TorsionAngle>(this.torsionAngles);
    }

    /**
     * Return a TorsionAngle registered with its id given as argument
     *
     * @param angleName
     * @return null if no TorsionAngle registered with this id
     */
    public TorsionAngle getTorsionAngle(String angleName) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        for (TorsionAngle angle : this.torsionAngles)
            if (angle.getName().equals(angleName))
                return angle;
        throw new BiologicalSymbolException(angleName + " is an unknown torsion angle for " + this.name, angleName, this);
    }

    public List<Atom> getAtoms() {
        return new ArrayList<Atom>(this.atoms);
    }

    /**
     * Return an Atom registered with its id given as argument. The id of an atom has to be according to the PDB file format
     *
     * @param atomName
     * @return null if no Atom registered with this id
     */
    public Atom getAtom(String atomName) {
        for (Atom a : this.atoms) {
            if (a.getName().equals(atomName))
                return a;
            if (a.getName().equals("O1P") && atomName.equals("OP1"))
                return a;
            if (a.getName().equals("OP1") && atomName.equals("O1P"))
                return a;
            if (a.getName().equals("O2P") && atomName.equals("OP2"))
                return a;
            if (a.getName().equals("OP2") && atomName.equals("O2P"))
                return a;
            if (a.getName().equals("O3P") && atomName.equals("OP3"))
                return a;
            if (a.getName().equals("OP3") && atomName.equals("O3P"))
                return a;
        }
        return null;
    }

    /**
     * This method is used internally to remove atoms during the construction of some specialized Residue3D concepts
     *
     * @param atomName
     */
    protected void removeAtom(String atomName) {
        Atom hit = null;
        for (Atom a : this.atoms)
            if (atomName.equals(a.getName()))
                hit = a;
        this.atoms.remove(hit);
    }

    public List<String> getAtomNames() {
        List<String> names = new ArrayList<String>(this.atoms.size());
        for (Atom a : this.atoms)
            names.add(a.getName());
        return names;
    }

    public List<String> getTorsionAngleNames() {
        List<String> names = new ArrayList<String>(this.torsionAngles.size());
        for (TorsionAngle a : this.torsionAngles)
            names.add(a.getName());
        return names;
    }

    public String getName() {
        return name;
    }

    public String getLabel() {
        TertiaryStructure ts = this.getParentFeatures(TertiaryStructure.class).get(0);
        Molecule m = this.getMolecules().iterator().next();
        return ts.getResidueLabel(this.getAnnotations().get(0).getLocation().getStart(),m);
    }

    abstract public Residue3D createNewResidue3D(char n) throws BiologicalSymbolException, ArchitectureException;

    public class Atom implements Serializable {

        private String name;
        private float[] coordinates;

        protected Atom(String atomName) {
            this.name = atomName.replace('*', '\'');
        }

        /**
         * The array of coordinates is reinitialized to null. After a call to this method, the hasCoordinatesFilled() method will return false
         */
        public void eraseCoordinates() {
            this.coordinates = null;
        }

        public String getName() {
            return this.name;
        }

        public void setCoordinates(float x, float y, float z) {
            if (this.coordinates == null) {
                this.coordinates = new float[]{x, y, z};
            } else {
                this.coordinates[0] = x;
                this.coordinates[1] = y;
                this.coordinates[2] = z;
            }
        }

        public Residue3D getResidue3D() {
            return Residue3D.this;
        }

        public float getX() {
            return this.coordinates[0];
        }

        public float getY() {
            return this.coordinates[1];
        }

        public float getZ() {
            return this.coordinates[2];
        }

        public boolean hasCoordinatesFilled() {
            return this.coordinates != null;
        }
    }

    public class TorsionAngle implements Serializable {

        private int value;
        private String name;

        protected TorsionAngle(String name) {
            this.name = name;
        }

        public void setValue(int value) {
            this.value = value;
        }

        public int getValue() {
            return this.value;
        }

        public String getName() {
            return this.name;
        }
    }
}
