package fr.unistra.ibmc.paradise.core.features;

import Jama.Matrix;
import fr.unistra.ibmc.paradise.ArchitectureException;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;

import java.util.List;

public abstract class RiboNucleotide3D extends Residue3D {

    public static final int C3ENDO = 0, C2ENDO = 1;

    protected int sugarPucker;

    protected RiboNucleotide3D(StructuralFeatureFactory factory, String name) {
        super(factory, name);
        this.sugarPucker = RiboNucleotide3D.C3ENDO;
        this.atoms.add(new Atom("P"));
        this.atoms.add(new Atom("O1P"));
        this.atoms.add(new Atom("O2P"));
        this.atoms.add(new Atom("O3P"));
        this.atoms.add(new Atom("OP1"));
        this.atoms.add(new Atom("OP2"));
        this.atoms.add(new Atom("OP3"));
        //if tri-phosphate (i.e. SARS PDB provided with S2S)
        this.atoms.add(new Atom("PA"));
        this.atoms.add(new Atom("O1A"));
        this.atoms.add(new Atom("O2A"));
        this.atoms.add(new Atom("O3A"));
        this.atoms.add(new Atom("PB"));
        this.atoms.add(new Atom("O1B"));
        this.atoms.add(new Atom("O2B"));
        this.atoms.add(new Atom("O3B"));
        this.atoms.add(new Atom("PG"));
        this.atoms.add(new Atom("O1G"));
        this.atoms.add(new Atom("O2G"));
        this.atoms.add(new Atom("O3G"));
        this.atoms.add(new Atom("C1'"));
        this.atoms.add(new Atom("H1'"));
        this.atoms.add(new Atom("C2'"));
        this.atoms.add(new Atom("O2'"));
        this.atoms.add(new Atom("1H2'"));
        this.atoms.add(new Atom("2HO'"));
        this.atoms.add(new Atom("C3'"));
        this.atoms.add(new Atom("H3'"));
        this.atoms.add(new Atom("O3'"));
        this.atoms.add(new Atom("C4'"));
        this.atoms.add(new Atom("H4'"));
        this.atoms.add(new Atom("O4'"));
        this.atoms.add(new Atom("C5'"));
        this.atoms.add(new Atom("1H5'"));
        this.atoms.add(new Atom("2H5'"));
        this.atoms.add(new Atom("O5'"));
        this.atoms.addAll(this.getDefaultBaseAtoms(false));
    }

    public Atom setAtomCoordinates(String atomName, float x, float y, float z) {
        Atom a;
        a = super.setAtomCoordinates(atomName, x, y, z);
        //TODO each time some Atom coordinates are registered, check if a new TorsionAngle can be calculated
        return a;
    }

    public int getSugarPucker() {
        return sugarPucker;
    }

    public void setSugarPucker(int sugarPucker) {
        this.sugarPucker = sugarPucker;
    }

    public RiboNucleotide3D createNewResidue3D(char n) throws BiologicalSymbolException, ArchitectureException {
        RiboNucleotide3D r;
        //copy the atoms of the sugar
        switch (Character.toUpperCase(n)) {
            case 'A':
                r = new Adenine3D(this.getFactory());
                break;
            case 'C':
                r = new Cytosine3D(this.getFactory());
                break;
            case 'G':
                r = new Guanine3D(this.getFactory());
                break;
            case 'U':
                r = new Uridine3D(this.getFactory());
                break;
            default:
                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Can only create a residue for 'A', 'C', 'G' or 'U'", n + "", this);
        }
        //first step, we copy the coordinates for the common atoms between the source residue and the dest one (ribose+phosphate)
        for (Atom atom : this.getAtoms()) {
            if (atom.hasCoordinatesFilled()) {
                //looking for a match,
                Atom atom2 = r.getAtom(atom.getName());
                if (atom2 != null)
                    atom2.setCoordinates(atom.getX(), atom.getY(), atom.getZ());
                /*
                else
                    System.out.println("null atom2 for "+atom.getName()+" in "+r);*/
            }
        }
        //now the atoms of the base => we need to fit the fold of the current residue to the dest one
        //first step, we calculate the transformation matrix between a reference (calculated by NAHELIX) of the current residue with the current residue
        Matrix t = this.getTransformationMatrix();
        //then we apply this matrix to 
        r.applyTransformationMatrix(t);
        return r;
    }

    /**
     * tludwig
     * Gets the transformation matrix from the reference base, to the actual base
     *
     * @return the transformation matrix
     */
    private Matrix getTransformationMatrix() throws BiologicalSymbolException {
        List<Atom> ref = getDefaultBaseAtoms(true);
        Atom s1 = ref.get(0);
        Atom s2 = ref.get(1);
        Atom s3 = ref.get(2);
        Atom s4 = ref.get(3);

        Atom d1 = this.getAtom(s1.getName());
        Atom d2 = this.getAtom(s2.getName());
        Atom d3 = this.getAtom(s3.getName());
        Atom d4 = this.getAtom(s4.getName());

        double[][] mr = {{s1.getX(), s2.getX(), s3.getX(), s4.getX()}, {s1.getY(), s2.getY(), s3.getY(), s4.getY()}, {s1.getZ(), s2.getZ(), s3.getZ(), s4.getZ()}, {1, 1, 1, 1}};
        double[][] mo = {{d1.getX(), d2.getX(), d3.getX(), d4.getX()}, {d1.getY(), d2.getY(), d3.getY(), d4.getY()}, {d1.getZ(), d2.getZ(), d3.getZ(), d4.getZ()}, {1, 1, 1, 1}};
        Matrix r = new Matrix(mr);
        Matrix o = new Matrix(mo);
        return o.times(r.inverse());
    }

    /**
     * tludwig
     * Apply a given transformation matrix to the atoms of the base
     *
     * @param t
     */
    private void applyTransformationMatrix(Matrix t) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        for (Atom atom : getDefaultBaseAtoms(true)) {
            Atom atom2 = this.getAtom(atom.getName());
            float[] v = new float[]{atom.getX(), atom.getY(), atom.getZ()};
            float[] f = new float[]
                    {
                            (float) (v[0] * t.get(0, 0) + v[1] * t.get(0, 1) + v[2] * t.get(0, 2) + t.get(0, 3)),
                            (float) (v[0] * t.get(1, 0) + v[1] * t.get(1, 1) + v[2] * t.get(1, 2) + t.get(1, 3)),
                            (float) (v[0] * t.get(2, 0) + v[1] * t.get(2, 1) + v[2] * t.get(2, 2) + t.get(2, 3))
                    };
            atom2.setCoordinates(f[0], f[1], f[2]);
        }
    }

    /**
     * tludwig
     * Return a list of atoms constituting the base group with their default coordinates generated by the NAHELIX algorithm
     */
    protected abstract List<Atom> getDefaultBaseAtoms(boolean withDefaultCoordinates);
}
