package fr.unistra.ibmc.paradise.core.features;

import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.ParadiseID;
import fr.unistra.ibmc.paradise.core.Source;
import fr.unistra.ibmc.paradise.core.utils.Residue;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;


public abstract class SecondaryStructure extends AbstractMainParadiseFeature {

    protected SecondaryStructure(ParadiseID paradiseID, StructuralFeatureFactory factory, String name, Source source) {
        super(paradiseID, factory, name, source);
    }

    public StructuralFeatureFactory getFactory() {
        return (StructuralFeatureFactory) super.getFactory();
    }

    public List<AtomAtomInteraction> getSingleHBonds() {
        return getSubFeatures(AtomAtomInteraction.class);
    }

    /**
     * Return the tertiary interactions only
     *
     * @return
     */
    //[fjossinet] the BaseBaseInteraction directly under a SecondaryStructure are tertiary interactions
    public List<BaseBaseInteraction> getTertiaryInteractions() {
        return this.getSubFeatures(BaseBaseInteraction.class);
    }

    public void addTertiaryInteraction(BaseBaseInteraction bbi) {
        this.addSubFeature(bbi);
    }


    public List<BaseBaseInteraction> getSecondaryInteractions() {
        List<BaseBaseInteraction> secondaryInteractions = new ArrayList<BaseBaseInteraction>();
        for (Helix h : this.getHelices())
            secondaryInteractions.addAll(h.getBaseBaseInteractions());
        return secondaryInteractions;
    }

    public List<BaseBaseInteraction> getBaseBaseInteractions() {
        List<BaseBaseInteraction> allInteractions = new ArrayList<BaseBaseInteraction>();
        allInteractions.addAll(this.getSubFeatures(BaseBaseInteraction.class));
        for (Helix h : this.getHelices())
            allInteractions.addAll(h.getBaseBaseInteractions());
        return allInteractions;
    }

    public List<AtomAtomInteraction> getAtomAtomInteractions() {
        List<AtomAtomInteraction> allInteractions = new ArrayList<AtomAtomInteraction>();
        allInteractions.addAll(this.getSubFeatures(AtomAtomInteraction.class));
        for (BaseBaseInteraction bbi : this.getBaseBaseInteractions())
            allInteractions.addAll(bbi.getAtomAtomInteractions());
        return allInteractions;
    }

    public List<SingleStrand> getSingleStrands() {
        return this.getSubFeatures(SingleStrand.class);
    }

    public void removeSingleStrand(SingleStrand ss) {
        for (Molecule m : ss.getMolecules())
            m.removeFeature(ss);
        this.removeSubFeature(ss);
    }


    public void removeBaseBaseInteraction(BaseBaseInteraction interaction) {
        interaction.removeAtomAtomInteractions();
        for (Molecule m : interaction.getMolecules())
            m.removeFeature(interaction);
        this.removeSubFeature(interaction);
    }

    public List<TertiaryStructure> getLinkedTertiaryStructures() {
        return this.getSubFeatures(TertiaryStructure.class);
    }

    /**
     * Return all the helices, even the pseudoknots
     *
     * @return
     */
    public List<Helix> getHelices() {
        return new ArrayList(this.getSubFeatures(Helix.class));
    }

    /**
     * Return all the helices vithout the pseudoknots
     *
     * @return
     */
    public List<Helix> getHelicesWithoutPKnots() {
        Set<Helix> allHelices = new HashSet<Helix>();
        for (Helix h : this.getSubFeatures(Helix.class))
            if (!h.isPseudoKnot())
                allHelices.add(h);
        return new ArrayList(allHelices);
    }

    public void removeHelix(Helix h) {
        for (BaseBaseInteraction interaction :h.getBaseBaseInteractions())
            h.removeBaseBaseInteraction(interaction);
        for (Molecule m : h.getMolecules())
            m.removeFeature(h);
        this.removeSubFeature(h);
    }

    public void removeSecondaryStructureDisplay(SecondaryStructureDisplay display) {
        for (Molecule m : display.getMolecules())
            m.removeFeature(display);
        super.removeSubFeature(display);
    }

    public SecondaryStructureDisplay getSecondaryStructureDisplay() {
        List<SecondaryStructureDisplay> displays = this.getSubFeatures(SecondaryStructureDisplay.class);
        if (displays.isEmpty())
            return null;
        else
            return displays.get(0);
    }

    /**
     * Return the StructuralDmomain containing the residue given as argument
     *
     * @param r
     * @return null if no such StructuralDomain
     */
    public StructuralDomain getEnclosingStructuralDomain(Residue r) {
        for (Helix h : this.getHelices())
            if (h.contains(r))
                return h;
        for (SingleStrand ss : this.getSingleStrands())
            if (ss.contains(r))
                return ss;
        return null;
    }

    public Residue getPairedResidueInSecondaryInteraction(Residue r) {
        for (Helix h : this.getHelices()) {
            for (BaseBaseInteraction interaction : h.getBaseBaseInteractions()) {
                if (interaction.getResidue().equals(r)) {
                    return interaction.getPartnerResidue();
                } else if (interaction.getPartnerResidue().equals(r)) {
                    return interaction.getResidue();
                }
            }
        }
        return null;
    }

    /**
     * A SecondaryStructure can be asociated to several structural alignments
     *
     * @param alignment
     */
    public void linkStructuralAlignment(StructuralAlignment alignment) {
        this.addSubFeature(alignment);
    }

    public List<StructuralAlignment> getLinkedStructuralAlignments() {
        return this.getSubFeatures(StructuralAlignment.class);
    }

    public List<Helix> getPseudoKnots() {
        List<Helix> pknots = new ArrayList<Helix>();
        for (Helix h : this.getHelices())
            if (h.isPseudoKnot())
                pknots.add(h);
        return pknots;
    }

    /**
     * Return the secondary structure for a given molecule as a bracket notation
     *
     * @param m the molecule for which we want the bracket notation
     * @return a String describing the secondary structure as a bracket notation
     */
    public String printAsBracketNotation(Molecule m) {
        char[] ret = new char[m.getLength()];
        //initialisation, unpaired residues
        for (int i = 0; i < m.getLength(); i++)
            ret[i] = '.';
        //pairing the residues
        for (Helix helix : this.getHelices()) {
            //TODO [fjossinet] URGENT => compare what helix.getMolecules().size() return with helix.getAnnotations().size() for a ss generated by a service like rnaview
            if (helix.annotates(m) && helix.getMolecules().size() == 1)
                for (BaseBaseInteraction bb : helix.getBaseBaseInteractions()) {
                    ret[bb.getFullLocation(m).getStart() - 1] = '(';
                    ret[bb.getFullLocation(m).getEnd() - 1] = ')';
                }
        }
        return new String(ret);
    }

}
