package fr.unistra.ibmc.paradise.core.features;

import fr.unistra.ibmc.paradise.core.Location;
import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.ParadiseID;
import fr.unistra.ibmc.paradise.core.Source;
import fr.unistra.ibmc.paradise.core.utils.Residue;

import java.util.*;

/**
 * A MultipleAlignment
 *
 * @author fjossinet
 */
public class StructuralAlignment extends AbstractMainParadiseFeature {

    private Map<Molecule, Integer> positions ;

    protected StructuralAlignment(ParadiseID paradiseID, IdentityFeatureFactory factory, String name, Source source) {
        super(paradiseID, factory, name, source);
        this.positions = new HashMap<Molecule, Integer>();
    }

    public IdentityFeatureFactory getFactory() {
        return (IdentityFeatureFactory) super.getFactory();
    }

    /**
     * Return the position for a molecule target which is homologous to a position of a molecule source
     * @param source
     * @param position
     * @param target
     * @return -1 if no homologous position
     */
    public int getHomologousPosition(Molecule source, int position, Molecule target) {
        for (StructuralIdentity si:this.getStructuralIdentities(true)) {
            Location sourceLocation = si.getFullLocation(source), targetLocation = si.getFullLocation(target),
                    sourceGaps = si.getGapsLocation(source), targetGaps = si.getGapsLocation(target);
            if (sourceLocation != null && sourceLocation.hasPosition(position) && targetLocation != null) {
                int sourcePos = sourceLocation.getStart()-1, targetPos = targetLocation.getStart()-1;
                for (int i = 1 ; i <= si.getLength() ; i++) {
                    if (sourceGaps == null || !sourceGaps.hasPosition(i)) {
                        sourcePos++;
                        if (targetGaps == null || !targetGaps.hasPosition(i)) {
                            targetPos++;
                            if (sourcePos == position)
                                return targetPos;
                        }
                    }
                }
            }
        }
        return -1;
    }

    public List<StructuralIdentity> getStructuralIdentities(boolean ordered) {
        if (ordered) {
            List<StructuralIdentity> ret = new ArrayList<StructuralIdentity>();
            for (StructuralIdentity identity : this.getSubFeatures(StructuralIdentity.class)) {
                boolean added = false;
                for (StructuralIdentity ri : ret) {
                    if (identity.isBefore(ri)) {
                        ret.add(ret.indexOf(ri), identity);
                        added = true;
                        break;
                    }
                }
                if (!added)
                    ret.add(identity);
            }
            return ret;
        } else
            return this.getSubFeatures(StructuralIdentity.class);
    }

    public void removeResidueIdentity(ResidueIdentity ri) {
        for (Molecule m : ri.getMolecules())
            m.removeFeature(ri);
        this.removeSubFeature(ri);
    }

    public void removeStructuralIdentity(StructuralIdentity ri) {
        for (Molecule m : ri.getMolecules())
            m.removeFeature(ri);
        this.removeSubFeature(ri);
    }

    public void linkSecondaryStructure(SecondaryStructure ss) {
        this.addSubFeature(ss);
    }

    public List<SecondaryStructure> getLinkedSecondaryStructures() {
        return this.getSubFeatures(SecondaryStructure.class);
    }

    public ResidueIdentity getResidueIdentity(Residue residue) {
        for (ResidueIdentity ri : this.getSubFeatures(ResidueIdentity.class)) {
            if (residue.getMolecule().getAnnotation(ri).getLocation().getStart() == residue.getAbsolutePosition())
                return ri;
        }
        return null;
    }

    public int getLength() {
        int length = 0;
        for (StructuralIdentity si:this.getStructuralIdentities(false))
            length+=si.getLength();
        return length;
    }

    public void setPositionInAlignment(Molecule m, int position) {
        this.positions.put(m,position);
    }

    public Integer getPositionInAlignment(Molecule m) {
        return this.positions.get(m);
    }

    public List<Molecule> getSortedMolecules() {
        if (this.positions.size() == 0) {
            Set<Molecule> molecules = new HashSet<Molecule>();
            for (StructuralIdentity si:this.getSubFeatures(StructuralIdentity.class))
                molecules.addAll(si.getMolecules());
            return new ArrayList<Molecule>(molecules);
        }
        else {
            List<Molecule> molecules = new ArrayList<Molecule>(this.positions.size());
            for (int i=0; i<this.positions.size();i++)
                molecules.add(null);
            for (Map.Entry<Molecule, Integer> e:this.positions.entrySet())
                molecules.set(e.getValue(),e.getKey());
            return molecules;
        }
    }
}
