package fr.unistra.ibmc.paradise.core.features;

import fr.unistra.ibmc.paradise.ArchitectureException;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;
import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.utils.Residue;

import java.util.*;

public final class StructuralFeatureFactory extends AbstractParadiseFeatureFactory {

    private static boolean CREATE_PKNOTS = false;

    public StructuralFeatureFactory(MoleculeFactory moleculeFactory, String description) {
        super(moleculeFactory,"Structure", description);
    }

    public Helix createHelix(String name, SecondaryStructure secondaryStructure, Molecule m, Location l) throws fr.unistra.ibmc.paradise.ArchitectureException {
        Helix h = null;
        if (l.getLength() < 4)
            throw new fr.unistra.ibmc.paradise.ArchitectureException("An Helix contains at least two base-pairs(4 residues)");
        int boundaries[] = l.getBoundaries();
        //[fjossinet] sometimes you can have an helix with two strands directly linked (for example an Helix 66-72, 72-79). This is the case with the RNAML produced by RNAVIEW for the PDB 1u9s.pdb)
        if (boundaries.length != 4) {
            //[fjossinet] to fix the problem of an helix with contiguous strands
            boundaries = new int[4];
            boundaries[0] = l.getStart();
            boundaries[1] = l.getStart() + l.getLength() / 2 - 1;
            boundaries[2] = l.getEnd() - l.getLength() / 2 + 1;
            boundaries[3] = l.getEnd();
            //throw new ArchitectureException("The location of an (intra)helix on 1 molecule must have 4 boundaries");
        }
        if (boundaries[1] - boundaries[0] != boundaries[3] - boundaries[2])
            throw new fr.unistra.ibmc.paradise.ArchitectureException("This helix has strands of different length : " + l);
        //an helix cannot overlap another one already registered
        for (Helix _h : secondaryStructure.getHelices())
            if (_h.annotates(m) && _h.getFullLocation(m).hasAtLeastOnePositionInCommonWith(l))
                throw new fr.unistra.ibmc.paradise.ArchitectureException("An helix cannot overlap another one already registered: helix " + _h.getFullLocation(m) + " in conflict with the upcoming helix " + l + " for molecule " + m.getName());
        h = new Helix(this, name);
        //is this new helix in a pseudoknot ??
        HELICES:
        for (Helix _h : secondaryStructure.getHelicesWithoutPKnots()) {
            if (_h.getMolecules().size() == 1 && _h.annotates(m)) {
                Location _l = _h.getFullLocation(m);
                int s1 = l.getStart();
                int e1 = l.getEnd();
                int s2 = _l.getStart();
                int e2 = _l.getEnd();
                if (e1 < e2 && s1 < s2 && s2 < e1 || e2 < e1 && s2 < s1 && s1 < e2) {
                    //the regular helix is the longest one or the first in the sequence
                    if (l.getLength() > _l.getLength())
                        _h.isPseudoKnot(true);
                    else if (_l.getLength() > l.getLength())
                        h.isPseudoKnot(true);
                        //same length
                    else if (s1 < s2)
                        _h.isPseudoKnot(true);
                    else
                        h.isPseudoKnot(true);
                }
            }
        }
        m.addFeature(h, l);
        secondaryStructure.addSubFeature(h);
        //adding BaseBaseInteractions
        for (int i = 0; i <= boundaries[1] - boundaries[0]; i++) {
            BaseBaseInteraction interaction = new CisWWInteraction(this, BaseBaseInteraction.RESIDUE_1, m, new Location(new Location(l.getStart() + i), new Location(l.getEnd() - i)));
            h.addSubFeature(interaction);
        }
        return h;
    }

    public Helix copyHelix(List<Molecule> targetMolecules, Helix referenceHelix, SecondaryStructure secondaryStructure) throws BiologicalSymbolException {
        Helix copyHelix;
        Iterator<Molecule> iterator = referenceHelix.getMolecules().iterator();
        Molecule _m = iterator.next();
        List<Molecule> _targetMolecules;
        if (iterator.hasNext()) {
            Molecule _m2 = iterator.next();
            copyHelix = this.createHelix(referenceHelix.getName(), secondaryStructure, _m.searchCopy(targetMolecules), new Location(referenceHelix.getFullLocation(_m)), _m2.searchCopy(targetMolecules), new Location(referenceHelix.getFullLocation(_m2)));
        } else {
            _targetMolecules = new ArrayList<Molecule>(targetMolecules);
            Molecule copyM = _m.searchCopy(_targetMolecules);
            _targetMolecules.remove(copyM);
            try {
                copyHelix = this.createHelix(referenceHelix.getName(), secondaryStructure, copyM, new Location(referenceHelix.getFullLocation(_m)));
            }
            catch (ArchitectureException e) {
                Molecule __m =  _m.searchCopy(_targetMolecules);
                copyHelix = this.createHelix(referenceHelix.getName(), secondaryStructure, __m, new Location(referenceHelix.getFullLocation(_m)));
            }
        }
        //the default basebaseinteraction in the copyHelix is replaced by a new one if the original one is a non-canical one
        for (BaseBaseInteraction referenceInteraction : referenceHelix.getBaseBaseInteractions())
            if (!referenceInteraction.isCanonical()) {
                this.copyBaseBaseInteraction(targetMolecules, referenceInteraction, copyHelix);
            }
        return copyHelix;
    }

    public Helix createHelix(String name, SecondaryStructure secondaryStructure, Molecule m1, Location l1, Molecule m2, Location l2) throws fr.unistra.ibmc.paradise.ArchitectureException {
        if (l1.getLength() + l2.getLength() < 4)
            throw new fr.unistra.ibmc.paradise.ArchitectureException("An Helix contains at least two base-pairs(4 residues)");
        if (l1.getLength() != l2.getLength())
            throw new fr.unistra.ibmc.paradise.ArchitectureException("This helix has strands of different length : " + l1 + " " + l2);
        //an helix cannot overlap another one already registered
        for (Helix _h : secondaryStructure.getHelices()) {
            if (_h.annotates(m1) && _h.getFullLocation(m1).hasAtLeastOnePositionInCommonWith(l1))
                throw new fr.unistra.ibmc.paradise.ArchitectureException("An helix cannot overlap another one already registered: helix " + _h.getFullLocation(m1) + " in conflict with the upcoming helix " + l1 + " for molecule " + m1.getName());
            if (_h.annotates(m2) && _h.getFullLocation(m2).hasAtLeastOnePositionInCommonWith(l2))
                throw new fr.unistra.ibmc.paradise.ArchitectureException("An helix cannot overlap another one already registered: helix " + _h.getFullLocation(m2) + " in conflict with the upcoming helix " + l2 + " for molecule " + m2.getName());
        }
        Helix helix = new Helix(this, name);
        //the current SecondaryStructure adds the Helix (and now this SecondaryStructure will also annotate a.getMolecule() for a.getLocation())
        m1.addFeature(helix, l1);
        m2.addFeature(helix, l2);
        secondaryStructure.addSubFeature(helix);
        //adding BaseBaseInteractions
        for (int i = 0; i < l1.getLength(); i++) {
            BaseBaseInteraction interaction = new CisWWInteraction(this, BaseBaseInteraction.RESIDUE_1, m1, new Location(l1.getStart() + i), m2, new Location(l2.getEnd() - i));
            //to avoid redundancy between secondary and tertiary interactions
            boolean found = false;
            for (BaseBaseInteraction _interaction : secondaryStructure.getTertiaryInteractions()) {
                if (interaction.getResidue().equals(_interaction.getResidue()) && interaction.getPartnerResidue().equals(_interaction.getPartnerResidue())) {
                    helix.addSubFeature(_interaction);
                    secondaryStructure.removeBaseBaseInteraction(interaction);
                    found = true;
                    break;
                }
            }
            if (!found)
                helix.addSubFeature(interaction);
        }
        return helix;
    }

    public SingleStrand createSingleStrand(String name, SecondaryStructure secondaryStructure, Molecule m, Location l) {
        SingleStrand ss = new SingleStrand(this, name);
        m.addFeature(ss, l);
        secondaryStructure.addSubFeature(ss);
        return ss;
    }

    public SingleStrand copySingleStrand(List<Molecule> targetMolecules, SingleStrand referenceSingleStrand, SecondaryStructure secondaryStructure) {
        Molecule _m = referenceSingleStrand.getMolecules().iterator().next();
        return this.createSingleStrand(referenceSingleStrand.getName(), secondaryStructure, _m.searchCopy(targetMolecules), new Location(referenceSingleStrand.getFullLocation(_m)));
    }

    public SecondaryStructureDisplay createSecondaryStructureDisplay(SecondaryStructure secondaryStructure, Molecule m, Location l) {
        SecondaryStructureDisplay display = new SecondaryStructureDisplay(this);
        m.addFeature(display, l);
        secondaryStructure.addSubFeature(display);
        return display;
    }

    public SecondaryStructureDisplay copySecondaryStructureDisplay(List<Molecule> targetMolecules, SecondaryStructureDisplay referenceDisplay, SecondaryStructure secondaryStructure) {
        Iterator<Molecule> iterator = referenceDisplay.getMolecules().iterator();
        Molecule _m = iterator.next();
        SecondaryStructureDisplay copyDisplay = this.createSecondaryStructureDisplay(secondaryStructure, _m.searchCopy(targetMolecules), new Location(referenceDisplay.getFullLocation(_m)));
        for (Residue2D referenceResidue2D : referenceDisplay.getResidues2D())
            this.copyResidue2D(targetMolecules, referenceResidue2D, copyDisplay);
        return copyDisplay;
    }

    public Residue2D createResidue2D(SecondaryStructureDisplay display, float x, float y, Molecule m, int position) {
        Residue2D r = new Residue2D(this, x, y);
        m.addFeature(r, new Location(position));
        display.addSubFeature(r);
        return r;
    }

    public Residue2D copyResidue2D(List<Molecule> targetMolecules, Residue2D referenceResidue2D, SecondaryStructureDisplay display) {
        Molecule _m = referenceResidue2D.getMolecules().iterator().next();
        return this.createResidue2D(display, referenceResidue2D.getX(), referenceResidue2D.getY(), _m.searchCopy(targetMolecules), referenceResidue2D.getFullLocation(_m).getStart());
    }

    //[fjossinet] to create an RNA SecondaryStructure without a starting location (only usable if this secondary structure is no created to be a sub-feature, since a feature can be a sub-feature only if it annotates at least one molecule)
    public RNASecondaryStructure createRNASecondaryStructure(ParadiseID paradiseID, String name, Source source) {
        RNASecondaryStructure secondaryStructure = new RNASecondaryStructure(paradiseID, this, name, source);
        return secondaryStructure;
    }

    public RNASecondaryStructure createRNASecondaryStructure(ParadiseID paradiseID, String name, Source source, Molecule m, Location l) {
        RNASecondaryStructure secondaryStructure = new RNASecondaryStructure(paradiseID, this, name, source);
        m.addFeature(secondaryStructure, l);
        return secondaryStructure;
    }

    /**
     * Create a new Secondary Structure and copy the subfeatures of the source secondary structure to the new target secondary structure
     * @param targetMolecules
     * @param sourceSecondaryStructure
     * @param copyPseudoKnots
     * @return
     * @throws BiologicalSymbolException
     */
    public SecondaryStructure copySecondaryStructure(List<Molecule> targetMolecules, SecondaryStructure sourceSecondaryStructure, boolean copyPseudoKnots) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        Iterator<Molecule> iterator = sourceSecondaryStructure.getMolecules().iterator();
        Molecule _m = iterator.next();
        RNASecondaryStructure targetSecondaryStructure = this.createRNASecondaryStructure(sourceSecondaryStructure.getParadiseID(), sourceSecondaryStructure.getName(), sourceSecondaryStructure.getSource(),_m.searchCopy(targetMolecules), new Location(sourceSecondaryStructure.getFullLocation(_m)));
        this.copySubFeaturesToSecondaryStructure(targetMolecules,sourceSecondaryStructure,targetSecondaryStructure,copyPseudoKnots);
        return targetSecondaryStructure;
    }

    /**
     * copy the subfeatures of the source secondary structure to the target secondary structure giben as argument
     * @param targetMolecules
     * @param sourceSecondaryStructure
     * @param targetSecondaryStructure
     * @param copyPseudoKnots
     * @throws BiologicalSymbolException
     */
    public void copySubFeaturesToSecondaryStructure(List<Molecule> targetMolecules,SecondaryStructure sourceSecondaryStructure,SecondaryStructure targetSecondaryStructure, boolean copyPseudoKnots) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        if (copyPseudoKnots) {
            for (Helix referenceHelix : sourceSecondaryStructure.getHelices())
                this.copyHelix(targetMolecules, referenceHelix, targetSecondaryStructure);
            for (SingleStrand referenceSingleStrand : sourceSecondaryStructure.getSingleStrands())
                this.copySingleStrand(targetMolecules, referenceSingleStrand, targetSecondaryStructure);
        } else {
            Map<Molecule, Location> locations = new HashMap<Molecule, Location>();
            for (Helix referenceHelix : sourceSecondaryStructure.getHelicesWithoutPKnots()) {
                Helix h = this.copyHelix(targetMolecules, referenceHelix, targetSecondaryStructure);
                for (Molecule m : h.getMolecules()) {
                    if (locations.containsKey(m))
                        //reduce the current location with the location of the helix for this molecule
                        locations.put(m, locations.get(m).differenceOf(h.getFullLocation(m)));
                    else
                        //create a new entry with the location of the full molecule reduced with the location of the helix for this molecule
                        locations.put(m, new Location(1, m.getLength()).differenceOf(h.getFullLocation(m)));
                }
            }
            //create for each molecule the new single-strands from the locations map
            int i = 0;
            for (Map.Entry<Molecule, Location> e : locations.entrySet()) {
                int[] boundaries = e.getValue().getBoundaries();
                for (int j = 0; j < boundaries.length - 1; j += 2) {
                    this.createSingleStrand("" + (i++), targetSecondaryStructure, e.getKey(), new Location(boundaries[j], boundaries[j + 1]));
                }
            }
            //now the basebaseinteractions linked to the pseudoknots are transformed as tertiary interactions
            for (Helix h : sourceSecondaryStructure.getPseudoKnots())
                for (BaseBaseInteraction interaction : h.getBaseBaseInteractions())
                    this.copyBaseBaseInteraction(targetMolecules, interaction, targetSecondaryStructure);
        }
        for (BaseBaseInteraction referenceInteraction : sourceSecondaryStructure.getTertiaryInteractions())
            this.copyBaseBaseInteraction(targetMolecules, referenceInteraction, targetSecondaryStructure);
        for (AtomAtomInteraction referenceAtomAtomInteraction : sourceSecondaryStructure.getSingleHBonds())
            this.copyAtomAtomInteraction(targetMolecules, referenceAtomAtomInteraction, targetSecondaryStructure);
    }

    public TertiaryStructure createTertiaryStructure(ParadiseID paradiseID, String name, Source source, Molecule m, Location l) {
        TertiaryStructure t = new TertiaryStructure(paradiseID, this, name, source);
        m.addFeature(t, l);
        return t;
    }

    //[fjossinet] to create a TertiaryStructure without a starting location (only usable if this tertiary structure is no created to be a sub-feature, since a feature can be a sub-feature only if it annotates at least one molecule)
    public TertiaryStructure createTertiaryStructure(ParadiseID paradiseID, String name, Source source) {
        TertiaryStructure t = new TertiaryStructure(paradiseID, this, name, source);
        return t;
    }

    public TertiaryStructure copyTertiaryStructure(List<Molecule> targetMolecules, TertiaryStructure referenceTertiaryStructure) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        Iterator<Molecule> iterator = referenceTertiaryStructure.getMolecules().iterator();
        Molecule _m = iterator.next();
        TertiaryStructure copyTertiaryStructure = this.createTertiaryStructure(referenceTertiaryStructure.getParadiseID(), referenceTertiaryStructure.getName(), referenceTertiaryStructure.getSource(), _m.searchCopy(targetMolecules), new Location(referenceTertiaryStructure.getFullLocation(_m)));
        for (Residue3D referenceresidue3D : referenceTertiaryStructure.getResidues3D())
            this.copyResidue3D(targetMolecules, referenceresidue3D, copyTertiaryStructure);
        return copyTertiaryStructure;
    }

    /**
     *
     * @param ss
     * @param atom1 has to be the atom corresponding to the residue at the start of the location l given as argument
     * @param atom2 has to be the atom corresponding to the residue at the end of the location l given as argument
     * @param interactionType
     * @param m
     * @param l
     * @return
     * @throws fr.unistra.ibmc.paradise.ArchitectureException
     */
    public AtomAtomInteraction createAtomAtomInteraction(SecondaryStructure ss, String atom1, String atom2, char interactionType, Molecule m, Location l) throws fr.unistra.ibmc.paradise.ArchitectureException {
        AtomAtomInteraction aai = new AtomAtomInteraction(this);
        aai.setAtom1(new Residue(l.getStart(),m),atom1);
        aai.setAtom2(new Residue(l.getEnd(),m),atom2);
        aai.setInteractionType(interactionType);
        m.addFeature(aai, l);
        ss.addSubFeature(aai);
        return aai;
    }

    public AtomAtomInteraction copyAtomAtomInteraction(List<Molecule> targetMolecules, AtomAtomInteraction referenceInteraction, SecondaryStructure ss) {
        AtomAtomInteraction copyAtomAtomInteraction;
        Iterator<Molecule> iterator = referenceInteraction.getMolecules().iterator();
        Molecule _m = iterator.next();
        if (iterator.hasNext()) {
            Molecule _m2 = iterator.next();
            copyAtomAtomInteraction = this.createAtomAtomInteraction(ss, referenceInteraction.getAtom1(), referenceInteraction.getAtom2(), referenceInteraction.getInteractionType(), _m.searchCopy(targetMolecules), new Location(referenceInteraction.getFullLocation(_m)), _m2.searchCopy(targetMolecules), new Location(referenceInteraction.getFullLocation(_m2)));
        } else
            copyAtomAtomInteraction = this.createAtomAtomInteraction(ss, referenceInteraction.getAtom1(), referenceInteraction.getAtom2(), referenceInteraction.getInteractionType(), _m.searchCopy(targetMolecules), new Location(referenceInteraction.getFullLocation(_m)));
        return copyAtomAtomInteraction;
    }

    /**
     *
     * @param bbi
     * @param atom1 has to be the atom corresponding to the residue at the start of the location l given as argument
     * @param atom2 has to be the atom corresponding to the residue at the end of the location l given as argument
     * @param interactionType
     * @param m
     * @param l
     * @return
     * @throws fr.unistra.ibmc.paradise.ArchitectureException
     */
    public AtomAtomInteraction createAtomAtomInteraction(BaseBaseInteraction bbi, String atom1, String atom2, char interactionType, Molecule m, Location l) throws fr.unistra.ibmc.paradise.ArchitectureException {
        AtomAtomInteraction aai = new AtomAtomInteraction(this);
        aai.setAtom1(new Residue(l.getStart(),m),atom1);
        aai.setAtom2(new Residue(l.getEnd(),m),atom2);
        aai.setInteractionType(interactionType);
        m.addFeature(aai, l);
        bbi.addSubFeature(aai);
        return aai;
    }

    public AtomAtomInteraction copyAtomAtomInteraction(List<Molecule> targetMolecules, AtomAtomInteraction referenceInteraction, BaseBaseInteraction bbi) {
        AtomAtomInteraction copyAtomAtomInteraction;
        Iterator<Molecule> iterator = referenceInteraction.getMolecules().iterator();
        Molecule _m = iterator.next();
        if (iterator.hasNext()) {
            Molecule _m2 = iterator.next();
            copyAtomAtomInteraction = this.createAtomAtomInteraction(bbi, referenceInteraction.getAtom1(), referenceInteraction.getAtom2(), referenceInteraction.getInteractionType(), _m.searchCopy(targetMolecules), new Location(referenceInteraction.getFullLocation(_m)), _m2.searchCopy(targetMolecules), new Location(referenceInteraction.getFullLocation(_m2)));
        } else
            copyAtomAtomInteraction = this.createAtomAtomInteraction(bbi, referenceInteraction.getAtom1(), referenceInteraction.getAtom2(), referenceInteraction.getInteractionType(), _m.searchCopy(targetMolecules), new Location(referenceInteraction.getFullLocation(_m)));
        return copyAtomAtomInteraction;
    }

    /**
     *
     * @param ss
     * @param atom1 has to be the atom corresponding to the residue at the start of the location l1 given as argument
     * @param atom2 has to be the atom corresponding to the residue at the start of the location l2 given as argument
     * @param interactionType
     * @param m1
     * @param l1
     * @param m2
     * @param l2
     * @return
     * @throws fr.unistra.ibmc.paradise.ArchitectureException
     */
    public AtomAtomInteraction createAtomAtomInteraction(SecondaryStructure ss, String atom1, String atom2, char interactionType, Molecule m1, Location l1, Molecule m2, Location l2) throws fr.unistra.ibmc.paradise.ArchitectureException {
        AtomAtomInteraction aai = new AtomAtomInteraction(this);
        aai.setAtom1(new Residue(l1.getStart(),m1),atom1);
        aai.setAtom2(new Residue(l2.getStart(),m2),atom2);
        aai.setInteractionType(interactionType);
        m1.addFeature(aai, l1);
        m2.addFeature(aai, l2);
        ss.addSubFeature(aai);
        return aai;
    }


    /**
     *
     * @param bbi
     * @param atom1 has to be the atom corresponding to the residue at the start of the location l1 given as argument
     * @param atom2 has to be the atom corresponding to the residue at the start of the location l2 given as argument
     * @param interactionType
     * @param m1
     * @param l1
     * @param m2
     * @param l2
     * @return
     * @throws ArchitectureException
     */
    public AtomAtomInteraction createAtomAtomInteraction(BaseBaseInteraction bbi, String atom1, String atom2, char interactionType, Molecule m1, Location l1, Molecule m2, Location l2) throws ArchitectureException {
        AtomAtomInteraction aai = new AtomAtomInteraction(this);
        aai.setAtom1(new Residue(l1.getStart(),m1),atom1);
        aai.setAtom2(new Residue(l2.getStart(),m2),atom2);
        aai.setInteractionType(interactionType);
        m1.addFeature(aai, l1);
        m2.addFeature(aai, l2);
        bbi.addSubFeature(aai);
        return aai;
    }

    public BaseBaseInteraction copyBaseBaseInteraction(List<Molecule> targetMolecules, BaseBaseInteraction referenceBaseBaseInteraction, SecondaryStructure secondaryStructure) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        BaseBaseInteraction copyBaseBaseInteraction;
        Iterator<Molecule> iterator = referenceBaseBaseInteraction.getMolecules().iterator();
        Molecule _m = iterator.next();
        if (iterator.hasNext()) {
            Molecule _m2 = iterator.next();
            copyBaseBaseInteraction = this.createBaseBaseInteraction(referenceBaseBaseInteraction.getEdge1(), referenceBaseBaseInteraction.getEdge2(), referenceBaseBaseInteraction.getOrientation(), referenceBaseBaseInteraction.getDominant(), secondaryStructure, _m.searchCopy(targetMolecules), new Location(referenceBaseBaseInteraction.getFullLocation(_m)), _m2.searchCopy(targetMolecules), new Location(referenceBaseBaseInteraction.getFullLocation(_m2)));
        } else
            copyBaseBaseInteraction = this.createBaseBaseInteraction(referenceBaseBaseInteraction.getEdge1(), referenceBaseBaseInteraction.getEdge2(), referenceBaseBaseInteraction.getOrientation(), referenceBaseBaseInteraction.getDominant(), secondaryStructure, _m.searchCopy(targetMolecules), new Location(referenceBaseBaseInteraction.getFullLocation(_m)));
        for (AtomAtomInteraction referenceAtomAtomInteraction : referenceBaseBaseInteraction.getAtomAtomInteractions())
            this.copyAtomAtomInteraction(targetMolecules, referenceAtomAtomInteraction, copyBaseBaseInteraction);
        return copyBaseBaseInteraction;
    }

    public BaseBaseInteraction copyBaseBaseInteraction(List<Molecule> targetMolecules, BaseBaseInteraction referenceBaseBaseInteraction, Helix helix) throws BiologicalSymbolException {
        BaseBaseInteraction copyBaseBaseInteraction;
        Iterator<Molecule> iterator = referenceBaseBaseInteraction.getMolecules().iterator();
        Molecule _m = iterator.next();
        if (iterator.hasNext()) {
            Molecule _m2 = iterator.next();
            copyBaseBaseInteraction = this.createBaseBaseInteraction(referenceBaseBaseInteraction.getEdge1(), referenceBaseBaseInteraction.getEdge2(), referenceBaseBaseInteraction.getOrientation(), referenceBaseBaseInteraction.getDominant(), helix, _m.searchCopy(targetMolecules) , new Location(referenceBaseBaseInteraction.getFullLocation(_m)), _m2.searchCopy(targetMolecules) , new Location(referenceBaseBaseInteraction.getFullLocation(_m2)));
        } else
            copyBaseBaseInteraction = this.createBaseBaseInteraction(referenceBaseBaseInteraction.getEdge1(), referenceBaseBaseInteraction.getEdge2(), referenceBaseBaseInteraction.getOrientation(), referenceBaseBaseInteraction.getDominant(), helix, _m.searchCopy(targetMolecules), new Location(referenceBaseBaseInteraction.getFullLocation(_m)));
        for (AtomAtomInteraction referenceAtomAtomInteraction : referenceBaseBaseInteraction.getAtomAtomInteractions())
            this.copyAtomAtomInteraction(targetMolecules, referenceAtomAtomInteraction, copyBaseBaseInteraction);
        return copyBaseBaseInteraction;
    }

    public BaseBaseInteraction createBaseBaseInteraction(char edge5, char edge3, char orientation, int dominantResidue, Helix helix, Molecule m, Location l) throws fr.unistra.ibmc.paradise.ArchitectureException, BiologicalSymbolException {
        BaseBaseInteraction bbi;
        if (dominantResidue != BaseBaseInteraction.RESIDUE_1 && dominantResidue != BaseBaseInteraction.RESIDUE_2 && dominantResidue != BaseBaseInteraction.RESIDUE_UNDEFINED)
            throw new fr.unistra.ibmc.paradise.ArchitectureException("Invalid value of dominantResidue, you must choose between BaseBaseInteraction.RESIDUE_1, BaseBaseInteraction.RESIDUE_2 or BaseBaseInteraction.RESIDUE_UNDEFINED");
        switch (edge5) {
            case BaseBaseInteraction.EDGE_HOOGSTEEN:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHHInteraction(this, dominantResidue, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHHInteraction(this, dominantResidue, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHSInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHSInteraction(this, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHWInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHWInteraction(this, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", edge3 + "", this);
                }
                break;
            case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSHInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSHInteraction(this, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSSInteraction(this, dominantResidue, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSSInteraction(this, dominantResidue, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSWInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSWInteraction(this, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", orientation + "", this);
                }
                break;
            case BaseBaseInteraction.EDGE_WATSON_CRICK:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWHInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWHInteraction(this, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWSInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWSInteraction(this, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWWInteraction(this, dominantResidue, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWWInteraction(this, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", edge3 + "", this);
                }
                break;
            default:
                throw new BiologicalSymbolException("Edge 5 unknown : " + edge5 + " choose between H, S & W", edge5 + "", this);
        }
        this.addBaseBaseInteraction(helix, bbi);
        return bbi;
    }

    public BaseBaseInteraction createBaseBaseInteraction(char edge5, char edge3, char orientation, int dominantResidue, Helix helix, Molecule m1, Location l1, Molecule m2, Location l2) throws fr.unistra.ibmc.paradise.ArchitectureException, BiologicalSymbolException {
        BaseBaseInteraction bbi;
        if (dominantResidue != BaseBaseInteraction.RESIDUE_1 && dominantResidue != BaseBaseInteraction.RESIDUE_2 && dominantResidue != BaseBaseInteraction.RESIDUE_UNDEFINED)
            throw new fr.unistra.ibmc.paradise.ArchitectureException("Invalid value of dominantResidue, you must choose between BaseBaseInteraction.RESIDUE_1, BaseBaseInteraction.RESIDUE_2 or BaseBaseInteraction.RESIDUE_UNDEFINED");

        switch (edge5) {
            case BaseBaseInteraction.EDGE_HOOGSTEEN:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHHInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHHInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHSInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHSInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHWInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHWInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", edge3 + "", this);
                }
                break;
            case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSHInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSHInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSSInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSSInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSWInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSWInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", orientation + "", this);
                }
                break;
            case BaseBaseInteraction.EDGE_WATSON_CRICK:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWHInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWHInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWSInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWSInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWWInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWWInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", edge3 + "", this);
                }
                break;
            default:
                throw new BiologicalSymbolException("Edge 5 unknown : " + edge5 + " choose between H, S & W", edge5 + "", this);
        }
        this.addBaseBaseInteraction(helix, bbi);
        return bbi;
    }

    public BaseBaseInteraction createBaseBaseInteraction(char edge5, char edge3, char orientation, int dominantResidue, SecondaryStructure secondaryStructure, Molecule m, Location l) throws ArchitectureException, fr.unistra.ibmc.paradise.BiologicalSymbolException {
        BaseBaseInteraction bbi;
        if (dominantResidue != BaseBaseInteraction.RESIDUE_1 && dominantResidue != BaseBaseInteraction.RESIDUE_2 && dominantResidue != BaseBaseInteraction.RESIDUE_UNDEFINED)
            throw new fr.unistra.ibmc.paradise.ArchitectureException("Invalid value of dominantResidue, you must choose between BaseBaseInteraction.RESIDUE_1, BaseBaseInteraction.RESIDUE_2 or BaseBaseInteraction.RESIDUE_UNDEFINED");
        switch (edge5) {
            case BaseBaseInteraction.EDGE_HOOGSTEEN:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHHInteraction(this, dominantResidue, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHHInteraction(this, dominantResidue, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHSInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHSInteraction(this, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHWInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHWInteraction(this, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", edge3 + "", this);
                }
                break;
            case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSHInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSHInteraction(this, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSSInteraction(this, dominantResidue, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSSInteraction(this, dominantResidue, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSWInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSWInteraction(this, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", orientation + "", this);
                }
                break;
            case BaseBaseInteraction.EDGE_WATSON_CRICK:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWHInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWHInteraction(this, m, l);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWSInteraction(this, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWSInteraction(this, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWWInteraction(this, dominantResidue, m, l);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWWInteraction(this, m, l);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", edge3 + "", this);
                }
                break;
            default:
                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Edge 5 unknown : " + edge5 + " choose between H, S & W", edge5 + "", this);
        }
        this.addBaseBaseInteraction(secondaryStructure, bbi);
        return bbi;
    }

    public BaseBaseInteraction createBaseBaseInteraction(char edge5, char edge3, char orientation, int dominantResidue, SecondaryStructure secondaryStructure, Molecule m1, Location l1, Molecule m2, Location l2) throws fr.unistra.ibmc.paradise.ArchitectureException, BiologicalSymbolException {
        BaseBaseInteraction bbi;
        if (dominantResidue != BaseBaseInteraction.RESIDUE_1 && dominantResidue != BaseBaseInteraction.RESIDUE_2 && dominantResidue != BaseBaseInteraction.RESIDUE_UNDEFINED)
            throw new fr.unistra.ibmc.paradise.ArchitectureException("Invalid value of dominantResidue, you must choose between BaseBaseInteraction.RESIDUE_1, BaseBaseInteraction.RESIDUE_2 or BaseBaseInteraction.RESIDUE_UNDEFINED");

        switch (edge5) {
            case BaseBaseInteraction.EDGE_HOOGSTEEN:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHHInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHHInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHSInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHSInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisHWInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransHWInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", edge3 + "", this);
                }
                break;
            case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSHInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSHInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSSInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSSInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisSWInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransSWInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", orientation + "", this);
                }
                break;
            case BaseBaseInteraction.EDGE_WATSON_CRICK:
                switch (edge3) {
                    case BaseBaseInteraction.EDGE_HOOGSTEEN:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWHInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWHInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_SUGAR_EDGE:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWSInteraction(this, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWSInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new fr.unistra.ibmc.paradise.BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    case BaseBaseInteraction.EDGE_WATSON_CRICK:
                        switch (orientation) {
                            case BaseBaseInteraction.ORIENTATION_CIS:
                                bbi = new CisWWInteraction(this, dominantResidue, m1, l1, m2, l2);
                                break;
                            case BaseBaseInteraction.ORIENTATION_TRANS:
                                bbi = new TransWWInteraction(this, m1, l1, m2, l2);
                                break;
                            default:
                                throw new BiologicalSymbolException("Orientation unknown : " + orientation + " choose between H, S & W", orientation + "", this);
                        }
                        break;
                    default:
                        throw new BiologicalSymbolException("Edge 3 unknown : " + edge3 + " choose between H, S & W", edge3 + "", this);
                }
                break;
            default:
                throw new BiologicalSymbolException("Edge 5 unknown : " + edge5 + " choose between H, S & W", edge5 + "", this);
        }
        this.addBaseBaseInteraction(secondaryStructure, bbi);
        return bbi;
    }

    private void addBaseBaseInteraction(Helix helix, BaseBaseInteraction interaction) {
        //[fjossinet] an helix cannot contains twice the same base-base interaction
        for (BaseBaseInteraction _interaction : helix.getBaseBaseInteractions())
            if (interaction.getResidue().equals(_interaction.getResidue()) && interaction.getPartnerResidue().equals(_interaction.getPartnerResidue()) || interaction.getResidue().equals(_interaction.getPartnerResidue()) && interaction.getPartnerResidue().equals(_interaction.getResidue())) {
                helix.removeBaseBaseInteraction(_interaction);
                break;
            }
        //if an Helix contains a base-base interaction, it cannot be attached to a secondary structure anymore
        List<SecondaryStructure> sss = interaction.getParentFeatures(SecondaryStructure.class);
        for (SecondaryStructure ss : sss)
            ss.removeBaseBaseInteraction(interaction);
        helix.addSubFeature(interaction);
    }

    private void addBaseBaseInteraction(SecondaryStructure ss, BaseBaseInteraction interaction) {
        for (Helix h : ss.getHelicesWithoutPKnots())
            for (BaseBaseInteraction _interaction : h.getBaseBaseInteractions()) {
                if (interaction.getResidue().equals(_interaction.getResidue()) && interaction.getPartnerResidue().equals(_interaction.getPartnerResidue()) || interaction.getResidue().equals(_interaction.getPartnerResidue()) && interaction.getPartnerResidue().equals(_interaction.getResidue())) {
                    h.removeBaseBaseInteraction(_interaction);
                    h.addSubFeature(interaction);
                    return;
                }
            }
        //[fjossinet] a secondary-structure cannot contains twice the same base-base interaction
        for (BaseBaseInteraction _interaction : ss.getTertiaryInteractions())
            if (interaction.getResidue().equals(_interaction.getResidue()) && interaction.getPartnerResidue().equals(_interaction.getPartnerResidue()) || interaction.getResidue().equals(_interaction.getPartnerResidue()) && interaction.getPartnerResidue().equals(_interaction.getResidue())) {
                ss.removeBaseBaseInteraction(_interaction);
                break;
            }
        ss.addSubFeature(interaction);
    }

    public Residue3D createResidue3D(TertiaryStructure tertiaryStructure, Molecule m, int position) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        String residue_symbol = m.getResidueAt(position);
        boolean isProtein = Protein.class.isInstance(m);
        Residue3D r;
        if (("A".equals(residue_symbol) || ("ATP".equals(residue_symbol))) && !isProtein)
            r = new Adenine3D(this);
        else if (("U".equals(residue_symbol) || ("UTP".equals(residue_symbol))) && !isProtein)
            r = new Uridine3D(this);
        else if (("G".equals(residue_symbol) || ("GTP".equals(residue_symbol))) && !isProtein)
            r = new Guanine3D(this);
        else if (("C".equals(residue_symbol) || ("CTP".equals(residue_symbol))) && !isProtein)
            r = new Cytosine3D(this);
        else if ("C".equals(residue_symbol))
            r = new Cysteine3D(this);
        else if ("H".equals(residue_symbol))
            r = new Histidine3D(this);
        else if ("I".equals(residue_symbol))
            r = new Isoleucine3D(this);
        else if ("M".equals(residue_symbol))
            r = new Methionine3D(this);
        else if ("S".equals(residue_symbol))
            r = new Serine3D(this);
        else if ("V".equals(residue_symbol))
            r = new Valine3D(this);
        else if ("A".equals(residue_symbol))
            r = new Alanine3D(this);
        else if ("G".equals(residue_symbol))
            r = new Glycine3D(this);
        else if ("L".equals(residue_symbol))
            r = new Leucine3D(this);
        else if ("P".equals(residue_symbol))
            r = new Proline3D(this);
        else if ("T".equals(residue_symbol))
            r = new Threonine3D(this);
        else if ("F".equals(residue_symbol))
            r = new Phenylalanine3D(this);
        else if ("R".equals(residue_symbol))
            r = new Arginine3D(this);
        else if ("Y".equals(residue_symbol))
            r = new Tyrosine3D(this);
        else if ("W".equals(residue_symbol))
            r = new Tryptophane3D(this);
        else if ("D".equals(residue_symbol))
            r = new AsparticAcid3D(this);
        else if ("N".equals(residue_symbol))
            r = new Asparagine3D(this);
        else if ("E".equals(residue_symbol))
            r = new GlutamicAcid3D(this);
        else if ("Q".equals(residue_symbol))
            r = new Glutamine3D(this);
        else if ("K".equals(residue_symbol))
            r = new Lysine3D(this);
        else if (isProtein)
            r = new UnknownAminoAcid3D(this, residue_symbol);
        else {
            System.out.println("Unknown residue "+residue_symbol);
            r = new UnknownNucleotide3D(this, residue_symbol);
        }
        m.addFeature(r, new Location(position));
        tertiaryStructure.addSubFeature(r);
        return r;
    }

    public Residue3D createResidue3D(TertiaryStructure tertiaryStructure, Residue3D referenceResidue3D, Molecule m, int position) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        Residue3D r3D = referenceResidue3D.createNewResidue3D(m.getResidueAt(position).charAt(0)); //the createNewResidue3D method provided by each Residue3D class copy the atom coordinates. It is not necessary to do it here
        m.addFeature(r3D, new Location(position));
        tertiaryStructure.addSubFeature(r3D);
        return r3D;
    }


    /**
     * Create a Residue3D from the 3D coordinates of an other Residue3D
     *
     * @param tertiaryStructure  the tertiary structure to which the new Residue3D will be added
     * @param referenceResidue3D the Residue3D that will be "cloned"
     * @return
     * @throws fr.unistra.ibmc.paradise.BiologicalSymbolException
     */
    public Residue3D copyResidue3D(List<Molecule> targetMolecules, Residue3D referenceResidue3D, TertiaryStructure tertiaryStructure) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
        Molecule m = referenceResidue3D.getMolecules().iterator().next();
        Residue3D newResidue3D = this.createResidue3D(tertiaryStructure, m.searchCopy(targetMolecules), referenceResidue3D.getFullLocation(m).getStart());
        for (Residue3D.Atom srcAtom : referenceResidue3D.getAtoms())
            if (srcAtom.hasCoordinatesFilled()) {
                Residue3D.Atom dstAtom = newResidue3D.getAtom(srcAtom.getName());
                dstAtom.setCoordinates(srcAtom.getX(), srcAtom.getY(), srcAtom.getZ());
            }
        return newResidue3D;
    }

}
