package fr.unistra.ibmc.paradise.core.features;

import fr.unistra.ibmc.paradise.core.Location;
import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.ParadiseID;
import fr.unistra.ibmc.paradise.core.Source;
import fr.unistra.ibmc.paradise.ArchitectureException;

import java.io.Serializable;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class TertiaryStructure extends AbstractMainParadiseFeature {

    private int model = 1; //important for the RMN data
    private NumberingSystem numberingSystem;

    protected TertiaryStructure(ParadiseID paradiseID, StructuralFeatureFactory factory, String name, Source source) throws ArchitectureException {
        super(paradiseID, factory, name, source);
        this.numberingSystem = new NumberingSystem();
    }

    public int getModel() {
        return model;
    }

    public void setModel(int model) {
        this.model = model;
    }

    public StructuralFeatureFactory getFactory() {
        return (StructuralFeatureFactory) super.getFactory();
    }

    public void linkSecondaryStructure(SecondaryStructure ss) {
        super.addSubFeature(ss);
    }

    public List<SecondaryStructure> getLinkedSecondaryStructures() {
        return this.getSubFeatures(SecondaryStructure.class);
    }

    /**
     * Return a Residue3D (if any) at the position given as argument.
     * Since a TertiaryStructure can annotate several Molecule, you have to precise the molecule as a second argument
     *
     * @param position from 1 to molecule's length
     * @param m        the molecule for which you want the Residue3D
     * @return null if no Residue3D stored at the position given as argument
     */
    public Residue3D getResidue3DAt(int position, Molecule m) {
        List<Residue3D> hits = this.getSubFeatures(Residue3D.class, m, new Location(position));
        if (hits.size() == 1)
            return hits.get(0);
        else
            return null;
    }

    /**
     * Retrun the first residue of the molecule m which has the label given as argument.
     * @param label
     * @param m
     * @return null if no Residue3D found with this label for the molecule m
     */
    public Residue3D getResidue3DWithLabel(String label, Molecule m) {
        Map<Integer,String> labels = this.numberingSystem.residueLabels.get(m);
        for (Map.Entry<Integer,String> e:labels.entrySet()) {
            if (label.equals(e.getValue()))
                return this.getResidue3DAt(e.getKey(),m);
        }
        return null;
    }

    public void removeResidue3D(Residue3D r) {
        for (Molecule m : r.getMolecules())
            m.removeFeature(r);
        super.removeSubFeature(r);
    }

    public List<Residue3D> getResidues3D() {
        return this.getSubFeatures(Residue3D.class);
    }

    public void addResidueLabel(String label, int absolutePosition, Molecule molecule) {
        Map<Integer, String> labels = this.numberingSystem.residueLabels.get(molecule);
        if (labels == null) {
            labels = new HashMap<Integer, String>();
            this.numberingSystem.residueLabels.put(molecule, labels);
        }
        labels.put(absolutePosition, label);
    }

    public void addMoleculeLabel(String label, Molecule molecule) {
        this.numberingSystem.moleculeLabels.put(molecule,label);
    }

    /**
     *
     * @param absolutePosition
     * @param molecule
     * @return the absolute position as label if no label associated to the absolute position given as argument in the numbering system attached to this tertiary structure
     */
    public String getResidueLabel(int absolutePosition, Molecule molecule) {
        Map<Integer, String> labels = this.numberingSystem.residueLabels.get(molecule);
        if (labels == null)
            return ""+absolutePosition;
        return labels.get(absolutePosition);
    }

    /**
     *
     * @param molecule
     * @return null if no label associated with the molecule given as argument in the numbering system attached to this tertiary structure
     */
    public String getMoleculeLabel(Molecule molecule) {
        return this.numberingSystem.moleculeLabels.get(molecule);
    }

    private class NumberingSystem implements Serializable {
        private Map<Molecule, Map<Integer, String>> residueLabels;
        private Map<Molecule, String> moleculeLabels;

        private NumberingSystem() {
            this.residueLabels = new HashMap<Molecule, Map<Integer, String>>();
            this.moleculeLabels = new HashMap<Molecule, String>();
        }
    }


}
