package fr.unistra.ibmc.paradise.core.io;

import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;

import javax.swing.*;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.util.Comparator;


public class BPSeqFileIO extends AbstractParadiseFileIO {

    public static final String[] suffixes= new String[]{"bpseq"};
    
    protected List<Molecule> molecules;
    protected List<SecondaryStructure> secondaryStructures;
    protected StructuralFeatureFactory structuralFeatureFactory;

    public BPSeqFileIO(MoleculeFactory moleculeFactory) {
        super("BPSeq File Parser", moleculeFactory);
        structuralFeatureFactory = this.createStructuralFeatureFactory();
    }

    public javax.swing.filechooser.FileFilter getFileFilter() {
        return new javax.swing.filechooser.FileFilter() {
            public boolean accept(File file) {
                if (file.isDirectory())
                    return true;
                for (String suffix:BPSeqFileIO.suffixes)
                    if (file.getName().endsWith(suffix))
                        return true;
                return false;
            }

            public String getDescription() {
                return "BPSeq files";
            }
        };
    }

    public List<SecondaryStructure> getSecondaryStructures() {
        return new ArrayList<SecondaryStructure>(secondaryStructures);
    }

    public List<StructuralAlignment> getStructuralAlignments() {
        return new ArrayList<StructuralAlignment>();
    }

    public List<TertiaryStructure> getTertiaryStructures() {
        return new ArrayList<TertiaryStructure>();
    }

    public List<Molecule> getMolecules() {
        return this.molecules;
    }

    public void parseFile(File f, fr.unistra.ibmc.paradise.tools.component.ProgressMonitor monitor) throws FileParsingException, InterruptedException {
        try {
            BufferedReader in = new BufferedReader(new FileReader(f));
            this.molecules = new ArrayList<Molecule>();
            this.secondaryStructures = new ArrayList<SecondaryStructure>();
            Molecule molecule = this.moleculeFactory.createRNA(new ParadiseID(), "RNA", Source.getFileAsSource(f) , "");
            this.molecules.add(molecule);
            List<Location> helices = new ArrayList<Location>();
            List<Location> tertiaryInteractions = new ArrayList<Location>();
            List<Location> allpairs = new ArrayList<Location>();
            String line = null;
            LINES: while ((line = in.readLine()) != null) {
                if (Thread.interrupted())
                    throw new InterruptedException();
                String[] tokens = line.split(" ");
                int base5, base3;
                try {
                    base5 = Integer.parseInt(tokens[0]);
                    base3 = Integer.parseInt(tokens[2]);
                    if (base3 != 0) {
                        Location l =new Location(new Location(base5), new Location(base3));
                        if (!allpairs.contains(l))
                            allpairs.add(l);
                    }
                }
                catch (NumberFormatException e) {
                    continue LINES;
                }
                if (tokens.length == 3) {
                    try {
                        molecule.addResidue(tokens[1]);
                    }
                    catch (BiologicalSymbolException e) {
                        String choice = null;
                        while (choice == null || choice.length() == 0) {
                            choice = (String) JOptionPane.showInputDialog(null,"Please choose an unmodified residue","Unknown Residue "+line.substring(17, 21).trim(), JOptionPane.WARNING_MESSAGE,null,new String[]{"A","U","G","C","?"},"A");
                            molecule.addModifiedResidue(line.substring(17, 21).trim(),choice);
                        }
                        molecule.addResidue(choice);
                    }

                }
            }
            in.close();            
            this.secondaryStructures.add(structuralFeatureFactory.createRNASecondaryStructure(new ParadiseID(), "Secondary Structure" , Source.getFileAsSource(f), molecule, new Location(1, molecule.getLength())));
            Collections.sort(allpairs, new Comparator() {
                public int compare(Object o, Object o1) {
                    return ((Location)o).getStart()-((Location)o1).getStart();
                }
            });
            Location allSingleStrandsLocation = new Location(1, molecule.getLength());
            Location l = null;
            for (int i = 0; i < allpairs.size()-1 ; i++) {
                Location l1 = allpairs.get(i), l2 = allpairs.get(i+1);
                if (l1.getStart()+1 == l2.getStart() && l1.getEnd()-1 == l2.getEnd()) { //if the basepairs are contiguous
                    if (l == null)
                        l = new Location(l1, l2);
                    else {
                        l.add(l1);
                        l.add(l2);
                    }
                }
                else {
                    if (l != null) {
                        if (l.getLength() == 2)
                            tertiaryInteractions.add(l);
                        else {
                            helices.add(l);
                            allSingleStrandsLocation.remove(l);
                        }
                    }
                    l = null;
                }
            }
            //last helix
            if (l != null) {
                if (l.getLength() == 2)
                    tertiaryInteractions.add(l);
                else {
                    helices.add(l);
                    allSingleStrandsLocation.remove(l);
                }
            }
            int i=1;
            for (Location _l:helices) {
                structuralFeatureFactory.createHelix("H" + (i ++), this.secondaryStructures.get(this.secondaryStructures.size()-1), molecule, _l);
                //System.out.println("create helix "+_l);
            }
            for (Location _l:tertiaryInteractions)
                structuralFeatureFactory.createBaseBaseInteraction(BaseBaseInteraction.EDGE_WATSON_CRICK, BaseBaseInteraction.EDGE_WATSON_CRICK, BaseBaseInteraction.ORIENTATION_CIS, BaseBaseInteraction.RESIDUE_1, this.secondaryStructures.get(this.secondaryStructures.size()-1), molecule, _l);
            int[] boundaries = allSingleStrandsLocation.getBoundaries();
            for (i = 0; i < boundaries.length; i += 2) {
                structuralFeatureFactory.createSingleStrand("SS" + (i + 1), this.secondaryStructures.get(this.secondaryStructures.size()-1), molecule, new Location(boundaries[i], boundaries[i + 1]));
            }
        }
        catch (Exception e) {
            throw new FileParsingException(e);
        }
    }

}
