package fr.unistra.ibmc.paradise.core.io;

import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.core.utils.Residue;
import fr.unistra.ibmc.paradise.tools.component.ProgressMonitor;
import fr.unistra.ibmc.paradise.ArchitectureException;

import java.io.*;
import java.util.*;

public class CTFileIO extends AbstractParadiseFileIO {

    public static final String[] suffixes= new String[]{"ct"};

    protected StructuralFeatureFactory structuralFeatureFactory;
    protected List<Molecule> molecules;
    protected List<SecondaryStructure> secondaryStructures;

    protected CTFileIO(String description, MoleculeFactory moleculeFactory) {
        super(description, moleculeFactory);
    }

    public CTFileIO(MoleculeFactory moleculeFactory) {
        super("CT File Parser", moleculeFactory);
        structuralFeatureFactory = this.createStructuralFeatureFactory();
    }

    public List<SecondaryStructure> getSecondaryStructures() {
        return new ArrayList<SecondaryStructure>(secondaryStructures);
    }

    public List<StructuralAlignment> getStructuralAlignments() {
        return new ArrayList<StructuralAlignment>();
    }

    public List<TertiaryStructure> getTertiaryStructures() {
        return new ArrayList<TertiaryStructure>();
    }

    public javax.swing.filechooser.FileFilter getFileFilter() {
        return new javax.swing.filechooser.FileFilter() {
            public boolean accept(File file) {
                if (file.isDirectory())
                    return true;
                for (String suffix:CTFileIO.suffixes)
                    if (file.getName().endsWith(suffix))
                        return true;
                return false;
            }

            public String getDescription() {
                return "CT files";
            }
        };
    }

    public List<Molecule> getMolecules() {
        return this.molecules;
    }


    public static void writeCTFile(File f,String sequence, String bracketNotation) throws FileNotFoundException {
        if (sequence.length() == bracketNotation.length()) {
            List<Location> allpairs = new ArrayList<Location>();
            Location locationInTheMolecule = new Location(), locationInTheAlignedSequence = new Location(); //locations to store the positions of the left brackets
            int moleculePosition = 0;
            for (int i = 0; i < bracketNotation.length() ; i++) {
                moleculePosition++;
                char c = bracketNotation.charAt(i);
                switch (c) {
                    case '(' :
                        locationInTheMolecule.add(moleculePosition) ;
                        locationInTheAlignedSequence.add(i);
                        break;
                    case ')' :
                        allpairs.add(new Location(new Location(locationInTheMolecule.getEnd()), new Location(moleculePosition)));
                        locationInTheMolecule.remove(locationInTheMolecule.getEnd()) ;
                        locationInTheAlignedSequence.remove(locationInTheAlignedSequence.getEnd());
                        break;
                }
            }
            Collections.sort(allpairs, new Comparator() {
                public int compare(Object o, Object o1) {
                    return ((Location)o).getStart()-((Location)o1).getStart();
                }
            });
            Location allSingleStrandsLocation = new Location(1, sequence.length());
            for (int i = 0; i < allpairs.size() ; i++)
                allSingleStrandsLocation.remove(allpairs.get(i));
            PrintWriter pw = new PrintWriter(f);
            pw.println(sequence.length() + " (generated with PARADISE)");
            for (int i = 0; i < sequence.length() ; i++) {
                if (allSingleStrandsLocation.hasPosition(i+1))
                    pw.println((i+1) + " " + sequence.charAt(i) + " " + i + " " + (i + 2) + " " + 0 + " " + (i + 1));
                else {
                    for (Location l: allpairs) {
                        if (l.getStart() == i+1) {
                            pw.println(l.getStart() + " " + sequence.charAt(l.getStart()-1) + " " + (l.getStart()-1) + " " + (l.getStart() + 1) + " " + l.getEnd() + " " + l.getStart());
                            break;
                        }
                        else if (l.getEnd() == i+1) {
                            pw.println(l.getEnd() + " " + sequence.charAt(l.getEnd()-1) + " " + (l.getEnd()-1) + " " + (l.getEnd() + 1) + " " + l.getStart() + " " + l.getEnd());
                            break;
                        }
                    }
                }
            }
            pw.close();
        }
        else
            throw new ArchitectureException("The RNA sequence is of different length than the bracket notation!!");
    }

    public static void writeCTFile(List<Residue> residues, List<BaseBaseInteraction> interactions, File f) throws FileNotFoundException, InterruptedException {
        PrintWriter pw = new java.io.PrintWriter(new java.io.FileOutputStream(f));
        pw.println(residues.size() + " (generated with PARADISE)");
        for (int i = 0; i < residues.size(); i++) {
            Residue r = residues.get(i);
            int partnerPosition = 0;
            for (BaseBaseInteraction bbi : interactions) {
                if (r.equals(bbi.getResidue())) {
                    partnerPosition = residues.indexOf(bbi.getPartnerResidue()) + 1;
                    break;
                } else if (r.equals(bbi.getPartnerResidue())) {
                    partnerPosition = residues.indexOf(bbi.getResidue()) + 1;
                    break;
                }
            }
            pw.println((i + 1) + " " + r.getMolecule().getResidueAt(r.getAbsolutePosition()) + " " + i + " " + (i + 2) + " " + partnerPosition + " " + (i + 1));
        }
        pw.close();
    }

    public void parseFile(File f, ProgressMonitor monitor) throws FileParsingException, InterruptedException {
        try {
            BufferedReader in = new BufferedReader(new FileReader(f));
            this.molecules = new ArrayList<Molecule>();
            this.secondaryStructures = new ArrayList<SecondaryStructure>();
            List<Location> helices = new ArrayList<Location>();
            List<Location> tertiaryInteractions = new ArrayList<Location>();
            List<Location> allpairs = new ArrayList<Location>();
            String line = null, head = null;
            //the head is the line just before the line describing the first residue in the sequence
            head = in.readLine();
            if (head != null) {
                StringTokenizer tok = new StringTokenizer(head);
                int nb = Integer.parseInt(tok.nextToken());

                char[] chars = new char[nb];

                int n = 0;
                while ((line = in.readLine()) != null) {
                    if (Thread.interrupted())
                        throw new InterruptedException();
                    tok = new StringTokenizer(line);
                    if (tok.countTokens() >= 5) {
                        tok.nextToken();
                        chars[n] = tok.nextToken().charAt(0);
                        tok.nextToken();
                        tok.nextToken();
                        int partner = Integer.parseInt(tok.nextToken());
                        if (partner != 0) {
                            Location l =new Location(new Location(n+1), new Location(partner));
                            if (!allpairs.contains(l))
                                allpairs.add(l);
                        }
                        n++;
                    }
                }

                if (n != nb)
                    throw new FileParsingException("According to the header of your file, there are " + nb + " residues, but i can only found " + n);

                String sequence = new String(chars);
                Molecule molecule = this.moleculeFactory.createRNA(new ParadiseID(), "RNA", Source.getFileAsSource(f), sequence);
                this.molecules.add(molecule);
                this.secondaryStructures.add(structuralFeatureFactory.createRNASecondaryStructure(new ParadiseID(), "Secondary Structure",Source.getFileAsSource(f), molecule, new Location(1, molecule.getLength())));


                Collections.sort(allpairs, new Comparator() {
                    public int compare(Object o, Object o1) {
                        return ((Location)o).getStart()-((Location)o1).getStart();
                    }
                });
                Location allSingleStrandsLocation = new Location(1, molecule.getLength());
                Location l = null;
                for (int i = 0; i < allpairs.size()-1 ; i++) {
                    Location l1 = allpairs.get(i), l2 = allpairs.get(i+1);
                    if (l1.getStart()+1 == l2.getStart() && l1.getEnd()-1 == l2.getEnd()) { //if the basepairs are contiguous
                        if (l == null)
                            l = new Location(l1, l2);
                        else {
                            l.add(l1);
                            l.add(l2);
                        }
                    }
                    else {
                        if (l != null) {
                            if (l.getLength() == 2)
                                tertiaryInteractions.add(l);
                            else {
                                helices.add(l);
                                allSingleStrandsLocation.remove(l);
                            }
                        }
                        l = null;
                    }
                }
                //last helix
                if (l != null) {
                    if (l.getLength() == 2)
                        tertiaryInteractions.add(l);
                    else {
                        helices.add(l);
                        allSingleStrandsLocation.remove(l);
                    }
                }
                int i=1;
                for (Location _l:helices) {
                    structuralFeatureFactory.createHelix("H" + (i ++), this.secondaryStructures.get(this.secondaryStructures.size()-1), molecule, _l);
                    //System.out.println("create helix "+_l);
                }
                for (Location _l:tertiaryInteractions)
                    structuralFeatureFactory.createBaseBaseInteraction(BaseBaseInteraction.EDGE_WATSON_CRICK, BaseBaseInteraction.EDGE_WATSON_CRICK, BaseBaseInteraction.ORIENTATION_CIS, BaseBaseInteraction.RESIDUE_1, this.secondaryStructures.get(this.secondaryStructures.size()-1), molecule, _l);
                int[] boundaries = allSingleStrandsLocation.getBoundaries();
                for (i = 0; i < boundaries.length; i += 2) {
                    structuralFeatureFactory.createSingleStrand("SS" + (i + 1), this.secondaryStructures.get(this.secondaryStructures.size()-1), molecule, new Location(boundaries[i], boundaries[i + 1]));
                }
            }
            in.close();
        }
        catch (InterruptedException e) {
            throw e;
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new FileParsingException(e);
        }
    }
}