package fr.unistra.ibmc.paradise.core.io;

import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.tools.component.ProgressMonitor;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;

import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Parse a FASTA file with an RNA sequence and an RNA Secondary Structure definition (described with the bracket notation).
 * The Secondary Structure definition has to be labelled as following:
 * >Secondary Structure (this header is described in the SECONDARY_STRUCTURE_HEADER static attribute of this class
 * (((...((..
 * >whatever
 * AATTCGCTACAGTA
 * This parser doesn't construct the Secondary Structure objects. It provides a method to recover the bracket notation for each sequence stored in the fasta file.
 * It these sequences has gzps, the bracket notation is modified according to the gaps.
 * At now this parser has a problem if the brackets describing the secondary structure contain a line starting with the '>' character (describing a non-canonical base-pair)
 * The molecular sequences can also contains gaps using the '-' character
 */
public class FastaFileIO extends CTFileIO {

    public static final String[] suffixes= new String[]{"fasta","fna","fa","fas","txt"};
    
    public static String SECONDARY_STRUCTURE_HEADER = ">Secondary Structure";
    Map<Molecule, String> moleculesToBrackets ;
    private List<StructuralAlignment> structuralAlignments;
    private IdentityFeatureFactory identityFeatureFactory;


    public FastaFileIO(MoleculeFactory moleculeFactory) {
        super("FASTA File parser", moleculeFactory);
    }

    @Override
    public List<StructuralAlignment> getStructuralAlignments() {
        return this.structuralAlignments;
    }

    public javax.swing.filechooser.FileFilter getFileFilter() {
        return new javax.swing.filechooser.FileFilter() {
            public boolean accept(File file) {
                if (file.isDirectory())
                    return true;
                for (String suffix:FastaFileIO.suffixes)
                    if (file.getName().endsWith(suffix))
                        return true;
                return false;
            }

            public String getDescription() {
                return "Fasta files";
            }
        };
    }

    public void parseFile(File f, ProgressMonitor monitor) throws FileParsingException {
        this.structuralFeatureFactory = this.createStructuralFeatureFactory();
        this.identityFeatureFactory = this.createIdentityFeatureFactory();
        this.moleculesToBrackets = new HashMap<Molecule, String>();
        this.molecules = new ArrayList<Molecule>();
        this.secondaryStructures = new ArrayList<SecondaryStructure>();
        this.structuralAlignments = new ArrayList<StructuralAlignment>();
        StructuralAlignment structuralAlignment = null;
        StructuralIdentity si = null;
        StringBuffer seq = new StringBuffer(), brackets = new StringBuffer();
        Location gapsPositions = null;
        String name = null;
        try {
            BufferedReader in = new BufferedReader(new FileReader(f));
            String line;
            while ((line = in.readLine()) != null) {
                if (line.startsWith(SECONDARY_STRUCTURE_HEADER)) {
                    while ((line = in.readLine()) != null && !line.startsWith(">"))
                        brackets.append(line.replace('<', '(').replace('>', ')'));
                }
                if (line.startsWith(">") && !line.startsWith(SECONDARY_STRUCTURE_HEADER)) {
                    if (seq.length() != 0 && name != null) {
                        String sequence = seq.toString();
                        char[] characters = sequence.toCharArray();
                        for (int i=0; i< characters.length;i++)
                            if (characters[i] == '-') {
                                gapsPositions.add(i+1);
                                if (structuralAlignment == null)
                                    structuralAlignment = this.identityFeatureFactory.createStructuralAligment(new ParadiseID(), "alignment", Source.getFileAsSource(f));
                            }
                        Molecule molecule = this.moleculeFactory.createRNA(new ParadiseID(), name, Source.getFileAsSource(f), sequence.replaceAll("-",""));
                        molecules.add(molecule);
                        if (structuralAlignment != null) {
                            if (si == null)
                                si = this.identityFeatureFactory.createStructuralIdentity(structuralAlignment,"si",molecule,new Location(1,molecule.getLength()));
                            else
                                molecule.addFeature(si,new Location(1,molecule.getLength()));
                            si.addGapsLocation(molecule,gapsPositions);
                            if (brackets.length() != 0) {
                                StringBuffer _brackets = new StringBuffer(brackets.toString());
                                int[] gaps = gapsPositions.getSinglePositions();
                                for (int i = gaps.length-1; i >= 0 ; i--)
                                    _brackets.deleteCharAt(gaps[i]-1);
                                moleculesToBrackets.put(molecule,_brackets.toString());
                            }
                        }
                        else if (brackets.length() != 0)
                            moleculesToBrackets.put(molecule,brackets.toString());
                    }
                    name = line.substring(1);
                    seq = new StringBuffer();
                    gapsPositions = new Location();
                }
                else
                    seq.append(line);
            }
            in.close();
            //the last sequence stored in the file
            if (seq.length() != 0 && name != null) {
                String sequence = seq.toString();
                char[] characters = sequence.toCharArray();
                for (int i=0; i< characters.length;i++)
                    if (characters[i] == '-') {
                        gapsPositions.add(i+1);
                        if (structuralAlignment == null)
                            structuralAlignment = this.identityFeatureFactory.createStructuralAligment(new ParadiseID(), "alignment", Source.getFileAsSource(f));
                    }

                Molecule molecule = this.moleculeFactory.createRNA(new ParadiseID(), name, Source.getFileAsSource(f), seq.toString().replaceAll("-",""));
                molecules.add(molecule);
                if (structuralAlignment != null) {
                    if (si == null)
                        si = this.identityFeatureFactory.createStructuralIdentity(structuralAlignment,"si",molecule,new Location(1,molecule.getLength()));
                    else
                        molecule.addFeature(si,new Location(1,molecule.getLength()));
                    si.addGapsLocation(molecule,gapsPositions);
                    if (brackets.length() != 0) {
                        StringBuffer _brackets = new StringBuffer(brackets.toString());
                        int[] gaps = gapsPositions.getSinglePositions();
                        for (int i = gaps.length-1; i >= 0 ; i--)
                            _brackets.deleteCharAt(gaps[i]-1);
                        moleculesToBrackets.put(molecule,_brackets.toString());
                    }
                }
                else if (brackets.length() != 0)
                    moleculesToBrackets.put(molecule,brackets.toString());
            }
            if (structuralAlignment != null) {
                this.structuralAlignments.add(structuralAlignment);
                this.secondaryStructures.get(0).linkStructuralAlignment(structuralAlignment);
            }
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new FileParsingException(e);
        }
    }

    /**
     * Return null if no bracket notation for the molecule given as argument
     * @param m
     * @return
     */
    public String getBracketNotation(Molecule m) {
        return this.moleculesToBrackets.get(m);
    }

    /**
     * Export a set of sequences using the FASTA format.
     * @param f the FASTA file to create
     * @param names the names of the sequences to export
     * @param sequences the sequences to export
     * @throws FileNotFoundException
     */
    public static void exportMoleculesAsFastaFile(File f, List<String> names, List<String> sequences) throws FileNotFoundException {
        PrintWriter pw = new PrintWriter(f);
        for (String sequence:sequences) {
            pw.println(">"+names.get(sequences.indexOf(sequence)));
            for(int c = 0 ; c < sequence.length() ; c+=80 ) {
                int d = Math.min(sequence.length(),c+80 );
                pw.print(sequence.substring(c,d)+"\n");
            }
        }
        pw.close();
    }

}
