package fr.unistra.ibmc.paradise.core.io;

import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;

import javax.swing.*;
import java.io.*;
import java.text.NumberFormat;
import java.util.*;

/**
 * A class mapping the data found in a PDB file with the PARADISE featuresCreated
 * User: fjossinet
 * Date: 5 janv. 2007
 */
public class PDBFileIO extends AbstractParadiseFileIO {

    public static final String[] suffixes= new String[]{"pdb"};

    public static boolean parseProteins;
    private static int LEFT_ALIGN = 0, RIGHT_ALIGN = 1;
    protected StructuralFeatureFactory tertiaryFactory;
    protected List<TertiaryStructure> tertiaryStructures;
    private Map<String, Molecule> molecules;

    public PDBFileIO(MoleculeFactory moleculeFactory) {
        super("PDB File parser", moleculeFactory);
    }

    public List<Molecule> getMolecules() {
        return new ArrayList<Molecule>(this.molecules.values());
    }

    /**
     * Serialize a set of Residue3D Feature using the PDB File Format. All the Residues3D have to be subfeatures of the same TertiaryStructure.
     * Only the first TertiaryStructure will be taken into account, other will be ignored
     *
     * @param residues
     * @param f
     * @param exportNumberingSystem if setted to false, the residues will be numbered according to their absolute positions
     * @return a Map between the ParadiseLSID of all the Molecule objects serialized and their id used in the PDB ouput file
     * @throws FileNotFoundException
     */
    public static Map<ParadiseID, String> writePDBFile(List<Residue3D> residues, boolean exportNumberingSystem, File f) throws FileNotFoundException, InterruptedException {
        Map<ParadiseID, String> mapping = new HashMap<ParadiseID, String>();
        PrintWriter pw = new java.io.PrintWriter(new java.io.FileOutputStream(f));
        int atomID = 0;
        NumberFormat coordFormat = NumberFormat.getInstance(Locale.ENGLISH);
        coordFormat.setMinimumFractionDigits(3);
        coordFormat.setMaximumFractionDigits(3);
        Location residueLocation = null;
        //we order the residues
        //[fjossinet] bug corrected, replace the previous version : order(structure.getResidues3D(), structure)
        residues = order(residues);
        char lastMoleculeName = 'A';
        TertiaryStructure ts = null;
        for (Residue3D residue : residues) {
            if (Thread.interrupted())
                throw new InterruptedException();
            Molecule m = residue.getMolecules().iterator().next();
            residueLocation = residue.getFullLocation(m);
            String residueName = m.getResidueAt(residueLocation.getStart());
            if (ts == null)
                ts = residue.getParentFeatures(TertiaryStructure.class).get(0);
            String moleculeLabel = ts.getMoleculeLabel(m);
            if (moleculeLabel == null) {
                moleculeLabel = ""+(++lastMoleculeName);
                ts.addMoleculeLabel(moleculeLabel,m);
            }
            mapping.put(residue.getMolecules().iterator().next().getParadiseID(), "" + m.getName().charAt(0));
            for (Residue3D.Atom a : residue.getAtoms()) {
                //not all the atoms registered during the construction of a Residue3D object have their coordinates filled
                if (a.hasCoordinatesFilled()) {
                    //the tag is left justified blank filled with 6 columns
                    pw.print(getPDBField(6, "ATOM", LEFT_ALIGN));
                    pw.print(getPDBField(11 - 7 + 1, "" + (++atomID), RIGHT_ALIGN));
                    pw.print("  ");
                    pw.print(getPDBField(16 - 13 + 1, a.getName().replace('\'', '*'), LEFT_ALIGN));
                    pw.print(getPDBField(20 - 18 + 1, residueName, RIGHT_ALIGN));
                    pw.print(getPDBField(1, " " + moleculeLabel, LEFT_ALIGN));
                    if (exportNumberingSystem)
                        pw.print(getPDBField(26 - 23 + 1, ts.getResidueLabel(residueLocation.getStart(),m), RIGHT_ALIGN));
                    else
                        pw.print(getPDBField(26 - 23 + 1, "" + residueLocation.getStart(), RIGHT_ALIGN));
                    pw.print(getPDBField(1, "", LEFT_ALIGN));
                    pw.print("   ");
                    pw.print(getPDBField(38 - 31 + 1, "" + coordFormat.format(a.getX()), RIGHT_ALIGN));
                    pw.print(getPDBField(46 - 39 + 1, "" + coordFormat.format(a.getY()), RIGHT_ALIGN));
                    pw.print(getPDBField(54 - 47 + 1, "" + coordFormat.format(a.getZ()), RIGHT_ALIGN));
                    //to fit to 80 columns for a PDB line
                    pw.println(getPDBField(26, "", LEFT_ALIGN));
                }
            }
            if (!hasNext(residue, residues) || changeMolecule(residue, getNext(residue, residues))) {
                pw.print(getPDBField(6, "TER", LEFT_ALIGN));
                pw.print(getPDBField(11 - 7 + 1, "" + (++atomID), RIGHT_ALIGN));
                pw.print("  ");
                pw.print(getPDBField(16 - 13 + 1, "", LEFT_ALIGN));
                pw.print(getPDBField(20 - 18 + 1, residueName, RIGHT_ALIGN));
                pw.print(getPDBField(1, " " + moleculeLabel, LEFT_ALIGN));
                if (exportNumberingSystem)
                    pw.print(getPDBField(26 - 23 + 1, ts.getResidueLabel(residueLocation.getStart(),m), RIGHT_ALIGN));
                else
                    pw.print(getPDBField(26 - 23 + 1, "" + residueLocation.getStart(), RIGHT_ALIGN));
                pw.print(getPDBField(1, "", LEFT_ALIGN));
                pw.print("   ");
                //to fit to 80 columns for a PDB line
                pw.print(getPDBField(38 - 31 + 1, "", RIGHT_ALIGN));
                pw.print(getPDBField(46 - 39 + 1, "", RIGHT_ALIGN));
                pw.print(getPDBField(54 - 47 + 1, "", RIGHT_ALIGN));
                pw.println(getPDBField(26, "", LEFT_ALIGN));
            }
        }
        pw.println("END   ");
        pw.close();

        return mapping;
    }

    private static List<Residue3D> order(List<Residue3D> residues) {
        List<Residue3D> ret = new ArrayList<Residue3D>();
        HashMap<Molecule, List<Residue3D>> lists = new HashMap<Molecule, List<Residue3D>>();
        //add each residue to a list associated with its molecule
        for (Residue3D res : residues) {
            Molecule m = res.getMolecules().iterator().next();
            if (!lists.containsKey(m))
                lists.put(m, new ArrayList<Residue3D>());
            orderAdd(res, lists.get(m));
        }
        //get the residues in order to return them
        for (List<Residue3D> l : lists.values())
            for (Residue3D res : l)
                ret.add(res);
        return ret;
    }

    private static void orderAdd(Residue3D res, List<Residue3D> list) {
        for (Residue3D lr : list)
            if (res.getAnnotations().get(0).getLocation().getStart() < lr.getAnnotations().get(0).getLocation().getStart()) {
                list.add(list.indexOf(lr), res);
                return;
            }
        list.add(res);
    }

    private static boolean changeMolecule(Residue3D r1, Residue3D r2) {
        return r1.getAnnotations().get(0).getMolecule() != r2.getAnnotations().get(0).getMolecule();
    }

    private static Residue3D getNext(Residue3D residue, List<Residue3D> residues) {
        int index = residues.indexOf(residue);
        return residues.get(index + 1);
    }

    private static boolean hasNext(Residue3D residue, List<Residue3D> residues) {
        int index = residues.indexOf(residue);
        return (index != -1 && index != residues.size() - 1);
    }

    /**
     * Return the TertiaryStructure feature constructed by the parser
     *
     * @return
     * @deprecated
     */
    //TODO remove the usage of this method. Now, due to the RMN data, a PDB parser return a list of several TertiaryStructures (only one if X-ray data)
    public TertiaryStructure getTertiaryStructure() {
        return this.tertiaryStructures.get(0);
    }

    public List<TertiaryStructure> getTertiaryStructures() {
        return new ArrayList<TertiaryStructure>(this.tertiaryStructures);
    }

    public List<SecondaryStructure> getSecondaryStructures() {
        return new ArrayList<SecondaryStructure>();
    }

    public List<StructuralAlignment> getStructuralAlignments() {
        return new ArrayList<StructuralAlignment>();
    }

    public void parseFile(File f, fr.unistra.ibmc.paradise.tools.component.ProgressMonitor monitor) throws FileParsingException, InterruptedException {
        this.molecules = new HashMap<String, Molecule>();
        this.tertiaryFactory = this.createStructuralFeatureFactory();
        //[fjossinet] since the ParadiseFileIO are instanciated only once when the Paradise framework is started, we have to reinitialize the list of tertiary structures
        if (this.tertiaryStructures == null)
            this.tertiaryStructures = new ArrayList<TertiaryStructure>();
        else
            this.tertiaryStructures.clear();
        TertiaryStructure currentTertiaryStructure = null;
        Molecule molecule = null;
        BufferedReader input = null;
        int currentModel = 1;
        try {
            input = new BufferedReader(new FileReader(f));
            String line = null;
            String tag = "fake";
            int nucleic_chain_id = 1, protein_chain_id = 1;
            String molecule_label = "fake";
            int nt_id = 0, aa_id = 0;
            String resId = "fake";
            Residue3D r = null;
            ModifiedResidue modified;
            //necessary to store the atoms parameters until the parser knowns if it parses a nucleotide or an aminoacid
            Map<String, float[]> atoms = new HashMap<String, float[]>();
            boolean isInsideNucleicAcid = false, isInsideProtein = false, ter_tag = false;
            LINES:
            while ((line = input.readLine()) != null) {
                if (Thread.interrupted())
                    throw new InterruptedException();
                try {
                    tag = line.substring(0, 6).trim();
                }
                catch (IndexOutOfBoundsException e) {
                    tag = "fake";
                }
                if ((tag.equalsIgnoreCase("ATOM") || tag.equalsIgnoreCase("HETATM")) && !line.substring(17, 21).trim().equals("HOH") /*sometimes the HOH HETATM are associated to a chain (example PDBID: 1agl)*/ && line.substring(21, 22).trim().length() != 0 /*only if the ATOM or HETATM line precises a molecule name*/) {
                    //new molecule so new residue
                    if (!molecule_label.equalsIgnoreCase(line.substring(21, 22).trim()) || ter_tag) {
                        //name.length==0 for H2O, Magnesium ions, .... For residues not inside a macromolecule
                        if ((isInsideNucleicAcid || isInsideProtein) && molecule_label.length() > 0) {
                            if (isInsideNucleicAcid) {
                                nucleic_chain_id++;
                                isInsideNucleicAcid = false;
                            } else {
                                protein_chain_id++;
                                isInsideProtein = false;
                            }
                        }
                        molecule_label = line.substring(21, 22).trim();
                        nt_id = 0;
                        aa_id = 0;
                        resId = line.substring(22, 27).trim();
                        r = null;
                        if (ter_tag)
                            ter_tag = false;
                    }
                    //only new residue
                    else if (!resId.equalsIgnoreCase(line.substring(22, 27).trim())) {
                        resId = line.substring(22, 27).trim();
                        r = null;
                    }
                    //residue is a base if the 04' atom is detected
                    if ((line.substring(12, 16).trim().equals("O4*") || line.substring(12, 16).trim().equals("O4'"))) {
                        nt_id++;
                        if (!isInsideNucleicAcid) {
                            isInsideNucleicAcid = true;
                            if (currentModel > 1)
                                molecule = this.molecules.get(molecule_label);
                            else {
                                molecule = this.moleculeFactory.createRNA(new ParadiseID(), molecule_label, Source.getFileAsSource(f), "");
                                this.molecules.put(molecule_label, molecule);
                            }
                            if (currentTertiaryStructure == null) {
                                currentTertiaryStructure = this.tertiaryFactory.createTertiaryStructure(new ParadiseID(), "Tertiary Structure", Source.getFileAsSource(f), molecule, new Location());
                                currentTertiaryStructure.setModel(currentModel);
                                this.tertiaryStructures.add(currentTertiaryStructure);
                            } else
                                molecule.addFeature(currentTertiaryStructure, new Location(1, molecule.getLength()));
                            currentTertiaryStructure.addMoleculeLabel(molecule_label, molecule);
                        }

                        //add the residue to the molecule
                        if (currentModel == 1) {
                            try {
                                molecule.addResidue(line.substring(17, 21).trim());
                            }
                            catch (BiologicalSymbolException e) {
                                String choice = null;
                                while (choice == null || choice.length() == 0) {
                                    choice = (String)JOptionPane.showInputDialog(null,"Please choose an unmodified residue","Unknown Residue "+line.substring(17, 21).trim(), JOptionPane.WARNING_MESSAGE,null,new String[]{"A","U","G","C", "?"},"A");
                                    molecule.addModifiedResidue(line.substring(17, 21).trim(),choice);
                                }
                                molecule.addResidue(choice);
                            }
                        }
                        r = this.tertiaryFactory.createResidue3D(currentTertiaryStructure, molecule, nt_id);

                        for (Map.Entry<String, float[]> e : atoms.entrySet()) {
                            //[fjossinet] PARADISE understand only C3' not C3* so....
                            r.setAtomCoordinates(e.getKey().replace('*', '\''), e.getValue()[0], e.getValue()[1], e.getValue()[2]);
                        }

                        atoms.clear();
                        //we register the new base in the NumberingSystem
                        currentTertiaryStructure.addResidueLabel(resId, nt_id, molecule);
                    }
                    //residue is an amino-acid if the CA atom is detected
                    else if (line.substring(12, 16).trim().equals("CA") && PDBFileIO.parseProteins) {
                        aa_id++;
                        if (!isInsideProtein) {
                            isInsideProtein = true;
                            if (currentModel > 1)
                                molecule = this.molecules.get(molecule_label);
                            else {
                                molecule = this.moleculeFactory.createProtein(new ParadiseID(), molecule_label, Source.getFileAsSource(f), "");
                                this.molecules.put(molecule_label, molecule);
                            }
                            if (currentTertiaryStructure == null) {
                                currentTertiaryStructure = this.tertiaryFactory.createTertiaryStructure(new ParadiseID(), ",Tertiary Structure", Source.getFileAsSource(f), molecule, new Location());
                                currentTertiaryStructure.setModel(currentModel);
                                this.tertiaryStructures.add(currentTertiaryStructure);
                            } else
                                molecule.addFeature(currentTertiaryStructure, new Location(1, molecule.getLength()));
                            currentTertiaryStructure.addMoleculeLabel(molecule_label, molecule);
                        }

                        //add the residue to the molecule
                        if (currentModel == 1)
                            try {
                                molecule.addResidue(line.substring(17, 21).trim());
                            }
                            catch (BiologicalSymbolException e) {
                                String choice = null;
                                while (choice == null || choice.length() == 0) {
                                    choice = (String)JOptionPane.showInputDialog(null,"Please choose an unmodified residue","Unknown Residue "+line.substring(17, 21).trim(), JOptionPane.WARNING_MESSAGE,null,new String[]{"A","U","G","C","?"},"A");
                                    molecule.addModifiedResidue(line.substring(17, 21).trim(),choice);
                                }
                                molecule.addResidue(choice);
                            }
                        r = this.tertiaryFactory.createResidue3D(currentTertiaryStructure, molecule, aa_id);

                        for (Map.Entry<String, float[]> e : atoms.entrySet()) {
                            r.setAtomCoordinates(e.getKey(), e.getValue()[0], e.getValue()[1], e.getValue()[2]);
                        }

                        atoms.clear();
                        //we register the new aa in the NumberingSystem
                        currentTertiaryStructure.addResidueLabel(resId, nt_id, molecule);
                    }
                    float[] coord = new float[]{Float.parseFloat(line.substring(30, 38).trim()),
                            Float.parseFloat(line.substring(38, 46).trim()),
                            Float.parseFloat(line.substring(46, 54).trim())};
                    if (r != null) {
                        r.setAtomCoordinates(line.substring(12, 16).trim().replace('*', '\''), coord[0], coord[1], coord[2]);
                        //System.out.println(line.substring(12, 16).trim());
                    }
                    //if the parser doesn't know the type of Residue currently parsed, we store the atoms in a temporary Map
                    else
                        atoms.put(line.substring(12, 16).trim(), coord);

                } else if (tag.equalsIgnoreCase("TER")) {
                    ter_tag = true;
                } else if (tag.equalsIgnoreCase("MODEL")) { //important for RMN data. Each model will be a new tertiary structure
                    currentTertiaryStructure = null;
                    currentModel = Integer.parseInt(line.split("\\s+")[1].trim());
                }
            }
        } catch (InterruptedException e) {
            throw e;
        } catch (Exception e) {
            throw new FileParsingException(e.getMessage(), e);
        }
    }

    public javax.swing.filechooser.FileFilter getFileFilter() {
        return new javax.swing.filechooser.FileFilter() {
            public boolean accept(File file) {
                if (file.isDirectory())
                    return true;
                for (String suffix:PDBFileIO.suffixes)
                    if (file.getName().endsWith(suffix))
                        return true;
                return false;
            }

            public String getDescription() {
                return "PDB files";
            }
        };
    }

    static private String getPDBField(int finalsize, String word, int align) {
        StringBuffer field = new StringBuffer();
        if (align == LEFT_ALIGN) {
            field.append(word);
            for (int i = 0; i < finalsize - word.length(); i++) {
                field.append(" ");
            }
        } else {
            for (int i = 0; i < finalsize - word.length(); i++) {
                field.append(" ");
            }
            field.append(word);
        }
        return field.toString();
    }
}
