package fr.unistra.ibmc.paradise.core.io;

import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.tools.ParadiseWorkingSession;
import fr.unistra.ibmc.paradise.tools.component.ProgressMonitor;
import fr.unistra.ibmc.paradise.utils.IOUtils;
import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.xpath.XPath;

import javax.swing.*;
import java.io.File;
import java.io.IOException;
import java.util.*;

public class RNAMLFileIO extends AbstractParadiseFileIO {

    public static final String[] suffixes= new String[]{"xml", "rnaml"};

    private List<Molecule> molecules;
    private List<RNASecondaryStructure> secondaryStructures;
    private List<StructuralAlignment> structuralAlignments;
    private List<TertiaryStructure> tertiaryStructures;
    private StructuralFeatureFactory structuralFactory;
    private IdentityFeatureFactory identityFactory;
    private Document document;

    public RNAMLFileIO(MoleculeFactory moleculeFactory) {
        super("RNAML File parser", moleculeFactory);
    }

    public void parseFile(File f, fr.unistra.ibmc.paradise.tools.component.ProgressMonitor monitor) throws FileParsingException, InterruptedException {
        this.molecules = new ArrayList<Molecule>();
        this.secondaryStructures = new ArrayList<RNASecondaryStructure>();
        this.structuralAlignments = new ArrayList<StructuralAlignment>();
        this.tertiaryStructures = new ArrayList<TertiaryStructure>();
        this.structuralFactory = this.createStructuralFeatureFactory();
        this.identityFactory = this.createIdentityFeatureFactory();
        try {
            IOUtils.writeToFile(f.getParent() + System.getProperty("file.separator") + "rnaml.dtd", RNAMLFileIO.class.getResourceAsStream("rnaml.dtd"));
            SAXBuilder builder = new SAXBuilder(false);
            this.document = builder.build(f);
            Element root = this.document.getRootElement();
            if (root.getAttributeValue("version") != null && root.getAttributeValue("version").equals("paradise 1.0")) {
                ParadiseRnaMLParser1_0 p = new ParadiseRnaMLParser1_0();
                p.parse(f, root);
                this.molecules.addAll(p.molecules.values());
                this.secondaryStructures.addAll(p.secondaryStructures.values());
                this.tertiaryStructures.addAll(p.tertiaryStructures.values());
                this.structuralAlignments.addAll(p.structuralAlignments.values());
            } else if (root.getAttributeValue("version") != null && root.getAttributeValue("version").equals("paradise 2.0")) {
                ParadiseRnaMLParser2_0 p = new ParadiseRnaMLParser2_0();
                p.parse(Source.getFileAsSource(f), root);
                this.molecules.addAll(p.molecules.values());
                this.secondaryStructures.addAll(p.secondaryStructures.values());
                this.tertiaryStructures.addAll(p.tertiaryStructures.values());
                this.structuralAlignments.addAll(p.structuralAlignments.values());
            } else {
                //to be able to read the RNAML files generated by the previous versions of S2S or thoses generated by RNAVIEW for example
                RnaMLParser p = new RnaMLParser();
                p.parse(f, root, monitor);
                this.molecules.addAll(p.molecules.values());
                this.secondaryStructures.addAll(p.secondaryStructures.values());
            }
        }
        catch (InterruptedException e) {
            throw e;
        } catch (Exception e) {
            throw new FileParsingException(e);
        }
    }

    public javax.swing.filechooser.FileFilter getFileFilter() {
        return new javax.swing.filechooser.FileFilter() {
            public boolean accept(File file) {
                if (file.isDirectory())
                    return true;
                for (String suffix:RNAMLFileIO.suffixes)
                    if (file.getName().endsWith(suffix))
                        return true;
                return false;
            }

            public String getDescription() {
                return "RNAML files";
            }
        };
    }

    /**
     * return all the RNASecondaryStructure features constructed
     *
     * @return
     */
    public List<SecondaryStructure> getSecondaryStructures() {
        return new ArrayList<SecondaryStructure>(this.secondaryStructures);
    }

    public List<StructuralAlignment> getStructuralAlignments() {
        return new ArrayList<StructuralAlignment>(structuralAlignments);
    }

    public List<TertiaryStructure> getTertiaryStructures() {
        return new ArrayList<TertiaryStructure>(tertiaryStructures);
    }

    public List<Molecule> getMolecules() {
        return new ArrayList<Molecule>(this.molecules);
    }

    private class RnaMLParser {

        private int helixCount, sstrandCount, alignmentCount,  modelCount, rnaSequenceCount, currentResidue = 1;
        private RNASecondaryStructure currentRNASecondaryStructure;
        private SecondaryStructureDisplay currentSecondaryStructureDisplay;
        private RNA currentRNA;
        private boolean interactions;
        private HashMap<Integer, ResidueIdentity> columnToResidueIdentity;
        private StructuralAlignment structuralAlignment;
        private Map<Integer, RNASecondaryStructure> secondaryStructures;
        private Map<Integer, Molecule> molecules;
        private Map<Integer, SecondaryStructureDisplay> secondaryStructureDisplays;

        private RnaMLParser() {
            this.secondaryStructures = new HashMap<Integer, RNASecondaryStructure>();
            this.molecules = new HashMap<Integer, Molecule>();
            this.secondaryStructureDisplays = new HashMap<Integer, SecondaryStructureDisplay>();
        }

        private void parse(File rnamlFile, Element root, ProgressMonitor monitor) throws fr.unistra.ibmc.paradise.BiologicalSymbolException, FileParsingException, IOException, JDOMException, InterruptedException {
            this.traverseJDOMElement(rnamlFile, root);
            //we are constructing the single-strands for each secondary structure
            Location single_strands_location;
            Location helixLocation;
            for (RNASecondaryStructure ss : secondaryStructures.values()) {
                for (Molecule m : ss.getMolecules()) {
                    single_strands_location = new Location(1, m.getLength());
                    for (Helix h : ss.getHelices()) {
                        //the location of this helix for the current molecule
                        helixLocation = h.getFullLocation(m);
                        if (helixLocation != null)
                            single_strands_location.remove(helixLocation);
                    }
                    //once all the locations of the helices attached to the current molecule removed, we have the boundaries of all the single-strands for this molecule
                    int sstrand_count = 0;
                    int[] boundaries = single_strands_location.getBoundaries();
                    for (int i = 0; i < boundaries.length; i += 2)
                        structuralFactory.createSingleStrand("SS" + (sstrand_count++), this.currentRNASecondaryStructure, m, new Location(boundaries[i], boundaries[i + 1]));
                }
            }
            // !!!!!!!!!!!! Compatibility with the previous version of S2S !!!!!!!!!!!!!!!!!!!!!
            //if there was an associated tertiary structure...
            //we check if there is an s2s comment in the root element of the rnaml file
            final org.jdom.Attribute comment = root.getAttribute("comment");
            if (comment != null) {
                String commentValue = comment.getValue().trim();
                if (commentValue.startsWith("s2s:")) {
                    PDBFileIO pdbParser = new PDBFileIO(moleculeFactory);
                    PDBFileIO.parseProteins = false;
                    pdbParser.parseFile(new File(rnamlFile.getParentFile(), commentValue.split("s2s:")[1]), monitor);
                    PDBFileIO.parseProteins = true;
                    //we add the molecules generated by the PDBFileIO to the list of the molecules generated by this parser
                    RNAMLFileIO.this.molecules.addAll(pdbParser.getMolecules());
                    TertiaryStructure tertiaryStructure = pdbParser.getTertiaryStructure();
                    Collection<Molecule> pdbMolecules = tertiaryStructure.getMolecules();
                    tertiaryStructure.getFactory().copyTertiaryStructure(Arrays.asList(this.molecules.values().toArray(new Molecule[]{})), tertiaryStructure);
                    //the tertiaryStructure is the parent feature of the secondarystructures stored in the rnaml file
                    for (RNASecondaryStructure ss : secondaryStructures.values())
                        tertiaryStructure.linkSecondaryStructure(ss);
                    //if the molecules generated by the PDB parser has been created by a ParadiseWorkingSession, we have to delete them since they are already stored in the working session with the parsing of the RNAML file
                    if (ParadiseWorkingSession.class.isInstance(moleculeFactory))
                        for (Molecule pdbMolecule : pdbMolecules)
                            ((ParadiseWorkingSession) moleculeFactory).removeMolecule(pdbMolecule);
                }

            }
        }

        private void traverseJDOMElement(File rnamlFile, final Element node) throws fr.unistra.ibmc.paradise.BiologicalSymbolException, FileParsingException, InterruptedException {
            if (Thread.interrupted())
                throw new InterruptedException();
            Element child = null;
            String name = null;

            for (Iterator i = node.getChildren().iterator(); i.hasNext();) {
                child = (Element) i.next();
                name = child.getName();
                if (name.equals("molecule")) {
                    this.rnaSequenceCount++;
                    this.modelCount = 0;
                    String rnaSequence = "";
                    Element sequence = child.getChild("sequence");
                    if (sequence != null) {
                        Element seqdata = sequence.getChild("seq-data");
                        if (seqdata != null)
                            rnaSequence = seqdata.getValue().trim().toUpperCase().replaceAll("\\s*", "");
                    }
                    this.currentRNA = moleculeFactory.createRNA(new ParadiseID(), "" + this.rnaSequenceCount, Source.getFileAsSource(rnamlFile), rnaSequence);
                    molecules.put(this.rnaSequenceCount, this.currentRNA);
                } else if (name.equals("model")) {
                    this.modelCount++;
                    //we recover a currentSecondaryStructure object with the modelCount id
                    this.currentRNASecondaryStructure = secondaryStructures.get(this.modelCount);
                    //if no currentSecondaryStructure is recovered, we create a new one
                    if (this.currentRNASecondaryStructure == null) {
                        this.currentRNASecondaryStructure = structuralFactory.createRNASecondaryStructure(new ParadiseID(), "?", Source.getFileAsSource(rnamlFile), this.currentRNA, new Location());
                        secondaryStructures.put(this.modelCount, this.currentRNASecondaryStructure);
                    }
                } else if (name.equals("str-analysis")) {
                    this.helixCount = 0;
                    this.sstrandCount = 0;
                } else if (name.equals("interactions")) {
                    this.interactions = true;
                    this.helixCount = 0;
                } else if (name.equals("base-pair")) {
                    this.defineBasePair(child);
                } else if (name.equals("helix")) {
                    this.helixCount++;
                    if (!this.interactions)
                        this.defineHelix(child);
                    else
                        this.defineInterHelix(child);
                } else if (name.equals("ss-base-coord")) {
                    this.constructBase(child);
                } else if (name.equals("coordinates")) {
                    this.constructSecondaryCoordinates(child);
                } else if (name.equals("secondary-structure-display")) {
                    //we recover a currentSecondaryStructureDisplay object with the modelCount id
                    this.currentSecondaryStructureDisplay = secondaryStructureDisplays.get(this.modelCount);
                    //if no currentSecondaryStructureDisplay is recovered, we create a new one
                    if (this.currentSecondaryStructureDisplay == null) {
                        this.currentSecondaryStructureDisplay = structuralFactory.createSecondaryStructureDisplay(this.currentRNASecondaryStructure, this.currentRNA, this.currentRNASecondaryStructure.getFullLocation(this.currentRNA));
                        secondaryStructureDisplays.put(this.modelCount, this.currentSecondaryStructureDisplay);
                    }
                } else if (child.getName().equals("alignment")) {
                    structuralAlignment = null;
                    columnToResidueIdentity = new HashMap<Integer, ResidueIdentity>();
                } else if (child.getName().equals("ali-sequence")) {
                    String sequence_name = child.getChild("molecule-id").getAttributeValue("ref").trim();
                    if (!FastaFileIO.SECONDARY_STRUCTURE_HEADER.equals(sequence_name)) {
                        String sequence = child.getChild("seq-data").getValue().trim();
                        if (structuralAlignment == null) {
                            structuralAlignment = identityFactory.createStructuralAligment(new ParadiseID(), "Alignment"+(++alignmentCount), Source.getFileAsSource(rnamlFile), this.currentRNA, new Location());
                            char[] cs = sequence.toCharArray();
                            int positionInSequence = 0;
                            for (int j = 0; j < cs.length; j++) {
                                if (cs[j] != '-') {
                                    positionInSequence++;
                                    columnToResidueIdentity.put(j, identityFactory.createResidueIdentity(structuralAlignment, this.currentRNA, new Location(positionInSequence)));
                                }
                            }
                            //We add this alignment to all the secondary structures constructed so far
                            for (SecondaryStructure ss : secondaryStructures.values())
                                ss.linkStructuralAlignment(this.structuralAlignment);
                        } else {
                            RNA rna = moleculeFactory.createRNA(new ParadiseID(), sequence_name, Source.getFileAsSource(rnamlFile), "");
                            char[] cs = sequence.toCharArray();
                            int positionInSequence = 0;
                            for (int j = 0; j < cs.length; j++) {
                                if (cs[j] != '-') {
                                    try {
                                        rna.addResidue("" + cs[j]);
                                    }
                                    catch (fr.unistra.ibmc.paradise.BiologicalSymbolException e) {
                                        String choice = null;
                                        while (choice == null || choice.length() == 0) {
                                            choice = (String) JOptionPane.showInputDialog(null,"Please choose an unmodified residue","Unknown Residue "+cs[j], JOptionPane.WARNING_MESSAGE,null,new String[]{"A","U","G","C","?"},"A");
                                            rna.addModifiedResidue(""+cs[j],choice);
                                        }
                                        rna.addResidue(choice);
                                    }
                                    positionInSequence++;
                                    ResidueIdentity ri = columnToResidueIdentity.get(j);
                                    if (ri != null)
                                        rna.addFeature(ri, new Location(positionInSequence));
                                    else
                                        columnToResidueIdentity.put(j, identityFactory.createResidueIdentity(structuralAlignment, rna, new Location(positionInSequence)));
                                }
                            }
                        }
                    }
                }

                //we avoid to treat some elements to avoid an OutOfMemory Error
                else if (child.getName().equals("base") ||
                        child.getName().equals("references") ||
                        child.getName().equals("database-entry") ||
                        child.getName().equals("analysis") ||
                        child.getName().equals("revision")) {
                    continue;
                }
                this.traverseJDOMElement(rnamlFile, child);
            }
        }

        private void constructBase(final Element b) {
            this.currentResidue = Integer.parseInt(b.getChild("base-id").getChild("position").getValue());
        }

        //TODO add only the boundaries of the Helices, not all the 2D coodinates from the RNAML file
        private void constructSecondaryCoordinates(final Element coord) throws FileParsingException {
            final String coordinates = coord.getValue();
            final java.util.StringTokenizer token = new java.util.StringTokenizer(coordinates, " ");
            final float[] baseCoord = new float[2];
            try {
                baseCoord[0] = Float.parseFloat(token.nextToken());
                baseCoord[1] = Float.parseFloat(token.nextToken());
            }
            catch (NumberFormatException nfe) {
                throw new FileParsingException(nfe);
            }
            structuralFactory.createResidue2D(this.currentSecondaryStructureDisplay, baseCoord[0], baseCoord[1], this.currentRNA, this.currentResidue);
        }


        private void defineBasePair(final Element bp) throws fr.unistra.ibmc.paradise.BiologicalSymbolException {
            final String base5id = bp.getChild("base-id-5p").getChild("base-id").getChild("position").getValue();
            final String base3id = bp.getChild("base-id-3p").getChild("base-id").getChild("position").getValue();
            final Element edge5 = bp.getChild("edge-5p");
            final Element edge3 = bp.getChild("edge-3p");
            final Element orientation = bp.getChild("bond-orientation");
            char _edge5 = '?', _edge3 = '?', _orientation = '?';

            if (orientation != null && edge5 != null && edge3 != null) {
                _orientation = orientation.getValue().trim().toUpperCase().toCharArray()[0];
                _edge5 = edge5.getValue().trim().toUpperCase().toCharArray()[0];
                _edge3 = edge3.getValue().trim().toUpperCase().toCharArray()[0];
            }
            if (_edge5 == '+' || _edge5 == '-')
                _edge5 = 'W';
            if (_edge3 == '+' || _edge3 == '-')
                _edge3 = 'W';
            String rna5id = null, rna3id = null;
            Element molecule_id = bp.getChild("base-id-5p").getChild("base-id").getChild("molecule-id");
            if (molecule_id != null) {
                rna5id = molecule_id.getAttributeValue("ref");
            }
            molecule_id = bp.getChild("base-id-3p").getChild("base-id").getChild("molecule-id");
            if (molecule_id != null) {
                rna3id = molecule_id.getAttributeValue("ref");
            }
            if (_orientation == '!' && _edge5 == '!' && _edge3 == '!') {
                AtomAtomInteraction interaction;
                /*
    if (!this.interactions)
        interaction = structuralFactory.createAtomAtomInteraction("?","?",AtomAtomInteraction.INTERACTION_HBOND,this.currentRNA,new Location(new Location(new Integer(base5id).intValue()),new Location(new Integer(base3id).intValue())));
    else
        interaction = structuralFactory.createAtomAtomInteraction("?","?",AtomAtomInteraction.INTERACTION_HBOND,getMoleculeByName(rna5id),new Location(new Integer(base5id).intValue()), getMoleculeByName(rna3id),new Location(new Integer(base3id).intValue())); */
                //TODO how to add that to the currentRNASecondaryStructure ? As a BaseBaseInteraction ? As an AtomAtomInteraction ?
            } else {
                BaseBaseInteraction interaction;
                /*
                if(BaseBaseInteraction.isAmbiguous(this.currentRNASequence.printSequence().substring(Integer.parseInt(base5id)-1,Integer.parseInt(base5id)).toCharArray()[0], this.currentRNASequence.printSequence().substring(Integer.parseInt(base3id)-1,Integer.parseInt(base3id)).toCharArray()[0], _edge5, _edge3, _orientation))
                    interaction = structuralFactory.createBaseBaseInteraction(_edge5, _edge3, _orientation, BaseBaseInteraction.RESIDUE_5_PRIME);//TODO here there is no way of knowing which residue will be dominant
                else
                    interaction = structuralFactory.createBaseBaseInteraction(_edge5, _edge3, _orientation);
                */
                if (!this.interactions)
                    //its an intra-bp
                    interaction = structuralFactory.createBaseBaseInteraction(_edge5, _edge3, _orientation, BaseBaseInteraction.RESIDUE_1, this.currentRNASecondaryStructure, this.currentRNA, new Location(new Location(new Integer(base5id).intValue()), new Location(new Integer(base3id).intValue())));//TODO here there is no way of knowing which residue will be dominant
                else
                    interaction = structuralFactory.createBaseBaseInteraction(_edge5, _edge3, _orientation, BaseBaseInteraction.RESIDUE_1, this.currentRNASecondaryStructure, getMoleculeByName(rna5id), new Location(new Integer(base5id).intValue()), getMoleculeByName(rna3id), new Location(new Integer(base3id).intValue()));//TODO here there is no way of knowing which residue will be dominant
            }
        }

        private void defineHelix(final Element h) throws FileParsingException {
            final int base5Id;
            final int base3Id;
            final int length;

            try {
                //no check for null exception, all the elements are necessary from an helix core according the DTD
                base5Id = Integer.parseInt(h.getChild("base-id-5p").getChild("base-id").getChild("position").getValue());
                base3Id = Integer.parseInt(h.getChild("base-id-3p").getChild("base-id").getChild("position").getValue());
                length = Integer.parseInt(h.getChild("length").getValue());
            }

            catch (NumberFormatException nfe) {
                throw new FileParsingException(nfe);
            }
            Helix helix = structuralFactory.createHelix(h.getAttributeValue("id"), this.currentRNASecondaryStructure, this.currentRNA, new Location(new Location(base5Id, base5Id + length - 1), new Location(base3Id, base3Id - length + 1)));
        }

        private void defineInterHelix(final Element h) {
            final String[] helixDetails = new String[6];
            //no check for null exception, all the elements are necessary from an helix core according the DTD
            helixDetails[0] = h.getChild("base-id-5p").getChild("base-id").getChild("position").getValue();
            helixDetails[2] = h.getChild("base-id-3p").getChild("base-id").getChild("position").getValue();
            helixDetails[4] = h.getChild("length").getValue();
            helixDetails[5] = h.getAttributeValue("id");

            Element molecule_id = h.getChild("base-id-5p").getChild("base-id").getChild("molecule-id");
            if (molecule_id != null) {
                helixDetails[1] = molecule_id.getAttributeValue("ref");
            } else {
                System.out.println("\nThe inter-helix base n" + helixDetails[0] + " 5boundarie lacks the molecule-id information");
                return;
            }
            molecule_id = h.getChild("base-id-3p").getChild("base-id").getChild("molecule-id");
            if (molecule_id != null) {
                helixDetails[3] = molecule_id.getAttributeValue("ref");
            } else {
                System.out.println("\nThe inter-helix base n" + helixDetails[2] + " 3boundarie lacks the molecule-id information");
                return;
            }
            structuralFactory.createHelix(h.getAttributeValue(helixDetails[5]), this.currentRNASecondaryStructure, getMoleculeByName(helixDetails[1]), new Location(new Integer(helixDetails[0]).intValue(), new Integer(helixDetails[0]).intValue() + new Integer(helixDetails[4]).intValue() - 1), getMoleculeByName(helixDetails[3]), new Location(new Integer(helixDetails[2]).intValue(), new Integer(helixDetails[2]).intValue() - new Integer(helixDetails[4]).intValue() + 1));
            //since to reference to a model inside the interactions markup, We consider that the interHelix apply to all RNASecondaryStructure features constructed so far
            //TODO this is an error in the DTD => we should try to improve that
        }


        private Molecule getMoleculeByName(String name) {
            for (Molecule m : this.molecules.values())
                if (name.equals(m.getName()))
                    return m;
            return null;
        }

    }

    private class ParadiseRnaMLParser1_0 {

        private int ssCount, tsCount;
        private RNA currentRNA;
        private boolean interMolecular;
        private HashMap<Integer, ResidueIdentity> columnToResidueIdentity;
        private StructuralAlignment currentStructuralAlignment;
        private TertiaryStructure currentTertiaryStructure;
        private RNASecondaryStructure currentSecondaryStructure;
        private SecondaryStructureDisplay currentSecondaryStructureDisplay;
        private Map<String, Molecule> molecules;
        private Map<String, RNASecondaryStructure> secondaryStructures;
        private Map<String, StructuralAlignment> structuralAlignments;
        private Map<String, TertiaryStructure> tertiaryStructures;

        private ParadiseRnaMLParser1_0() {
            this.secondaryStructures = new HashMap<String, RNASecondaryStructure>();
            this.structuralAlignments = new HashMap<String, StructuralAlignment>();
            this.tertiaryStructures = new HashMap<String, TertiaryStructure>();
            this.molecules = new HashMap<String, Molecule>();
        }

        private void parse(final File rnamlFile, Element root) throws fr.unistra.ibmc.paradise.BiologicalSymbolException, FileParsingException, IOException, JDOMException, InterruptedException {
            this.traverseJDOMElement(rnamlFile,root);
            //we are constructing the single-strands for each secondary structure
            Location single_strands_location;
            Location helixLocation;
            for (RNASecondaryStructure ss : secondaryStructures.values()) {
                for (Molecule m : ss.getMolecules()) {
                    single_strands_location = new Location(1, m.getLength());
                    for (Helix h : ss.getHelices()) {
                        //the location of this helix for the current molecule
                        helixLocation = h.getFullLocation(m);
                        if (helixLocation != null)
                            single_strands_location.remove(helixLocation);
                    }
                    //once all the locations of the helices attached to the current molecule removed, we have the boundaries of all the single-strands for this molecule
                    int sstrand_count = 0;
                    int[] boundaries = single_strands_location.getBoundaries();
                    for (int i = 0; i < boundaries.length; i += 2)
                        structuralFactory.createSingleStrand("SS" + (sstrand_count++), ss, m, new Location(boundaries[i], boundaries[i + 1]));
                }
            }
        }

        private void traverseJDOMElement(final File rnamlFile, final Element node) throws fr.unistra.ibmc.paradise.BiologicalSymbolException, FileParsingException, InterruptedException {
            if (Thread.interrupted())
                throw new InterruptedException();
            Element child = null;
            String name = null;

            for (Iterator i = node.getChildren().iterator(); i.hasNext();) {
                child = (Element) i.next();
                name = child.getName();
                if (name.equals("molecule")) {
                    String rnaSequence = "", moleculeName = "";
                    Element identity = child.getChild("identity");
                    if (identity != null) {
                        Element nameEl = identity.getChild("name");
                        moleculeName = nameEl != null ? nameEl.getValue() : "" + molecules.size();
                    }
                    Element sequence = child.getChild("sequence");
                    if (sequence != null) {
                        Element seqdata = sequence.getChild("seq-data");
                        if (seqdata != null)
                            rnaSequence = seqdata.getValue().trim();
                    }
                    this.currentRNA = moleculeFactory.createRNA(new ParadiseID(), moleculeName, Source.getFileAsSource(rnamlFile), rnaSequence);
                    molecules.put(child.getAttributeValue("id"), this.currentRNA);
                } else if (name.equals("tertiary-structure")) {
                    //perhaps we have already constructed an empty Tertiary Structure because a Secondary structure referenced it
                    this.currentTertiaryStructure = tertiaryStructures.get(new ParadiseID());
                    //if not we create a new one
                    if (this.currentTertiaryStructure == null) {
                        this.currentTertiaryStructure = structuralFactory.createTertiaryStructure(new ParadiseID(), "Tertiary Structure"+(++tsCount), Source.getFileAsSource(rnamlFile));
                        tertiaryStructures.put(child.getAttributeValue("id"), this.currentTertiaryStructure);
                    }
                    for (Object o : child.getChildren("base")) {
                        Element base = (Element) o;
                        int position = Integer.parseInt(base.getAttributeValue("base-id"));
                        Residue3D r = structuralFactory.createResidue3D(this.currentTertiaryStructure, this.currentRNA, position);
                        for (Object e : base.getChildren("atom")) {
                            Element atom = (Element) e;
                            r.setAtomCoordinates(atom.getAttributeValue("type"), Float.parseFloat(atom.getAttributeValue("x")), Float.parseFloat(atom.getAttributeValue("y")), Float.parseFloat(atom.getAttributeValue("z")));
                        }
                    }
                } else if (name.equals("structure-annotation")) {
                    if (child.getAttribute("alignment-id") != null) {
                        StructuralAlignment al = structuralAlignments.get(child.getAttributeValue("alignment-id"));
                        //if this alignment is not already constructed
                        if (al == null) {
                            //we generate an empty one
                            al = identityFactory.createStructuralAligment(new ParadiseID(), "?", Source.getFileAsSource(rnamlFile), this.currentRNA, new Location());
                            structuralAlignments.put(child.getAttributeValue("id"), al);
                        }
                    }
                    if (interMolecular) {
                        this.currentSecondaryStructure = secondaryStructures.get(child.getAttributeValue("id"));
                        for (Object e : child.getChildren("base-pair")) {
                            Element bp = (Element) e;
                            structuralFactory.createBaseBaseInteraction(bp.getAttributeValue("edge1").toCharArray()[0], bp.getAttributeValue("edge2").toCharArray()[0], bp.getAttributeValue("orientation").toCharArray()[0], BaseBaseInteraction.RESIDUE_1, this.currentSecondaryStructure, molecules.get(new ParadiseID()), new Location(Integer.parseInt(bp.getAttributeValue("base1-id"))), molecules.get(new ParadiseID()), new Location(Integer.parseInt(bp.getAttributeValue("base2-id"))));
                        }
                        for (Object e : child.getChildren("atoms-interaction")) {
                            Element atomInteraction = (Element) e;
                            structuralFactory.createAtomAtomInteraction(this.currentSecondaryStructure, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], molecules.get(new ParadiseID()), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id"))), molecules.get(new ParadiseID()), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base2-id"))));
                        }
                    } else {
                        //perhaps we have already constructed an empty Secondary Structure because a Tertiary structure referenced it
                        this.currentSecondaryStructure = secondaryStructures.get(child.getAttributeValue("id"));
                        //if not we create a new one
                        if (this.currentSecondaryStructure == null) {
                            this.currentSecondaryStructure = structuralFactory.createRNASecondaryStructure(new ParadiseID(), "?", Source.getFileAsSource(rnamlFile), this.currentRNA, new Location());
                            secondaryStructures.put(child.getAttributeValue("id"), this.currentSecondaryStructure);
                        }
                        for (Object e : child.getChildren("base-pair")) {
                            Element bp = (Element) e;
                            structuralFactory.createBaseBaseInteraction(bp.getAttributeValue("edge5").toCharArray()[0], bp.getAttributeValue("edge3").toCharArray()[0], bp.getAttributeValue("orientation").toCharArray()[0], BaseBaseInteraction.RESIDUE_1, this.currentSecondaryStructure, this.currentRNA, new Location(Integer.parseInt(bp.getAttributeValue("base5-id")), Integer.parseInt(bp.getAttributeValue("base3-id"))));
                        }
                        for (Object e : child.getChildren("atoms-interaction")) {
                            Element atomInteraction = (Element) e;
                            structuralFactory.createAtomAtomInteraction(this.currentSecondaryStructure, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], this.currentRNA, new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id")), Integer.parseInt(atomInteraction.getAttributeValue("base2-id"))));
                        }
                        //if this secondarystructure is linked to a tertiarystructure
                        if (child.getAttributeValue("tertiary-structure-id") != null) {
                            TertiaryStructure ts = tertiaryStructures.get(child.getAttributeValue("tertiary-structure-id"));
                            //if this tertiarystructure is not already constructed
                            if (ts == null) {
                                //we generate an empty one
                                ts = structuralFactory.createTertiaryStructure(new ParadiseID(), "TS"+(++tsCount),Source.getFileAsSource(rnamlFile), this.currentRNA, new Location());
                                tertiaryStructures.put(child.getAttributeValue("id"), ts);
                            }
                            ts.linkSecondaryStructure(this.currentSecondaryStructure);
                        }
                    }
                } else if (name.equals("helix")) {
                    Helix h;
                    if (interMolecular) {
                        int _5end = Integer.parseInt(child.getAttributeValue("base5-id")),
                                _3end = Integer.parseInt(child.getAttributeValue("base3-id")),
                                length = Integer.parseInt(child.getAttributeValue("length"));
                        h = structuralFactory.createHelix(child.getAttributeValue("id"), this.currentSecondaryStructure, molecules.get(new ParadiseID()), new Location(_5end, _5end + length - 1), molecules.get(new ParadiseID()), new Location(_3end - length + 1, _3end));
                        //can be null if the interhelix construction has been descativated
                        if (h != null)
                            for (Object o : child.getChildren("base-pair")) {
                                Element bp = (Element) o;
                                BaseBaseInteraction interaction = structuralFactory.createBaseBaseInteraction(bp.getAttributeValue("edge1").toCharArray()[0], bp.getAttributeValue("edge2").toCharArray()[0], bp.getAttributeValue("orientation").toCharArray()[0], BaseBaseInteraction.RESIDUE_1, this.currentSecondaryStructure, molecules.get(new ParadiseID()), new Location(Integer.parseInt(bp.getAttributeValue("base1-id"))), molecules.get(new ParadiseID()), new Location(Integer.parseInt(bp.getAttributeValue("base2-id"))));
                                for (Object e : bp.getChildren("atoms-interaction")) {
                                    Element atomInteraction = (Element) e;
                                    structuralFactory.createAtomAtomInteraction(interaction, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], molecules.get(new ParadiseID()), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id"))), molecules.get(new ParadiseID()), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base2-id"))));
                                }
                            }
                    } else {
                        int _5end = Integer.parseInt(child.getAttributeValue("base5-id")),
                                _3end = Integer.parseInt(child.getAttributeValue("base3-id")),
                                length = Integer.parseInt(child.getAttributeValue("length"));
                        h = structuralFactory.createHelix(child.getAttributeValue("id"), this.currentSecondaryStructure, this.currentRNA, new Location(new Location(_5end, _5end + length - 1), new Location(_3end - length + 1, _3end)));
                        for (Object o : child.getChildren("base-pair")) {
                            Element bp = (Element) o;
                            BaseBaseInteraction interaction = structuralFactory.createBaseBaseInteraction(bp.getAttributeValue("edge5").toCharArray()[0], bp.getAttributeValue("edge3").toCharArray()[0], bp.getAttributeValue("orientation").toCharArray()[0], BaseBaseInteraction.RESIDUE_1, this.currentSecondaryStructure, this.currentRNA, new Location(new Location(Integer.parseInt(bp.getAttributeValue("base5-id"))), new Location(Integer.parseInt(bp.getAttributeValue("base3-id")))));
                            for (Object e : bp.getChildren("atoms-interaction")) {
                                Element atomInteraction = (Element) e;
                                structuralFactory.createAtomAtomInteraction(interaction, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], this.currentRNA, new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id")), Integer.parseInt(atomInteraction.getAttributeValue("base2-id"))));
                            }
                        }
                    }
                } else if (name.equals("inter-molecular-interactions"))
                    this.interMolecular = true;
                else if (name.equals("secondary-structure-display")) {
                    this.currentSecondaryStructureDisplay = structuralFactory.createSecondaryStructureDisplay(this.currentSecondaryStructure, this.currentRNA, this.currentSecondaryStructure.getFullLocation(this.currentRNA));
                    for (Object o : child.getChildren("ss-base-coord")) {
                        Element coord = (Element) o;
                        structuralFactory.createResidue2D(this.currentSecondaryStructureDisplay, Float.parseFloat(coord.getAttributeValue("x")), Float.parseFloat(coord.getAttributeValue("y")), this.currentRNA, Integer.parseInt(coord.getAttributeValue("base-id")));
                    }
                } else if (child.getName().equals("alignment")) {
                    //perhaps we have already constructed an empty StructuralAlignment because a Secondary structure referenced it
                    this.currentStructuralAlignment = structuralAlignments.get(child.getAttributeValue("id"));
                    //if not we create a new one
                    if (this.currentStructuralAlignment == null) {
                        this.currentStructuralAlignment = identityFactory.createStructuralAligment(new ParadiseID(),"Structural Alignment", Source.getFileAsSource(rnamlFile));
                        structuralAlignments.put(child.getAttributeValue("id"), this.currentStructuralAlignment);
                    }
                    columnToResidueIdentity = new HashMap<Integer, ResidueIdentity>();
                    for (Object o : child.getChildren("ali-sequence")) {
                        Element c = (Element) o;
                        Molecule m = molecules.get(c.getAttributeValue("molecule-id"));
                        String sequence = c.getValue();
                        char[] cs = sequence.toCharArray();
                        int positionInSequence = 0;
                        for (int j = 0; j < cs.length; j++) {
                            if (cs[j] != '-') {
                                positionInSequence++;
                                ResidueIdentity ri = columnToResidueIdentity.get(j);
                                if (ri != null)
                                    m.addFeature(ri, new Location(positionInSequence));
                                else
                                    columnToResidueIdentity.put(j, identityFactory.createResidueIdentity(this.currentStructuralAlignment, m, new Location(positionInSequence)));
                            }
                        }
                        if (c.getAttribute("structural-annotation-id") != null) {
                            RNASecondaryStructure _ss = secondaryStructures.get(c.getAttributeValue("structural-annotation-id"));
                            if (_ss == null) {
                                _ss = structuralFactory.createRNASecondaryStructure(new ParadiseID(), "SS"+(++ssCount),Source.getFileAsSource(rnamlFile), this.currentRNA, new Location());
                                secondaryStructures.put(child.getAttributeValue("id"), _ss);
                            }
                            this.currentStructuralAlignment.linkSecondaryStructure(_ss);
                        }
                    }
                    if (child.getAttribute("structure-annotation-id") != null) {
                        RNASecondaryStructure _ss = secondaryStructures.get(child.getAttributeValue("structure-annotation-id"));
                        //since we have added the ali-sequence to the currentStructuralAlignment, we're sure that the alignement annotates a molecule before to be added as a subfeature
                        _ss.linkStructuralAlignment(this.currentStructuralAlignment);
                    }
                }
                this.traverseJDOMElement(rnamlFile, child);
            }
        }

    }

    private class ParadiseRnaMLParser2_0 {

        private int ssCount, tsCount, alignmentCount;
        private HashMap<Integer, ResidueIdentity> columnToResidueIdentity;
        private Map<String, Molecule> molecules;
        private Map<String, RNASecondaryStructure> secondaryStructures;
        private Map<String, StructuralAlignment> structuralAlignments;
        private Map<String, TertiaryStructure> tertiaryStructures;

        private ParadiseRnaMLParser2_0() {
            this.secondaryStructures = new HashMap<String, RNASecondaryStructure>();
            this.structuralAlignments = new HashMap<String, StructuralAlignment>();
            this.tertiaryStructures = new HashMap<String, TertiaryStructure>();
            this.molecules = new HashMap<String, Molecule>();
        }

        private void parse(final Source source, Element root) throws fr.unistra.ibmc.paradise.BiologicalSymbolException, FileParsingException, IOException, JDOMException, InterruptedException {
            this.traverseJDOMElement(source, root);
            //we are constructing the single-strands for each secondary structure
            Location single_strands_location;
            Location helixLocation;
            for (RNASecondaryStructure ss : secondaryStructures.values()) {
                for (Molecule m : ss.getMolecules()) {
                    single_strands_location = new Location(1, m.getLength());
                    for (Helix h : ss.getHelices()) {
                        //the location of this helix for the current molecule
                        helixLocation = h.getFullLocation(m);
                        if (helixLocation != null)
                            single_strands_location.remove(helixLocation);
                    }
                    //once all the locations of the helices attached to the current molecule removed, we have the boundaries of all the single-strands for this molecule
                    int sstrand_count = 0;
                    int[] boundaries = single_strands_location.getBoundaries();
                    for (int i = 0; i < boundaries.length; i += 2)
                        structuralFactory.createSingleStrand("SS" + (sstrand_count++), ss, m, new Location(boundaries[i], boundaries[i + 1]));
                }
            }
        }

        private void traverseJDOMElement(final Source source, final Element node) throws fr.unistra.ibmc.paradise.BiologicalSymbolException, FileParsingException, InterruptedException {
            if (Thread.interrupted())
                throw new InterruptedException();
            Element child = null;
            String name = null;

            for (Iterator i = node.getChildren().iterator(); i.hasNext();) {
                child = (Element) i.next();
                name = child.getName();
                if (name.equals("molecule")) {
                    String moleculeSequence = "", moleculeName = "";
                    Element identity = child.getChild("identity");
                    String type = "rna";
                    Attribute _type = child.getAttribute("type");
                    if (_type != null)
                        type = _type.getValue();
                    if (identity != null) {
                        Element nameEl = identity.getChild("name");
                        moleculeName = nameEl != null ? nameEl.getValue() : "rna" + molecules.size();
                    }
                    Element sequence = child.getChild("sequence");
                    if (sequence != null) {
                        Element seqdata = sequence.getChild("seq-data");
                        if (seqdata != null)
                            moleculeSequence = seqdata.getValue().trim();
                    }
                    if ("rna".equals(type))
                        this.molecules.put(child.getAttributeValue("id"), moleculeFactory.createRNA(new ParadiseID(), moleculeName,source, moleculeSequence));
                    else if ("protein".equals(type))
                        this.molecules.put(child.getAttributeValue("id"), moleculeFactory.createProtein(new ParadiseID(), moleculeName, source , moleculeSequence));
                } else if (name.equals("tertiary-structure")) {
                    //perhaps we have already constructed an empty Tertiary Structure because a Secondary structure referenced it
                    TertiaryStructure ts = tertiaryStructures.get(child.getAttributeValue("id"));
                    //if not we create a new one
                    if (ts == null) {
                        ts = structuralFactory.createTertiaryStructure(new ParadiseID(), "TS"+(++tsCount),source);
                        tertiaryStructures.put(child.getAttributeValue("id"), ts);
                    }
                    ts.setSource(source);
                    ts.setName(child.getAttributeValue("name"));
                    for (Object o : child.getChildren("base")) {
                        Element base = (Element) o;
                        int position = Integer.parseInt(base.getAttributeValue("base-id"));
                        Residue3D r = structuralFactory.createResidue3D(ts, molecules.get(base.getAttributeValue("molecule-id")), position);
                        for (Object e : base.getChildren("atom")) {
                            Element atom = (Element) e;
                            r.setAtomCoordinates(atom.getAttributeValue("type"), Float.parseFloat(atom.getAttributeValue("x")), Float.parseFloat(atom.getAttributeValue("y")), Float.parseFloat(atom.getAttributeValue("z")));
                        }
                    }
                } else if (name.equals("structure-annotation")) {
                    //perhaps we have already constructed an empty Secondary Structure because a Tertiary structure referenced it
                    RNASecondaryStructure ss = secondaryStructures.get(child.getAttributeValue("id"));
                    //if not we create a new one
                    if (ss == null) {
                        ss = structuralFactory.createRNASecondaryStructure(new ParadiseID(), "SS"+(++ssCount),source);
                        secondaryStructures.put(child.getAttributeValue("id"), ss);
                    }
                    ss.setSource(source);
                    ss.setName(child.getAttributeValue("name"));
                    for (Object e : child.getChildren("base-pair")) {
                        Element bp = (Element) e;
                        BaseBaseInteraction interaction = null;
                        if (bp.getAttributeValue("molecule1-id").equals(bp.getAttributeValue("molecule2-id")))
                            interaction = structuralFactory.createBaseBaseInteraction(bp.getAttributeValue("edge1").toCharArray()[0], bp.getAttributeValue("edge2").toCharArray()[0], bp.getAttributeValue("orientation").toCharArray()[0], BaseBaseInteraction.RESIDUE_1, ss, molecules.get(bp.getAttributeValue("molecule1-id")), new Location(new Location(Integer.parseInt(bp.getAttributeValue("base1-id"))), new Location(Integer.parseInt(bp.getAttributeValue("base2-id")))));
                        else
                            interaction = structuralFactory.createBaseBaseInteraction(bp.getAttributeValue("edge1").toCharArray()[0], bp.getAttributeValue("edge2").toCharArray()[0], bp.getAttributeValue("orientation").toCharArray()[0], BaseBaseInteraction.RESIDUE_1, ss, molecules.get(bp.getAttributeValue("molecule1-id")), new Location(Integer.parseInt(bp.getAttributeValue("base1-id")), Integer.parseInt(bp.getAttributeValue("base1-id"))), molecules.get(bp.getAttributeValue("molecule2-id")), new Location(Integer.parseInt(bp.getAttributeValue("base2-id")), Integer.parseInt(bp.getAttributeValue("base2-id"))));
                        for (Object _e : bp.getChildren("atoms-interaction")) {
                            Element atomInteraction = (Element) _e;
                            if (atomInteraction.getAttributeValue("molecule1-id").equals(atomInteraction.getAttributeValue("molecule2-id")))
                                structuralFactory.createAtomAtomInteraction(interaction, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], molecules.get(atomInteraction.getAttributeValue("molecule1-id")), new Location(new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id"))), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base2-id")))));
                            else
                                structuralFactory.createAtomAtomInteraction(interaction, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], molecules.get(atomInteraction.getAttributeValue("molecule1-id")), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id"))), molecules.get(atomInteraction.getAttributeValue("molecule2-id")), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base2-id"))));
                        }
                    }
                    for (Object e : child.getChildren("atoms-interaction")) {
                        Element atomInteraction = (Element) e;
                        if (atomInteraction.getAttributeValue("molecule1-id").equals(atomInteraction.getAttributeValue("molecule2-id")))
                            structuralFactory.createAtomAtomInteraction(ss, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], molecules.get(atomInteraction.getAttributeValue("molecule1-id")), new Location(new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id"))), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base2-id")))));
                        else
                            structuralFactory.createAtomAtomInteraction(ss, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], molecules.get(atomInteraction.getAttributeValue("molecule1-id")), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id")), Integer.parseInt(atomInteraction.getAttributeValue("base1-id"))), molecules.get(atomInteraction.getAttributeValue("molecule2-id")), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base2-id")), Integer.parseInt(atomInteraction.getAttributeValue("base2-id"))));
                    }
                    for (Object e : child.getChildren("helix")) {
                        Element helixElement = (Element) e;
                        int _5end = Integer.parseInt(helixElement.getAttributeValue("base5-id")),
                                _3end = Integer.parseInt(helixElement.getAttributeValue("base3-id")),
                                length = Integer.parseInt(helixElement.getAttributeValue("length"));
                        Helix h = null;
                        if (helixElement.getAttributeValue("molecule1-id").equals(helixElement.getAttributeValue("molecule2-id"))) {
                            h = structuralFactory.createHelix(helixElement.getAttributeValue("id"), ss, molecules.get(helixElement.getAttributeValue("molecule1-id")), new Location(new Location(_5end, _5end + length - 1), new Location(_3end - length + 1, _3end)));
                        } else {
                            h = structuralFactory.createHelix(helixElement.getAttributeValue("id"), ss, molecules.get(helixElement.getAttributeValue("molecule1-id")), new Location(_5end, _5end + length - 1), molecules.get(helixElement.getAttributeValue("molecule2-id")), new Location(_3end - length + 1, _3end));
                        }
                        //can be null if the interhelix construction has been desactivated
                        if (h != null)
                            for (Object o : helixElement.getChildren("base-pair")) {
                                Element bp = (Element) o;
                                BaseBaseInteraction interaction = null;
                                if (bp.getAttributeValue("molecule1-id").equals(bp.getAttributeValue("molecule2-id")))
                                    interaction = structuralFactory.createBaseBaseInteraction(bp.getAttributeValue("edge1").toCharArray()[0], bp.getAttributeValue("edge2").toCharArray()[0], bp.getAttributeValue("orientation").toCharArray()[0], BaseBaseInteraction.RESIDUE_1, h, molecules.get(bp.getAttributeValue("molecule1-id")), new Location(new Location(Integer.parseInt(bp.getAttributeValue("base1-id"))), new Location(Integer.parseInt(bp.getAttributeValue("base2-id")))));
                                else
                                    interaction = structuralFactory.createBaseBaseInteraction(bp.getAttributeValue("edge1").toCharArray()[0], bp.getAttributeValue("edge2").toCharArray()[0], bp.getAttributeValue("orientation").toCharArray()[0], BaseBaseInteraction.RESIDUE_1, h, molecules.get(bp.getAttributeValue("molecule1-id")), new Location(Integer.parseInt(bp.getAttributeValue("base1-id"))), molecules.get(bp.getAttributeValue("molecule2-id")), new Location(Integer.parseInt(bp.getAttributeValue("base2-id"))));
                                for (Object _e : bp.getChildren("atoms-interaction")) {
                                    Element atomInteraction = (Element) _e;
                                    if (atomInteraction.getAttributeValue("molecule1-id").equals(atomInteraction.getAttributeValue("molecule2-id")))
                                        structuralFactory.createAtomAtomInteraction(interaction, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], molecules.get(atomInteraction.getAttributeValue("molecule1-id")), new Location(new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id"))), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base2-id")))));
                                    else
                                        structuralFactory.createAtomAtomInteraction(interaction, atomInteraction.getAttributeValue("atom1-type"), atomInteraction.getAttributeValue("atom2-type"), atomInteraction.getAttributeValue("type").toCharArray()[0], molecules.get(atomInteraction.getAttributeValue("molecule1-id")), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base1-id"))), molecules.get(atomInteraction.getAttributeValue("molecule2-id")), new Location(Integer.parseInt(atomInteraction.getAttributeValue("base2-id"))));
                                }
                            }
                    }
                    for (Object e : child.getChildren("secondary-structure-display")) {
                        Element displayElement = (Element) e;
                        Molecule m = ss.getMolecules().iterator().next();
                        SecondaryStructureDisplay display = structuralFactory.createSecondaryStructureDisplay(ss, m, ss.getFullLocation(m));
                        for (Object o : displayElement.getChildren("ss-base-coord")) {
                            Element coord = (Element) o;
                            structuralFactory.createResidue2D(display, Float.parseFloat(coord.getAttributeValue("x")), Float.parseFloat(coord.getAttributeValue("y")), molecules.get(coord.getAttributeValue("molecule-id")), Integer.parseInt(coord.getAttributeValue("base-id")));
                        }
                    }
                    if (child.getAttribute("alignment-id") != null) {
                        StructuralAlignment al = structuralAlignments.get(child.getAttributeValue("alignment-id"));
                        try {

                            //if this alignment is not already constructed
                            if (al == null) {
                                //we generate an empty one
                                //we have to find at least one molecule annotated by this upcoming alignment
                                Molecule m = molecules.get(((org.jdom.Element) XPath.selectSingleNode(document, "/rnaml//alignment[@id=" + child.getAttributeValue("alignment-id") + "]")).getAttributeValue("molecule-ids").split(" ")[0]);
                                al = identityFactory.createStructuralAligment(new ParadiseID(), "Alignment"+(++alignmentCount),source,m, new Location());
                                structuralAlignments.put(child.getAttributeValue("alignment-id"), al);
                            }
                            al.linkSecondaryStructure(ss);
                        }
                        catch (JDOMException e) {
                            throw new FileParsingException(e);
                        }
                    }
                    //if this secondarystructure is linked to a tertiarystructure
                    if (child.getAttribute("tertiary-structure-id") != null) {
                        TertiaryStructure ts = tertiaryStructures.get(child.getAttributeValue("tertiary-structure-id"));
                        try {
                            //if this tertiarystructure is not already constructed
                            if (ts == null) {
                                //we generate an empty one
                                //we have to find at least one molecule annotated by this upcoming tertiary-structure
                                Molecule m = molecules.get(((org.jdom.Element) XPath.selectSingleNode(document, "/rnaml//tertiary-structure[@id=" + child.getAttributeValue("tertiary-structure-id") + "]")).getAttributeValue("molecule-ids").split(" ")[0]);
                                ts = structuralFactory.createTertiaryStructure(new ParadiseID(), "TS"+(++tsCount),source, m, new Location());
                                tertiaryStructures.put(child.getAttributeValue("tertiary-structure-id"), ts);
                            }
                            //TODO [fjossinet] temporary trick to add a secondary structure even if it annotates nothing (for example, the secondary structure of a protein is an empty element and consequently annotates no molecule
                            if (ss.annotatesNothing())
                                for (Molecule m : ts.getMolecules())
                                    m.addFeature(ss, new Location(ts.getFullLocation(m)));
                            ts.linkSecondaryStructure(ss);
                        }
                        catch (JDOMException e) {
                            throw new FileParsingException(e);
                        }
                    }
                } else if (child.getName().equals("alignment")) {
                    //perhaps we have already constructed an empty StructuralAlignment because a Secondary structure referenced it
                    StructuralAlignment alignment = structuralAlignments.get(child.getAttributeValue("id"));
                    //if not we create a new one
                    if (alignment == null) {
                        alignment = identityFactory.createStructuralAligment(new ParadiseID(), "Alignment"+(++alignmentCount),source);
                        structuralAlignments.put(child.getAttributeValue("id"), alignment);
                    }
                    alignment.setName(child.getAttributeValue("name"));
                    alignment.setSource(source);
                    columnToResidueIdentity = new HashMap<Integer, ResidueIdentity>();
                    for (Object o : child.getChildren("ali-sequence")) {
                        Element c = (Element) o;
                        Molecule m = molecules.get(c.getAttributeValue("molecule-id"));
                        String sequence = c.getValue();
                        char[] cs = sequence.toCharArray();
                        int positionInSequence = 0;
                        for (int j = 0; j < cs.length; j++) {
                            if (cs[j] != '-') {
                                positionInSequence++;
                                ResidueIdentity ri = columnToResidueIdentity.get(j);
                                if (ri != null)
                                    m.addFeature(ri, new Location(positionInSequence));
                                else
                                    columnToResidueIdentity.put(j, identityFactory.createResidueIdentity(alignment, m, new Location(positionInSequence)));
                            }
                        }
                        if (c.getAttribute("structural-annotation-id") != null) {
                            RNASecondaryStructure _ss = secondaryStructures.get(c.getAttributeValue("structure-annotation-id"));
                            if (_ss == null) {
                                //we generate an empty one
                                //we have to find at least one molecule annotated by this upcoming tertiary-structure
                                try {
                                    Molecule _m = molecules.get(((org.jdom.Element) XPath.selectSingleNode(document, "/rnaml//structural-annotation[@id=" + child.getAttributeValue("structural-annotation-id") + "]")).getAttributeValue("molecule-ids").split(" ")[0]);
                                    _ss = structuralFactory.createRNASecondaryStructure(new ParadiseID(), "SS"+(++ssCount),source, _m, new Location());
                                    secondaryStructures.put(c.getAttributeValue("structural-annotation-id"), _ss);
                                }
                                catch (JDOMException e) {
                                    throw new FileParsingException(e);
                                }
                            }
                            alignment.linkSecondaryStructure(_ss);
                        }
                    }
                    if (child.getAttribute("structure-annotation-id") != null) {
                        RNASecondaryStructure _ss = secondaryStructures.get(child.getAttributeValue("structure-annotation-id"));
                        //since we have added the ali-sequence to the currentStructuralAlignment, we're sure that the alignement annotates a molecule before to be added as a subfeature
                        _ss.linkStructuralAlignment(alignment);
                    }
                }
                this.traverseJDOMElement(source, child);
            }
        }

    }

}
