package fr.unistra.ibmc.paradise.core.io;

import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.tools.component.ProgressMonitor;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.util.Comparator;

public class StockholmFileIO extends AbstractParadiseFileIO {

    public static final String[] suffixes= new String[]{"stk","txt"};

    private List<StructuralAlignment> structuralAlignments;
    private IdentityFeatureFactory identityFeatureFactory;
    protected StructuralFeatureFactory structuralFeatureFactory;
    protected List<Molecule> molecules;
    protected List<SecondaryStructure> secondaryStructures;

    public StockholmFileIO(MoleculeFactory moleculeFactory) {
        super("Stockholm File parser", moleculeFactory);
    }

    public javax.swing.filechooser.FileFilter getFileFilter() {
        return new javax.swing.filechooser.FileFilter() {
            public boolean accept(File file) {
                if (file.isDirectory())
                    return true;
                for (String suffix:StockholmFileIO.suffixes)
                    if (file.getName().endsWith(suffix))
                        return true;
                return false;
            }

            public String getDescription() {
                return "Stockholm files";
            }
        };
    }

    public List<SecondaryStructure> getSecondaryStructures() {
        return this.secondaryStructures;
    }

    public List<TertiaryStructure> getTertiaryStructures() {
        return new ArrayList<TertiaryStructure>();
    }

    public List<Molecule> getMolecules() {
        return this.molecules;
    }

    public List<StructuralAlignment> getStructuralAlignments() {
        return this.structuralAlignments;
    }

    public void parseFile(File f, ProgressMonitor monitor) throws FileParsingException {
        this.structuralFeatureFactory = this.createStructuralFeatureFactory();
        this.identityFeatureFactory = this.createIdentityFeatureFactory();
        this.molecules = new ArrayList<Molecule>();
        List<String> alignedSequences = new ArrayList<String>();
        this.secondaryStructures = new ArrayList<SecondaryStructure>();
        this.structuralAlignments = new ArrayList<StructuralAlignment>();
        StructuralAlignment structuralAlignment = this.identityFeatureFactory.createStructuralAligment(new ParadiseID(), "alignment", Source.getFileAsSource(f));
        StructuralIdentity si = null;
        this.structuralAlignments.add(structuralAlignment);
        StringBuffer brackets = new StringBuffer();
        Location gapsPositions = null;
        try {
            BufferedReader in = new BufferedReader(new FileReader(f));
            String line;
            boolean ssConsReached = false;
            while ((line = in.readLine()) != null) {
                if (line.startsWith("#=GC SS_cons")) {
                    ssConsReached = true;
                    String[] tokens = line.trim().split("#=GC SS_cons");
                    brackets.append(tokens[1].trim().replace('<', '(').replace('>', ')'));
                    for (StructuralIdentity _si:structuralAlignment.getStructuralIdentities(false)) {
                        for (Molecule m:_si.getMolecules()) {
                            this.generateSecondaryStructure(f,m, alignedSequences.get(molecules.indexOf(m)),brackets);
                            this.secondaryStructures.get(secondaryStructures.size()-1).linkStructuralAlignment(structuralAlignment);
                        }
                    }
                    break;
                }
                else if (line.startsWith("# STOCKHOLM"))
                    continue;
                else if (!ssConsReached) {
                    if (line.trim().length() == 0) //if an empty line
                        continue;
                    line = line.replace('.','-'); //if the dot character has been chosen for the gaps
                    String[] tokens = line.trim().split("\\s+");
                    String moleculeName = tokens[0], alignedSequence = tokens[1];
                    gapsPositions = new Location();
                    char[] characters = alignedSequence.toCharArray();
                    for (int i=0; i< characters.length;i++)
                        if (characters[i] == '-')
                            gapsPositions.add(i+1);
                    Molecule molecule = this.moleculeFactory.createRNA(new ParadiseID(), moleculeName , Source.getFileAsSource(f),alignedSequence.replaceAll("-",""));
                    molecules.add(molecule);
                    alignedSequences.add(alignedSequence);
                    if (si == null)
                        si = this.identityFeatureFactory.createStructuralIdentity(structuralAlignment,"si",molecule,new Location(1,molecule.getLength()));
                    else
                        molecule.addFeature(si,new Location(1,molecule.getLength()));
                    si.addGapsLocation(molecule,gapsPositions);
                }
            }
            in.close();
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new FileParsingException(e);
        }
    }

    private void generateSecondaryStructure(File f, Molecule m, String alignedSequence, StringBuffer brackets) throws FileParsingException, BiologicalSymbolException {
        //System.out.println(brackets.toString());
        //System.out.println(alignedSequence);
        List<Location> helices = new ArrayList<Location>();
        List<Location> tertiaryInteractions = new ArrayList<Location>();
        List<Location> allpairs = new ArrayList<Location>();
        if (brackets.length() == alignedSequence.length()) {
            Location locationInTheMolecule = new Location(), locationInTheAlignedSequence = new Location(); //locations to store the positions of the left brackets
            this.secondaryStructures.add(structuralFeatureFactory.createRNASecondaryStructure(new ParadiseID(), "SS "+f.getName() ,Source.getFileAsSource(f), m, new Location(1, m.getLength())));
            int moleculePosition = 0;
            for (int i = 0; i < brackets.length() ; i++) {
                if (alignedSequence.charAt(i) != '-')
                    moleculePosition++;
                char c = brackets.charAt(i);
                switch (c) {
                    case '(' :
                        locationInTheMolecule.add(moleculePosition) ;
                        locationInTheAlignedSequence.add(i);
                        break;
                    case ')' :
                        if (alignedSequence.charAt(locationInTheAlignedSequence.getEnd()) != '-' && alignedSequence.charAt(i) != '-') {
                            allpairs.add(new Location(new Location(locationInTheMolecule.getEnd()), new Location(moleculePosition)));
                        }
                        locationInTheMolecule.remove(locationInTheMolecule.getEnd()) ;
                        locationInTheAlignedSequence.remove(locationInTheAlignedSequence.getEnd());
                        break;
                }
            }
            Collections.sort(allpairs, new Comparator() {
                public int compare(Object o, Object o1) {
                    return ((Location)o).getStart()-((Location)o1).getStart();
                }
            });
            Location allSingleStrandsLocation = new Location(1, m.getLength());
            Location l = null;
            for (int i = 0; i < allpairs.size()-1 ; i++) {
                Location l1 = allpairs.get(i), l2 = allpairs.get(i+1);
                if (l1.getStart()+1 == l2.getStart() && l1.getEnd()-1 == l2.getEnd()) { //if the basepairs are contiguous
                    if (l == null)
                        l = new Location(l1, l2);
                    else {
                        l.add(l1);
                        l.add(l2);
                    }
                }
                else {
                    if (l != null) {
                        if (l.getLength() == 2)
                            tertiaryInteractions.add(l);
                        else {
                            helices.add(l);
                            allSingleStrandsLocation.remove(l);
                        }
                    }
                    l = null;
                }
            }
            //last helix
            if (l != null) {
                if (l.getLength() == 2)
                    tertiaryInteractions.add(l);
                else {
                    helices.add(l);
                    allSingleStrandsLocation.remove(l);
                }
            }
            int i=1;
            for (Location _l:helices) {
                structuralFeatureFactory.createHelix("H" + (i ++), this.secondaryStructures.get(this.secondaryStructures.size()-1), m, _l);
                //System.out.println("create helix "+_l);
            }
            for (Location _l:tertiaryInteractions)
                structuralFeatureFactory.createBaseBaseInteraction(BaseBaseInteraction.EDGE_WATSON_CRICK, BaseBaseInteraction.EDGE_WATSON_CRICK, BaseBaseInteraction.ORIENTATION_CIS, BaseBaseInteraction.RESIDUE_1, this.secondaryStructures.get(this.secondaryStructures.size()-1), m, _l);
            int[] boundaries = allSingleStrandsLocation.getBoundaries();
            for (i = 0; i < boundaries.length; i += 2) {
                structuralFeatureFactory.createSingleStrand("SS" + (i + 1), this.secondaryStructures.get(this.secondaryStructures.size()-1), m, new Location(boundaries[i], boundaries[i + 1]));
            }
        }
        else
            throw new FileParsingException("One sequence has not the same size than its brackets");
    }

}
