package fr.unistra.ibmc.paradise.services.analysis.ibmc.nahelix;

import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.analysis.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.core.io.PDBFileIO;
import fr.unistra.ibmc.paradise.core.io.FileParsingException;
import fr.unistra.ibmc.paradise.utils.*;
import fr.unistra.ibmc.paradise.utils.matrix.Matrix;
import fr.unistra.ibmc.paradise.utils.matrix.StackingMatrix;
import fr.unistra.ibmc.paradise.tools.component.DefaultProgressMonitor;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.util.*;

public class IBMCNahelixAnalysis extends AbstractAnalysis {

    private int offset = 0;
    private TertiaryStructure ts;

    protected IBMCNahelixAnalysis(MoleculeFactory factory) {
        super("Nahelix (Local Service)", factory);
        this.addCategory(Analysis.TERTIARY_STRUCTURE_PREDICTION);
        this.addParadiseFeatureFactory(this.createStructuralFeatureFactory());
    }

    public void run(List<Molecule> molecules, Parameters parameters) throws AnalysisException, NoSolutionException {
        this.offset = 0;
        this.ts = null;
        try {
            Set<Helix> helices = new HashSet<Helix>();
            Set<SingleStrand> sstrands = new HashSet<SingleStrand>();
            Set<BaseBaseInteraction> interactions = new HashSet<BaseBaseInteraction>();
            for (Molecule m:molecules) {
                helices.addAll(m.getSelectedFeatures(Helix.class));
                sstrands.addAll(m.getSelectedFeatures(SingleStrand.class));
                interactions.addAll(m.getSelectedFeatures(BaseBaseInteraction.class));
            }
            //for each helix, we calculate the 3D fold and it is added to the current Tertiary Structure
            for (Helix h:helices)
                this.calculate3DFold(h, this.offset++);
            //for each sstrand, we calculate the 3D fold and it is added to the current Tertiary Structure            
            for (SingleStrand ss:sstrands)
                this.calculate3DFold(ss, this.offset++);
            for (BaseBaseInteraction interaction:interactions)
                this.calculate3DFold(interaction, this.offset++);
            //the single-strands are stacked on their neighbor helix
            for (SingleStrand strand : sstrands) {
                Molecule m = strand.getMolecules().iterator().next();
                Location l =  strand.getFullLocation(m);
                int start = l.getStart();
                int end = l.getEnd();
                if (start == 1) //if the strand starts in 1, there are no preceding helix
                {
                    if (end != m.getLength())//if there is a following helix
                    {
                        Residue3D r = getResidue3D(end + 1, m, this.ts);
                        //stack before
                        if (r != null)
                            this.stackStrandBeforeResidue(strand, r, m, this.ts);
                    }
                } else {
                    Residue3D r = getResidue3D(start - 1, m, this.ts);
                    //stack after
                    if (r != null)
                        this.stackStrandAfterResidue(strand, r, m, this.ts);
                }
            }
            for (Molecule m:molecules) {
                m.removeAllSelectedFeatures();
                m.addSelectedFeature(this.ts);
            }
        }
        catch (Exception e) {
            throw new AnalysisException(e);
        }
    }


    public void calculate3DFold(ParadiseFeature feature, int offset) throws NahelixException, InterruptedException {
        int step = 25 * offset;
        File tmpPdb = null;
        try {
            tmpPdb = IOUtils.createTemporaryFile("nahelix");
        } catch (IOException e) {
            throw new NahelixException(e);
        }
        String seq1;
        String seq2;
        int start1;
        int start2;
        PDBFileIO pdbParser;
        TertiaryStructure structuralDomain3D;
        StructuralFeatureFactory factory = this.getParadiseFeatureFactory(StructuralFeatureFactory.class);

        if (feature instanceof Helix) {
            Helix h = (Helix) feature;
            if (h.getMolecules().size() == 2) {
                Iterator<Molecule> it = h.getMolecules().iterator();
                Molecule m1 = it.next(), m2 = it.next();
                Location l1 = h.getFullLocation(m1), l2 = h.getFullLocation(m2);
                seq1 = m1.printSequence(l1);
                seq2 = GeneralUtils.reverse(m2.printSequence(l2));

                NahelixAlgorithm nahelixAlgorithm = new NahelixAlgorithm();
                nahelixAlgorithm.getRNAasPDBFile(l1.getStart(), l2.getStart(), seq1, seq2, tmpPdb, step);
                //the PDB file generated by the NaHelix algorithm is parsed
                pdbParser = new PDBFileIO(this.getMoleculeFactory());
                try {
                    pdbParser.parseFile(tmpPdb, new DefaultProgressMonitor());
                } catch (FileParsingException e) {
                    throw new NahelixException(e);
                }
                structuralDomain3D = pdbParser.getTertiaryStructures().get(0);
                int residueParsed = 0;
                //the residues3D generated for the current Structural Domain are gathered in the central TertiaryStructure
                int start = l1.getStart(), end = l1.getEnd();
                for (int i = start; i <= end; i++) {
                    Residue3D residue = structuralDomain3D.getResidues3D().get(residueParsed++);
                    try {
                        if (this.ts == null)
                            this.ts = this.getParadiseFeatureFactory(StructuralFeatureFactory.class).createTertiaryStructure(new ParadiseID(), "Tertiary Structure", Source.getAnalysisAsSource(this.description), m1, new Location(i));
                        factory.createResidue3D(this.ts, residue, m1, i);
                    } catch (BiologicalSymbolException e) {
                        throw new NahelixException(e);
                    }
                }
                start = l2.getStart();
                end = l2.getEnd();
                for (int i = start; i <= end; i++) {
                    Residue3D residue = structuralDomain3D.getResidues3D().get(residueParsed++);
                    try {
                        if (this.ts == null)
                            this.ts = this.getParadiseFeatureFactory(StructuralFeatureFactory.class).createTertiaryStructure(new ParadiseID(), "Tertiary Structure", Source.getAnalysisAsSource(this.description), m2, new Location(i));
                        factory.createResidue3D(this.ts, residue, m2, i);
                    } catch (BiologicalSymbolException e) {
                        throw new NahelixException(e);
                    }
                }
            }
            else {
                Molecule m = h.getMolecules().iterator().next();
                Location l = h.getFullLocation(m);
                seq1 = m.printSequence(new Location(l.getStart(), l.getStart()+l.getLength()/2-1));
                seq2 = GeneralUtils.reverse(m.printSequence(new Location(l.getEnd()-l.getLength()/2+1, l.getEnd())));

                NahelixAlgorithm nahelixAlgorithm = new NahelixAlgorithm();
                nahelixAlgorithm.getRNAasPDBFile(l.getStart(), l.getEnd()-l.getLength()/2+1, seq1, seq2, tmpPdb, step);
                //the PDB file generated by the NaHelix algorithm is parsed
                pdbParser = new PDBFileIO(this.getMoleculeFactory());
                try {
                    pdbParser.parseFile(tmpPdb, new DefaultProgressMonitor());
                } catch (FileParsingException e) {
                    throw new NahelixException(e);
                }
                structuralDomain3D = pdbParser.getTertiaryStructures().get(0);
                //the residues3D generated for the current Structural Domain are gathered in the central TertiaryStructure
                int residueParsed = 0;
                int start = l.getStart(), end = l.getStart()+l.getLength()/2-1;
                for (int i = start; i <= end; i++) {
                    Residue3D residue = structuralDomain3D.getResidues3D().get(residueParsed++);
                    try {
                        if (this.ts == null)
                            this.ts = this.getParadiseFeatureFactory(StructuralFeatureFactory.class).createTertiaryStructure(new ParadiseID(), "Tertiary Structure", Source.getAnalysisAsSource(this.description), m, new Location(i));
                        factory.createResidue3D(this.ts, residue, m, i);
                    } catch (BiologicalSymbolException e) {
                        throw new NahelixException(e);
                    }
                }
                start = l.getEnd()-l.getLength()/2+1;
                end = l.getEnd();
                for (int i = start; i <= end; i++) {
                    Residue3D residue = structuralDomain3D.getResidues3D().get(residueParsed++);
                    try {
                        if (this.ts == null)
                            this.ts = this.getParadiseFeatureFactory(StructuralFeatureFactory.class).createTertiaryStructure(new ParadiseID(), "Tertiary Structure", Source.getAnalysisAsSource(this.description), m, new Location(i));
                        factory.createResidue3D(this.ts, residue, m, i);
                    } catch (BiologicalSymbolException e) {
                        throw new NahelixException(e);
                    }
                }
            }

        } else if (feature instanceof SingleStrand) {
            SingleStrand ss = (SingleStrand)feature;
            Molecule m = ss.getMolecules().iterator().next();
            int start = feature.getFullLocation(m).getStart();
            int end = feature.getFullLocation(m).getEnd();
            seq1 = m.printSequence(feature.getFullLocation(m));
            seq2 = "";
            for (int i = 0; i < seq1.length(); i++)
                seq2 += 'X';
            start1 = start;
            start2 = end + 1;
            NahelixAlgorithm nahelixAlgorithm = new NahelixAlgorithm();
            nahelixAlgorithm.getRNAasPDBFile(start1, start2, seq1, seq2, tmpPdb, step);
            //the PDB file generated by the NaHelix algorithm is parsed
            pdbParser = new PDBFileIO(this.getMoleculeFactory());
            try {
                pdbParser.parseFile(tmpPdb, new DefaultProgressMonitor());
            } catch (FileParsingException e) {
                throw new NahelixException(e);
            }
            structuralDomain3D = pdbParser.getTertiaryStructures().get(0);
            int residueParsed = 0;
            //the residues3D generated for the current Structural Domain are gathered in the central TertiaryStructure
            for (int i = start; i <= end; i++) {
                Residue3D residue = structuralDomain3D.getResidues3D().get(residueParsed++);
                try {
                    if (this.ts == null)
                        this.ts = this.getParadiseFeatureFactory(StructuralFeatureFactory.class).createTertiaryStructure(new ParadiseID(), "Tertiary Structure", Source.getAnalysisAsSource(this.description), m, new Location(i));
                    factory.createResidue3D(this.ts, residue, m, i);
                } catch (BiologicalSymbolException e) {
                    throw new NahelixException(e);
                }
            }
        }  else if (feature instanceof BaseBaseInteraction) {
            /*
            BaseBaseInteraction interaction = (BaseBaseInteraction) feature;
            File basePair = null;
            try {
                basePair = new File(Paradise.getTertiaryData("basepairs/CHWC.pdb"));                
                PDBFileIO parser = new PDBFileIO(this.moleculeFactory);
                parser.parseFile(basePair,new DefaultProgressMonitor());
                for (Molecule m:parser.getTertiaryStructures().get(0).getMolecules()) {
                    if ("1".equals(m.getName())) {
                        Residue r = interaction.getResidue();
                        if (this.ts == null)
                            this.ts = this.getParadiseFeatureFactory(StructuralFeatureFactory.class).createTertiaryStructure(new ParadiseID(), "Tertiary Structure", Source.getAnalysisAsSource(this.description), r.getMolecule(), new Location(r.getAbsolutePosition()));
                        Residue3D r3D = (Residue3D)m.getAnnotations(Residue3D.class).iterator().next().getFeature();
                        r3D.getFactory().createResidue3D(this.ts,r3D,r.getMolecule(), r.getAbsolutePosition());
                    }
                    else if ("2".equals(m.getName())) {
                        Residue r = interaction.getPartnerResidue();
                        if (this.ts == null)
                            this.ts = this.getParadiseFeatureFactory(StructuralFeatureFactory.class).createTertiaryStructure(new ParadiseID(), "Tertiary Structure", Source.getAnalysisAsSource(this.description), r.getMolecule(), new Location(r.getAbsolutePosition()));
                        Residue3D r3D = (Residue3D)m.getAnnotations(Residue3D.class).iterator().next().getFeature();
                        r3D.getFactory().createResidue3D(this.ts,r3D,r.getMolecule(), r.getAbsolutePosition());
                    }
                }
            } catch (FileParsingException e) {
                throw new NahelixException(e);
            } catch (BiologicalSymbolException e) {
                throw new NahelixException(e);
            }*/
        }

    }

    private final int BEFORE = 1;
    private final int AFTER = 2;

    Matrix getTransformationMatrixStack(Residue3D ref, Residue3D other, int direction) {
        Residue3D.Atom oC2;
        Residue3D.Atom oC3;
        Residue3D.Atom oC4;
        Residue3D.Atom oO3;
        Residue3D.Atom rC2;
        Residue3D.Atom rC3;
        Residue3D.Atom rC4;
        Residue3D.Atom rO3;
        oC2 = other.getAtom("C2'");
        oC3 = other.getAtom("C3'");
        oC4 = other.getAtom("C4'");
        oO3 = other.getAtom("O3'");
        rC2 = ref.getAtom("C2'");
        rC3 = ref.getAtom("C3'");
        rC4 = ref.getAtom("C4'");
        rO3 = ref.getAtom("O3'");
        Matrix O;
        Matrix C;
        if (oC2.hasCoordinatesFilled() && oC3.hasCoordinatesFilled() && oC4.hasCoordinatesFilled() && oC3.hasCoordinatesFilled()) {
            float[] oC2O3 = {oC2.getX(), oC2.getY(), oC2.getZ()};
            float[] oC3O3 = {oC3.getX(), oC3.getY(), oC3.getZ()};
            float[] oC4O3 = {oC4.getX(), oC4.getY(), oC4.getZ()};
            float[] oO3O3 = {oO3.getX(), oO3.getY(), oO3.getZ()};
            O = StackingMatrix.getResidueToReferenceMatrix(oC2O3, oC3O3, oC4O3, oO3O3);
        } else
            return null;
        if (rC2.hasCoordinatesFilled() && rC3.hasCoordinatesFilled() && rC4.hasCoordinatesFilled() && rC3.hasCoordinatesFilled()) {
            float[] rC2O3 = {rC2.getX(), rC2.getY(), rC2.getZ()};
            float[] rC3O3 = {rC3.getX(), rC3.getY(), rC3.getZ()};
            float[] rC4O3 = {rC4.getX(), rC4.getY(), rC4.getZ()};
            float[] rO3O3 = {rO3.getX(), rO3.getY(), rO3.getZ()};
            C = StackingMatrix.getReferenceToResidueMatrix(rC2O3, rC3O3, rC4O3, rO3O3);
        } else
            return null;

        if (O != null && C != null) {
            if (direction == BEFORE) {
                Matrix P = StackingMatrix.getToPreviousMatrix();
                return C.times(P.times(O));
                //warning !!
                //this order is very important because P/N are static and relative to a newly create residue
                //if P/N is applied after C, then the stack relative to the 2nd stand of an helix will be opposite
            }

            if (direction == AFTER) {
                Matrix N = StackingMatrix.getToNextMatrix();
                return C.times(N.times(O));
            }
        }
        return null;
    }

    private void applyTransformation(Residue3D.Atom atom, Matrix matrix) {
        float[] c = StackingMatrix.applyMatrixToPoint(matrix, atom.getX(), atom.getY(), atom.getZ());
        atom.setCoordinates(c[0], c[1], c[2]);
    }

    private void applyTransformation(Residue3D residue, Matrix matrix) {
        for (Residue3D.Atom atom : residue.getAtoms()) {
            if (atom.hasCoordinatesFilled()) {
                applyTransformation(atom, matrix);
            }
        }
    }

    private void stackStrandBeforeResidue(SingleStrand strand, Residue3D refResidue, Molecule m, TertiaryStructure ts) {
        int start = strand.getFullLocation(m).getStart();
        int end = strand.getFullLocation(m).getEnd();
        Residue3D r = getResidue3D(end, m, ts);
        if (r!=null)  {
            Matrix matrix = getTransformationMatrixStack(refResidue, r, BEFORE);
            for (int i = start; i <= end; i++)
                applyTransformation(getResidue3D(i, m, ts), matrix);
        }
    }

    private void stackStrandAfterResidue(SingleStrand strand, Residue3D refResidue, Molecule m, TertiaryStructure ts) {
        int start = strand.getFullLocation(m).getStart();
        int end = strand.getFullLocation(m).getEnd();
        Residue3D r = getResidue3D(start, m, ts);
        if (r!=null)  {
            Matrix matrix = getTransformationMatrixStack(refResidue, r, AFTER);
            for (int i = start; i <= end; i++)
                applyTransformation(getResidue3D(i, m, ts), matrix);
        }
    }

    private Residue3D getResidue3D(int i, Molecule m, TertiaryStructure ts) {
        for (Residue3D r : ts.getResidues3D())
            if (r.getFullLocation(m).getStart() == i)
                return r;
        return null;
    }

    private static class NahelixAlgorithm {

        private static final int P = 1;
        private static final int O5 = 3;


        private int algoStr1;
        private int algoStr2;
        private int algoStr3;
        private String algoSeq1;
        private String algoSeq2;
        private String algoSeq3;
        private int algoLength;
        private File algoOutputFile;

        private static double con = Math.PI / 180;

        /**
         * Prints the result of the construction in a .hd file format
         *
         * @param x    float[] - list of x
         * @param y    float[] - list of y
         * @param z    float[] - list of z
         * @param name String[] - list of atom names
         * @param seq  char[] - list of base names
         * @param rnum int[] - list of residue number
         */
        private void save(float[] x, float[] y, float[] z, String[] name, char seq[], int rnum[], float offsetX, float offsetY, float offsetZ) throws ParsingException {
            try {
                PrintWriter out = new PrintWriter(new FileWriter(algoOutputFile));
                char chain = 'A';
                char oldChain = chain;
                int i = 0;
                int offset = 0;

                while (name[i] != null) {
                    if (rnum[i] > algoSeq1.length() + algoStr1 - 1) {
                        //             Printing.debug("rnum > algoLength "+rnum[i]+" > "+Nahelix.algoSeq1.length());
                        chain = 'B';
                    }
                    if (oldChain != chain) {
                        String ligne = PDB.getTerString(i + 1 + offset, seq[i - 1] + "", oldChain, rnum[i - 1]);
                        out.println(ligne);
                        offset++;
                    }
                    /*(int nb, String name, String type, char chId, int num, float x, float y, float z)*/
                    if (seq[i] != 'X') {
                        String ligne = PDB.getAtomString(i + 1 + offset, name[i], seq[i] + "", chain, rnum[i], x[i] + offsetX, y[i] + offsetY, z[i] + offsetZ);
                        out.println(ligne);
                        //     System.out.println(ligne);

                    }
                    i++;
                    oldChain = chain;
                }
                String ligne = PDB.getTerString(i + 1 + offset, seq[i - 1] + "", oldChain, rnum[i - 1]);
                out.println(ligne);
                out.close();
            }
            catch (IOException e) {
                System.err.println("In Nahelix.print() : can't write to file : " + algoOutputFile);
                e.printStackTrace();
            }
        }

        private void getRNAasPDBFile(int str1, int str2, String seq1, String seq2, File outputFile, float offsetX) throws NahelixException {
            algoOutputFile = outputFile;
            HELIXaBr(str1, str2, seq1, seq2, P, offsetX, 0, 0);
        }

        /**
         * Builds a double helix of TYPE_ADNA, TYPE_BDNA or TYPE_RNA
         *
         * @param str1      int - index of the first residue of the first strand
         * @param str2      int - index of the first residue of the second strand
         * @param seq1      String - sequence of the first strand
         * @param seq2      String - sequence of the second strand
         * @param endType   int - end type (P, O3P or O5')
         */
        private void HELIXaBr(int str1, int str2, String seq1, String seq2, int endType, float offsetX, float offsetY, float offsetZ) throws NahelixException {
            algoSeq1 = seq1;
            algoSeq2 = seq2;
            algoStr1 = str1;
            algoStr2 = str2;
            int size = (seq1.length()+seq2.length())*24;
            float x[] = new float[size];
            float y[] = new float[size];
            float z[] = new float[size];
            String l3f[] = new String[size];
            char seq[] = new char[size];
            int rnum[] = new int[size];

            int jc = 0, jd, jci, jdi;
            jci = 0;
            jdi = 2;

            if (endType == NahelixAlgorithm.P) {
                jci = 1;
            }
            if (endType == NahelixAlgorithm.O5) {
                jci = 4;
                jdi = 5;
            }

            float omega = 0;

            float omeg = 32.7f, rh = 2.81f;
            double[] theta = new double[]{72.8, 70.5, 71.2, 76.6, 60.1, 55.8, 46.9, 41.4, 45.3, 30.2, 33.2, 36.9};
            double[] r = new double[]{9.54, 8.71, 9.59, 7.50, 8.49, 9.75, 9.68, 8.86, 9.15, 10.15, 8.82, 8.55};
            double[] zi = new double[]{-2.44, -3.75, -4.94, -3.80, -3.48, -3.35, -3.10, -4.02, -1.77, -3.14, -3.17, -1.77};
            String[] l3r = new String[]{"O3' ", "P   ", "O1P ", "O2P ", "O5' ", "C5' ", "C4' ", "C3' ", "O4' ", "O2' ", "C2' ", "C1' "};

            seq1 = seq1.toUpperCase();
            seq2 = seq2.toUpperCase();

            int coe = -1;
            int nfi1 = 0;

            for (int k = 0; k < seq1.length(); k++) {
                if (seq1.charAt(k) != 'X') {
                    break;
                }
                nfi1++;
            }

            double arg = coe * (theta[0] - omega - omeg * nfi1) * con;
            x[0] = (float) (r[0] * Math.cos(arg));
            y[0] = (float) (r[0] * Math.sin(arg));
            z[0] = (float) (coe * (zi[0] - rh * nfi1));
            l3f[0] = "O3P ";
            seq[0] = seq1.charAt(nfi1);
            rnum[0] = str1 + nfi1;

            char base;
            for (int i = 1; i <= seq1.length(); i++) {
                base = seq1.charAt(i - 1);
                if (i == 1) {
                    jd = jdi;
                    jc = jc - jci;
                } else {
                    jd = 2;
                }

                if (base != 'X') {
                    for (int j1 = jd - 1; j1 < r.length; j1++) {
                        arg = coe * (theta[j1] - omega) * con;
                        x[jc + j1] = (float) (r[j1] * Math.cos(arg));
                        y[jc + j1] = (float) (r[j1] * Math.sin(arg));
                        z[jc + j1] = (float) (coe * (zi[j1] - rh * (i - 1)));
                        l3f[jc + j1] = l3r[j1];
                        seq[jc + j1] = base;
                        rnum[jc + j1] = str1 + i - 1;
                    }
                    jc += r.length;

                    Atoms at = new Atoms();

                    switch (base) {
                        case 'A':
                            at = AA(coe, i - 1, omega, rh);
                            break;
                        case 'C':
                            at = CA(coe, i - 1, omega, rh);
                            break;
                        case 'G':
                            at = GA(coe, i - 1, omega, rh);
                            break;
                        case 'U':
                            at = UA(coe, i - 1, omega, rh);
                            break;
                        case 'T':
                            at = TA(coe, i - 1, omega, rh);
                            break;
                        default:
                            Printing.warning("Nucleotide unknow : " + base);
                    }

                    for (int j = 0; j < at.nb; j++) {
                        x[jc + j] = at.x[j];
                        y[jc + j] = at.y[j];
                        z[jc + j] = at.z[j];
                        l3f[jc + j] = at.name[j];
                        seq[jc + j] = base;
                        rnum[jc + j] = str1 + i - 1;
                    }

                    jc += at.nb;
                }

                omega += omeg;

                if (base != 'X') {
                    arg = coe * (theta[0] - omega) * con;
                    x[jc] = (float) (r[0] * Math.cos(arg));
                    y[jc] = (float) (r[0] * Math.sin(arg));
                    z[jc] = (float) (coe * (zi[0] - rh * i));
                    l3f[jc] = l3r[0];
                    seq[jc] = base;
                    rnum[jc] = str1 + i - 1;
                }

            }
            coe = 1;
            int nfi2 = 0;
            omega = omega - omeg;
            //int ij = 0;

            for (int k = 0; k < seq2.length(); k++) {
                if (seq2.charAt(k) != 'X') {
                    break;
                }
                nfi2++;
            }

            arg = coe * (theta[0] + omega - omeg * nfi2) * con;
            x[jc + 1] = (float) (r[0] * Math.cos(arg));
            y[jc + 1] = (float) (r[0] * Math.sin(arg));
            z[jc + 1] = (float) (coe * (zi[0] + rh * (seq2.length() - 1 - nfi2)));
            l3f[jc + 1] = "O3P ";
            seq[jc + 1] = seq2.charAt(seq2.length() - 1);
            rnum[jc + 1] = str2;
            jc++;

            for (int i = seq2.length(); i >= 1; i--) {
                base = seq2.charAt(i - 1);
                if (i == seq2.length()) {
                    jd = jdi;
                    jc = jc - jci;
                } else {
                    jd = 2;
                }
                if (base != 'X') {
                    for (int j1 = jd - 1; j1 < r.length; j1++) {
                        arg = coe * (theta[j1] + omega) * con;
                        x[jc + j1] = (float) (r[j1] * Math.cos(arg));
                        y[jc + j1] = (float) (r[j1] * Math.sin(arg));
                        z[jc + j1] = (float) (coe * (zi[j1] + rh * (i - 1)));
                        l3f[jc + j1] = l3r[j1];
                        seq[jc + j1] = base;
                        rnum[jc + j1] = str2 + seq2.length() - i;
                    }
                    jc += r.length;

                    Atoms at = new Atoms();

                    switch (base) {
                        case 'A':
                            at = AA(coe, i - 1, omega, rh);
                            break;
                        case 'C':
                            at = CA(coe, i - 1, omega, rh);
                            break;
                        case 'G':
                            at = GA(coe, i - 1, omega, rh);
                            break;
                        case 'U':
                            at = UA(coe, i - 1, omega, rh);
                            break;
                        case 'T':
                            at = TA(coe, i - 1, omega, rh);
                            break;
                    }

                    for (int j = 0; j < at.nb; j++) {
                        x[jc + j] = at.x[j];
                        y[jc + j] = at.y[j];
                        z[jc + j] = at.z[j];
                        l3f[jc + j] = at.name[j];
                        seq[jc + j] = base;
                        rnum[jc + j] = str2 + seq2.length() - i;
                    }
                    jc += at.nb;

                }

                omega = omega - omeg;

                if (base != 'X') {
                    arg = coe * (theta[0] + omega) * con;
                    x[jc] = (float) (r[0] * Math.cos(arg));
                    y[jc] = (float) (r[0] * Math.sin(arg));
                    z[jc] = (float) (coe * (zi[0] + rh * (i - 2)));
                    l3f[jc] = l3r[0];
                    seq[jc] = base;
                    rnum[jc] = str2 + seq2.length() - i;
                }
            }
            try {
                save(x, y, z, l3f, seq, rnum, offsetX, offsetY, offsetZ);
            }
            catch (ParsingException e) {
                throw new NahelixException(e);
            }

        }

        private Atoms AA(int coe, int i, float omega, float rh) throws NahelixException {
            double[] theta = {37.2, 47.0, 46.2, 30.9, 15.6, 17.4, 5.5, 8.7, 18.0, 27.6};
            double[] r = {7.12, 6.39, 5.15, 5.14, 3.08, 4.39, 5.28, 6.54, 7.15, 6.44};
            double[] zi = {-1.39, -1.33, -.95, -.75, -.05, -.34, -.24, -.54, -.93, -1.01};
            String[] name = {"N9  ", "C8  ", "N7  ", "C5  ", "N6  ", "C6  ", "N1  ", "C2  ", "N3  ", "C4  "};
            return new Atoms(theta, r, zi, name, coe, i, omega, rh);
        }

        private Atoms CA(int coe, int i, float omega, float rh) throws NahelixException {
            double[] theta = {37.2, 47.6, 50.6, 36.1, 36.8, 24.4, 19.9, 26.8};
            double[] r = {7.12, 6.53, 5.26, 3.23, 4.50, 5.32, 7.51, 6.61};
            double[] zi = {-1.39, -1.38, -1.03, -.31, -.67, -.68, -1.06, -1.04};
            String[] name = {"N1  ", "C6  ", "C5  ", "N4  ", "C4  ", "N3  ", "O2  ", "C2  "};
            return new Atoms(theta, r, zi, name, coe, i, omega, rh);
        }

        private Atoms GA(int coe, int i, float omega, float rh) throws NahelixException {
            double[] theta = {37.2, 47., 46., 30.5, 13.9, 17.1, 5.5, .7, 8.5, 18., 27.5};
            double[] r = {7.12, 6.36, 5.11, 5.13, 3.16, 4.34, 5.35, 7.58, 6.67, 7.20, 6.44};
            double[] zi = {-1.39, -1.32, -.94, -.74, -.04, -.33, -.26, -.42, -.55, -.94, -1.01};
            String[] name = {"N9  ", "C8  ", "N7  ", "C5  ", "O6  ", "C6  ", "N1  ", "N2  ", "C2  ", "N3  ", "C4  "};
            return new Atoms(theta, r, zi, name, coe, i, omega, rh);
        }

        private Atoms TA(int coe, int i, float omega, float rh) throws NahelixException {
            double[] theta = {37.2, 47.7, 50.8, 37.1, 37.8, 25.2, 19.6, 26.9, 67.7};
            double[] r = {7.12, 6.54, 5.28, 3.24, 4.42, 5.31, 7.46, 6.63, 5.19};
            double[] zi = {-1.39, -1.38, -1.04, -.32, -.66, -.69, -1.04, -1.04, -1.08};
            String[] name = {"N1  ", "C6  ", "C5  ", "O4  ", "C4  ", "N3  ", "O2  ", "C2  ", "M5  "};
            return new Atoms(theta, r, zi, name, coe, i, omega, rh);
        }

        private Atoms UA(int coe, int i, float omega, float rh) throws NahelixException {
            double[] theta = {37.2, 47.7, 50.8, 37.1, 37.8, 25.2, 19.6, 26.9};
            double[] r = {7.12, 6.54, 5.28, 3.24, 4.42, 5.31, 7.46, 6.63};
            double[] zi = {-1.39, -1.38, -1.04, -.32, -.66, -.69, -1.04, -1.04};
            String[] name = {"N1  ", "C6  ", "C5  ", "O4  ", "C4  ", "N3  ", "O2  ", "C2  "};
            return new Atoms(theta, r, zi, name, coe, i, omega, rh);
        }

        private class Atoms {
            private float[] x;
            private float[] y;
            private float[] z;
            private String[] name;
            private int nb = 0;

            private Atoms() {
            }

            /**
             * Builds an atom set, with cardinal coordinates
             *
             * @param theta double[] - theta angle of each atom
             * @param r     double[] - radius of each atom
             * @param zi    double[] - cylindric offset of each atom
             * @param name2 String[] - name of each atom
             * @param coe   int - a coefficient
             * @param i     int - position of the base in the sequence
             * @param omega float - rotation step
             * @param rh    float - vertical offset
             */
            private Atoms(double[] theta, double[] r, double[] zi, String[] name2, int coe, int i, float omega, float rh) throws NahelixException {
                if ((theta.length == r.length) && (zi.length == r.length) && (name2.length == r.length)) {
                    x = new float[theta.length];
                    y = new float[theta.length];
                    z = new float[theta.length];
                    name = new String[theta.length];
                    for (int w = 0; w < theta.length; w++) {
                        double arg = coe * (theta[w] + coe * omega) * NahelixAlgorithm.con;
                        x[w] = (float) (r[w] * Math.cos(arg));
                        y[w] = (float) (r[w] * Math.sin(arg));
                        z[w] = (float) (coe * (zi[w] + coe * rh * i));
                        name[w] = name2[w];
                        nb = x.length;
                    }
                } else {
                    throw new NahelixException("Wrong number of elements in one of the arrays");
                }
            }
        }

    }
}