package fr.unistra.ibmc.paradise.services.analysis.ibmc.rnart;

import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.MoleculeFactory;
import fr.unistra.ibmc.paradise.core.utils.Residue;
import fr.unistra.ibmc.paradise.core.analysis.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.core.io.PDBFileIO;
import fr.unistra.ibmc.paradise.utils.HD;
import fr.unistra.ibmc.paradise.utils.NumberFormat;
import fr.unistra.ibmc.paradise.tools.component.DefaultProgressMonitor;
import fr.unistra.ibmc.paradise.Paradise;

import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.*;

public class IBMCRnartAnalysis extends AbstractAnalysis {

    protected IBMCRnartAnalysis(MoleculeFactory factory) {
        super("Rnart (Local Service)", factory);
        this.addCategory(Analysis.TERTIARY_STRUCTURE_REFINEMENT);
        this.addParadiseFeatureFactory(this.createStructuralFeatureFactory());
    }

    private List<Molecule> molecules;
    private Map<Integer, RiboNucleotide3D> all3DNucleotides;
    private SecondaryStructure ss = null;
    private TertiaryStructure ts = null;

    public void run(List<Molecule> molecules, Parameters parameters) throws AnalysisException, NoSolutionException {
        this.molecules = molecules;
        try {
            File rnartDir =  new File(Paradise.getTmpDirectory(),"rnart"+System.nanoTime());
            rnartDir.mkdir();
            File    globalPDB = new File(rnartDir,"GLOBAL.PDB"),
                    atomFile = new File(rnartDir,"ATOMS.DAT"),
                    nuclinFile = new File(rnartDir,"NUCLIN.DAT"),
                    hbndFile = new File(rnartDir,"HBND.DAT");

            String mode = null;
            String threshold = null;

            if (parameters != null) {
                for (Parameter param : parameters.getParameters()) {
                    if (param.getName().equals("mode"))
                        mode = param.getValue();

                    if (param.getName().equals("threshold"))
                        threshold = param.getValue();
                }
            }

            if (mode == null)
                throw new AnalysisException("The mode parameter is null");
            if (threshold == null)
                throw new AnalysisException("The threshold parameter is null");

            if (molecules.size() > 0) {
                ts = molecules.get(0).getSelectedFeatures(TertiaryStructure.class).get(0);
                ss = molecules.get(0).getSelectedFeatures(SecondaryStructure.class).get(0);

                this.all3DNucleotides = sortNucleotides();
                List<String> _3dMolecules = new ArrayList<String>(); // will contains the molecules for which we have Residue3D objects

                //WRITE GLOBAL.PDB FILE                
                this.status = "Writing GLOBAL.PDB";
                List<Residue3D> residues = new ArrayList<Residue3D>(this.all3DNucleotides.size());
                Molecule currentMolecule = null;
                int currentPosition = -1;
                StringBuffer currentSequence = null;
                for (int i = 1 ; i<= this.all3DNucleotides.size();i++) { //in this loop we fill the sorted list of Residue3D objects to create the GLOBAL.PDB file and we use it to create the sequence strings corresponding to these residues.
                    Residue3D r = this.all3DNucleotides.get(i);
                    residues.add(r);
                    Molecule _m = r.getMolecules().iterator().next();
                    int position = r.getAnnotations().iterator().next().getLocation().getStart();
                    if (!_m.equals(currentMolecule) || currentPosition != position -1) {
                        if (currentSequence != null)
                            _3dMolecules.add(currentSequence.toString());
                        currentSequence = new StringBuffer();
                    }
                    currentMolecule = _m;
                    currentPosition = position;
                    currentSequence.append(r.getName().charAt(0));    
                }
                //the last molecule
                if (currentSequence != null)
                    _3dMolecules.add(currentSequence.toString());
                PDBFileIO.writePDBFile(residues, true, globalPDB);

                //WRITE NUCLIN.DAT FILE
                this.status = "Writing NUCLIN.DAT";
                String nuclindat = "";

                nuclindat += "Rnart Service\n\n";
                nuclindat += mergeMolecules(_3dMolecules) + "\n";

                nuclindat += "    3\n";
                String tmp = "";

                //here we don't add the basepairs, this will be taken care off via the AtomAtomInteraction in HBND.DAT

                nuclindat += "\n" + "MODEL\n" + "    0    2    0    0    0\n" + "    4   " + all3DNucleotides.size() + "\n";

                List<Integer> sugars = getSugars();

                for (int i = 0; i < sugars.size(); i++) {
                    tmp += " " + sugars.get(i);
                    if ((i % 35) == 34) {
                        nuclindat += tmp + "\n";
                        tmp = "";
                    }
                }
                if (!tmp.equals(""))
                    nuclindat += tmp + "\n";

                PrintWriter out = new PrintWriter(new FileWriter(nuclinFile));
                out.print(nuclindat);
                out.flush();
                out.close();

                //WRITE HBND.DAT FILE
                this.status = "Writing HBND.DAT";

                String hbnddat = "";
                List<String> hbndLines = getHbndLines();

                for (String hbndline : hbndLines)
                    hbnddat += hbndline + "\n";

                out = new PrintWriter(new FileWriter(hbndFile));
                out.print(hbnddat);
                out.flush();
                out.close();

                //WRITE ATOMS.DAT FILE
                this.status = "Writing ATOMS.DAT";
                HD.convertPDBasHD(atomFile, globalPDB);

                this.status = "Calling NUCMULT";
                NucMult nm = new NucMult(this,mode, threshold, rnartDir.getAbsolutePath()+System.getProperty("file.separator"));
                int pass = nm.getNumberOfPass();
                String p = pass + "";
                if (pass < 10)
                    p = "0" + p;
                File f = new File(rnartDir, "ATOMS_" + p + ".hd.pdb");

                PDBFileIO reader = new PDBFileIO(this.getMoleculeFactory());
                reader.parseFile(f, new DefaultProgressMonitor());
                Molecule resultMolecule = reader.getMolecules().get(0); //here i remember that we merge all the molecule in a single one for the RNART job (mergeMolecules() method in this class)

                for (Molecule molecule : molecules)
                    molecule.removeAllSelectedFeatures();

                for (Molecule.Annotation annotation : resultMolecule.getAnnotations(Residue3D.class)) {
                    Residue3D rnartR = (Residue3D) annotation.getFeature();
                    Residue3D localR = all3DNucleotides.get(Integer.parseInt(rnartR.getLabel())); //we're using the unique label created with the sortNucleotides() method in this class to do the mapping
                    if (localR != null) {
                        for (Residue3D.Atom rnartA : rnartR.getAtoms()) {
                            if (rnartA.hasCoordinatesFilled()) {
                                Residue3D.Atom localA = localR.getAtom(rnartA.getName());
                                //System.out.println("rnartR.getName() "+rnartR.getName());
                                //System.out.println("rnartA.getName() "+rnartA.getName());
                                //System.out.println("localR.getName() "+localR.getName());
                                //System.out.println("localA.getName() "+localA.getName());
                                localA.setCoordinates(rnartA.getX(), rnartA.getY(), rnartA.getZ());
                            }
                        }
                        localR.getAnnotations().get(0).getMolecule().addSelectedFeature(localR);
                    } else {
                        System.err.println("Warning : No residue find with label " + rnartR.getLabel());
                    }
                }
            }
        }
        catch (Exception e) {
            throw new AnalysisException(e);
        }
    }




    private String mergeMolecules(List<String> molecules) {

        StringBuffer out = new StringBuffer();
        for (String m : molecules) {
            if (m.length() > 0) {
                out.append(m.charAt(0));
                int line = 1;
                for (int i = 1; i < m.length(); i++) {
                    out.append(m.charAt(i));
                    line++;
                    if (line == 70 && i < m.length() - 1) {
                        line = 0;
                        out.append("-\n");
                    }
                }
                out.append("\n");
            }
        }

        return out.toString();
    }

    List<Integer> order(Set<Integer> set) {
        List<Integer> list = new ArrayList<Integer>(set.size());
        for (int v : set) {
            boolean last = true;
            for (int i : list) {
                if (v < i) {
                    last = false;
                    list.add(list.indexOf(i), v);
                    break;
                }
            }
            if (last)
                list.add(v);
        }
        return list;
    }

    List<Integer> getSugars() throws AnalysisException {
        List<Integer> ret = new ArrayList<Integer>();
        for (Integer key : order(all3DNucleotides.keySet()))
            ret.add(all3DNucleotides.get(key).getSugarPucker());
        return ret;
    }

    List<String> getHbndLines() throws AnalysisException {
        List<String> ret = new ArrayList<String>();

        for (AtomAtomInteraction iaai : ss.getSubFeaturesRecursively(AtomAtomInteraction.class)) {
            double maxDist = 0;
            double minDist = 0;

            switch (iaai.getInteractionType()) {
                case AtomAtomInteraction.INTERACTION_HBOND:
                    minDist = 2.7f;
                    maxDist = 3.1f;
                    break;
                case AtomAtomInteraction.INTERACTION_CH:
                    minDist = 2.8f;
                    maxDist = 4.0f;
                    break;
                case AtomAtomInteraction.INTERACTION_WATER:
                    minDist = 3.3f;
                    maxDist = 5.0f;
                    break;
                case AtomAtomInteraction.INTERACTION_MAGNESIUM:
                    minDist = 1.8f; //???
                    maxDist = 2.2f; //???
                    break;
                case AtomAtomInteraction.INTERACTION_SODIUM:
                    minDist = 3.0f;
                    maxDist = 3.5f;
                    break;
            }


            Iterator<Residue> it = iaai.getResidues().iterator();
            Residue _r1 = it.next(),
                    _r2 = it.next();
            Residue3D   r1 = ts.getResidue3DAt(_r1.getAbsolutePosition(), _r1.getMolecule()),
                    r2 = ts.getResidue3DAt(_r2.getAbsolutePosition(), _r2.getMolecule());

            if (r1 != null && r2 != null) { // can be null since not all the 2D has been generated as 3D

                String fNName = "   " + r1.getName().charAt(0);
                String fNNum = NumberFormat.intFormat(Integer.parseInt(r1.getLabel()), 3);
                String fAName = null;
                if (iaai.getResidue1().getSymbol() == r1.getName().charAt(0))  //test to be sure that we get an atom attached to this residue (important for intermolecular interaction because we don't know from which chain is atom1 or atom2)
                    fAName= iaai.getAtom1();
                else
                    fAName= iaai.getAtom2();
                while (fAName.length() < 4)
                    fAName += " ";

                String sNName = " " + r2.getName().charAt(0);
                String sNNum = NumberFormat.intFormat(Integer.parseInt(r2.getLabel()), 3);
                String sAName = null;                
                if (iaai.getResidue1().getSymbol() == r2.getName().charAt(0)) //test to be sure that we get an atom attached to this residue (important for intermolecular interaction because we don't know from which chain is atom1 or atom2)
                    sAName= iaai.getAtom1();
                else
                    sAName= iaai.getAtom2();
                while (sAName.length() < 4)
                    sAName += " ";

                String min = NumberFormat.doubleFormat(minDist, 6, 1);
                String max = "   " + NumberFormat.doubleFormat(maxDist, 6, 1);

                ret.add(fNName + fNNum + fAName + sNName + sNNum + sAName + min + max);
            }
        }
        return ret;
    }

    /**
     * Return a collection of Residue3D objects.
     * Each object is linked to a unique label.
     * In the returned collection, the objects are ckustered by molecule, and, in each cliusted ordered according to their absolute position in their molecule
     * @return
     * @throws AnalysisException
     */
    private Map<Integer, RiboNucleotide3D> sortNucleotides() throws AnalysisException {
        Map<Integer, RiboNucleotide3D> ret = new HashMap<Integer, RiboNucleotide3D>();
        Map<Molecule, List<RiboNucleotide3D>> lists = new HashMap<Molecule, List<RiboNucleotide3D>>();

        for (Molecule molecule : molecules)
            lists.put(molecule, new ArrayList<RiboNucleotide3D>());

        for (Residue3D res : ts.getResidues3D())
            if (res instanceof RiboNucleotide3D) {
                Molecule.Annotation a = res.getAnnotations().get(0);
                addSorted((RiboNucleotide3D) res, lists.get(a.getMolecule()));
            }

        int id = 1; // to produce an unique label for each Residue3D objects
        for (Molecule molecule : molecules)
            for (RiboNucleotide3D res : lists.get(molecule)) {
                Molecule.Annotation a = res.getAnnotations().get(0);
                ts.addResidueLabel(""+id, a.getLocation().getStart(), a.getMolecule());
                ret.put(id, res);
                id++;
            }

        return ret;
    }

    private void addSorted(RiboNucleotide3D res, List<RiboNucleotide3D> list) throws AnalysisException {
        for (RiboNucleotide3D lr : list)
            if (getNumerotationFor(res) < getNumerotationFor(lr)) {
                list.add(list.indexOf(lr), res);
                return;
            }
        list.add(res);
    }

    private int getNumerotationFor(Residue3D residue) throws AnalysisException {
        int numVal = residue.getAnnotations().get(0).getLocation().getStart();
        if (numVal > 999 || numVal < 1)
            throw new AnalysisException("Sorry, due to limitation of the external algorithm used to perform this services, all residues numbers must be between 1 and 999");
        return numVal;
    }
}
