package fr.unistra.ibmc.paradise.services.analysis.ibmc.rnaview;

import fr.unistra.ibmc.paradise.core.*;
import fr.unistra.ibmc.paradise.core.analysis.*;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.core.io.PDBFileIO;
import fr.unistra.ibmc.paradise.core.io.RNAMLFileIO;
import fr.unistra.ibmc.paradise.utils.IOUtils;
import fr.unistra.ibmc.paradise.tools.component.DefaultProgressMonitor;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;

import org.apache.commons.io.FileUtils;

public class IBMCRnaviewAnalysis extends AbstractAnalysis {

    public IBMCRnaviewAnalysis(MoleculeFactory factory) {
        super("RNAVIEW (ARN/IBMC Web Service)",factory);
        this.addCategory(Analysis.TERTIARY_STRUCTURE_ANNOTATION);
    }

    public void run(List<Molecule> molecules,Parameters parameters) throws AnalysisException, NoSolutionException {
        long start = System.currentTimeMillis();
        try {
            //We store all the TertiaryStructure features selected
            List<Residue3D> residues = new ArrayList<Residue3D>();
            for(Molecule molecule : molecules) {
                if (Thread.interrupted())
                    throw new AnalysisException("The analysis has been cancelled");
                for (ParadiseFeature r:molecule.getSelectedFeatures())
                    residues.add((Residue3D)r);
            }
            //#### trick for RNAVIEW ###################
            //some PDB files have missing atoms for residues. RNAVIEW export in the RNAML file only the residues with all the atoms
            //consequently, RNAVIEW produces in these cases deleted version of the original molecule
            // this is a problem for PARADISE to find the homologuous molecule between the sequence stored in the RNAML output and the original PDB file
            // to avoid that, we set the coordinates of the missing atoms to 0 0 0. With such coordinates, RNAVIEW will detect nothing for them (no interaction) but it is enough to have these residues in the RNAML output
            List<Residue3D.Atom> atomsFilled = new ArrayList<Residue3D.Atom>();
            for (Residue3D r:residues) {
                if (Thread.interrupted())
                    throw new AnalysisException("The analysis has been cancelled");
                List<Residue3D.Atom> atoms = r.getAtoms();
                for (Residue3D.Atom a: atoms) {
                    if (    (Adenine3D.class.isInstance(r) && (a.getName().equals("N9") || a.getName().equals("C8") || a.getName().equals("N7") || a.getName().equals("C5") || a.getName().equals("C6") || a.getName().equals("N6") || a.getName().equals("N1") || a.getName().equals("C2") || a.getName().equals("N3") || a.getName().equals("C4"))
                            ||  Guanine3D.class.isInstance(r) && (a.getName().equals("N9") || a.getName().equals("C8") || a.getName().equals("N7") || a.getName().equals("C5") || a.getName().equals("C6") || a.getName().equals("O6") || a.getName().equals("N1") || a.getName().equals("C2") || a.getName().equals("N2") || a.getName().equals("N3") || a.getName().equals("C4"))
                            ||  Uridine3D.class.isInstance(r) && (a.getName().equals("N1") || a.getName().equals("C2") || a.getName().equals("O2") || a.getName().equals("N3") || a.getName().equals("C4") || a.getName().equals("04") || a.getName().equals("C5") || a.getName().equals("C6"))
                            ||  Cytosine3D.class.isInstance(r) && (a.getName().equals("N1") || a.getName().equals("C2") || a.getName().equals("O2") || a.getName().equals("N3") || a.getName().equals("C4") || a.getName().equals("N4") || a.getName().equals("C5") || a.getName().equals("C6")))
                            && !a.hasCoordinatesFilled()) {
                        a.setCoordinates(0,0,0);
                        atomsFilled.add(a);
                    }
                }
            }
            //####### end of trick ###########
            this.status = "Generation of the PDB file";
            File f = IOUtils.createTemporaryFile("rnaview");
            PDBFileIO.writePDBFile(residues, true, f);
            //######### trick deleted ############
            // to avoid to produce later a PDB file with atoms containing coordinates of 0 0 0
            for (Residue3D.Atom a:atomsFilled)
                a.eraseCoordinates();
            //###################################
            this.status = "PDB generated in "+(System.currentTimeMillis()-start)/1000+"s";
            RNAMLFileIO rnamlParser = new RNAMLFileIO(this.moleculeFactory);
            start = System.currentTimeMillis();
            this.status = "RNAVIEW is running";

            RnaviewServiceLocator locator = new RnaviewServiceLocator();
            Rnaview_PortType rnaview = locator.getRnaview();
            String s = FileUtils.readFileToString(f);
            String pdbAnnotation = rnaview.annotatePDB(s); //practical function from the Apache Commons IO project
            this.status ="RNAVIEW done in "+(System.currentTimeMillis()-start)/1000+"s";

            if (pdbAnnotation == null)
                throw new NoSolutionException("No base-pairs found, no 2D structure plotted (may be a single strand??)");
            else {
                start = System.currentTimeMillis();
                this.status ="Parsing of RNAVIEW output";
                File pdbAnnotationFile = IOUtils.createTemporaryFile("rnaview");
                FileUtils.writeStringToFile(pdbAnnotationFile,pdbAnnotation);
                rnamlParser.parseFile(pdbAnnotationFile, new DefaultProgressMonitor());
                this.status = "Parsing done in "+(System.currentTimeMillis()-start)/1000+"s";
                //only one secondaryStructure generated by RNAVIEW with a tertiaryStructure
                SecondaryStructure ss = rnamlParser.getSecondaryStructures().get(0);
                //[fjossinet] it's more logical that the origin of the secondary structure is the RNAVIEW analysis rather than the RNAML parser
                ss.setParadiseID(new ParadiseID());
                ss.setName("Secondary Structure");
                ss.setSource(Source.getAnalysisAsSource(this.description));
                //next instruction is a crucial point
                //if the secondary structure contains at least 2 molecules with the same sequence, the system will always use the first one. Find a way to improve that...
                //in principle the molecules sent to rnaview are sorted like the molecules returned
                //so I fake the molecules returned to give them the same PARADISE ID than the molecule sent
                List<Molecule> returnedMolecules = new ArrayList<Molecule>(ss.getMolecules());
                for (int i=0; i<returnedMolecules.size();i++) {
                    Molecule returnedMolecule = returnedMolecules.get(i), localMolecule = molecules.get(i);
                    if (returnedMolecule.printSequence().equals(localMolecule.printSequence())) //a security
                        returnedMolecule.getParadiseID().setValue(localMolecule.getParadiseID().getValue());
                }
                //second point: sometimes, RNAVIEW merges 2 sequences from the original sequence in one sequence. So it's impossible for PARADISE to recover the original molecule already in memory
                ss.getFactory().copySecondaryStructure(molecules,ss, false);
                for (Molecule m:molecules) {
                    if (Thread.interrupted())
                        throw new AnalysisException("The analysis has been cancelled");
                    m.removeAllSelectedFeatures();
                    m.addSelectedFeature(ss);
                }
            }
        }
        catch (AnalysisException e) {
            throw e;
        }
        catch (NoSolutionException e) {
            throw e;
        }
        catch (Exception e) {
            throw new AnalysisException(e);
        }
    }
}
