package fr.unistra.ibmc.paradise.services.community;

import fr.unistra.ibmc.paradise.core.analysis.Parameters;
import fr.unistra.ibmc.paradise.services.ParadisePlatform;
import fr.unistra.ibmc.paradise.services.ParadisePlatformAgent;

import fr.unistra.ibmc.paradise.services.community.SubscribeCommunityBehaviour;

import fr.unistra.ibmc.paradise.services.community.tools.XMLTools;
import fr.unistra.ibmc.paradise.Paradise;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import jade.wrapper.StaleProxyException;

import org.apache.log4j.Priority;

import java.io.File;
import java.io.IOException;
import java.util.*;

import jade.core.AID;
import jade.lang.acl.ACLMessage;

/**
 *
 * This class of agent is used for community fonctionnalities of the program
 * LOGIN / UNLOG / REGISTER / Getting connected users
 *
 * it also contain the variabled used to call the ParadiseCommunityAgent's services
 *
 * @author F. Pozzobon
 *
 */
public class ParadiseCommunityAgent extends ParadisePlatformAgent {

    private static final long serialVersionUID = -6636640922422847230L;

    // The services
    public static final String LOGIN="Login",
            UNLOG = "Unlog",
            REGISTER="Register",
            GET_CONNECTED_USERS = "Get_Connected_users";

    //
    public static final String CONTACT_CONNECTED = "contact_connected",
            CONTACT_DISCONNECTED = "contact_disconnected";

    // all the categories of services that can do the ParadiseCommunityAgent
    public static final String categories[] = {LOGIN, UNLOG, REGISTER, GET_CONNECTED_USERS};

    // the name of the XML file that contain users
    protected static String xmlUserFile = "users.xml";

    // the path of the XML file
    protected String xmlTotalPath;
    protected Map<AID, Contact> connectedUsers;
    protected Parameters parameters;
    protected ParadisePlatform platform;

    //all the existing users in the server
    protected Map<String, String> existingUsers;

    // The XML document used
    protected Document userDoc;


    /**
     * A ParadiseCommunityAgent needs a ParadisePlatform 
     * to manage contacts.
     *
     */
    public ParadiseCommunityAgent(ParadisePlatform platform, String name, String categories[]) throws StaleProxyException {
        super(platform, name, categories);
        this.xmlTotalPath = Paradise.getParadiseDirectory()+System.getProperty("file.separator")+ xmlUserFile;
        this.connectedUsers = Collections.synchronizedMap(new HashMap<AID, Contact>());
        this.parameters = new Parameters();
        this.existingUsers = new HashMap<String, String>();
        this.userDoc = null;

        //Reading the XML file which contain all the users registered in the server
        this.loadExistingUsers();
    }



    /**
     *
     * Load existing users from XML file
     *
     */
    protected void loadExistingUsers()
    {

        // Check if the folder exist, if not we create it
        File xmlFile = new File(this.xmlTotalPath);
        if(xmlFile.exists()) {
            userDoc = XMLTools.readXMLFile(xmlTotalPath);

            if(userDoc != null) {
                NodeList list = userDoc.getElementsByTagName("User_List");
                if(list.getLength() == 0){

                    //ParadisePlatform.getPlatform().writeLogMessage(this.getClass(), Priority.WARN, "User List doesn't exist in the user XML File", "Loading XML File", this.name, null);
                }
                else
                {
                    //ParadisePlatform.getPlatform().writeLogMessage(this.getClass(), Priority.INFO, "Loading users from XML File", "", this.getName(), null);
                    for(int i = 0; i < list.getLength();++i)
                    {

                        NodeList userNodeList = ((Element)(list.item(i))).getElementsByTagName("User");

                        for(int j=0;j<userNodeList.getLength();++j)
                        {
                            String name = XMLTools.getText(userNodeList.item(j),"User_Name");
                            String pass = XMLTools.getText(userNodeList.item(j),"User_Password");

                            if (pass !=null && name!=null)
                            {
                                this.existingUsers.put(name, pass);
                            }
                            else
                            {
                                //ParadisePlatform.getPlatform().writeLogMessage(this.getClass(), Priority.WARN, "User missing informations : "+name+" "+pass, "Loading XML File", this.name, null);
                            }

                        }

                    }
                }

            }
        }

        else {
            if(XMLTools.createXMLDocument(this.xmlTotalPath, "<?xml version=\"1.0\" ?>\n<User_List>\n</User_List>"))
                //we call again the method if we have added the file
                this.loadExistingUsers();
        }


    }

    /**
     * Create a new user in the XML if he doesn't exist
     * return true in this case
     * either return false
     * @param userName
     * @param password
     * @return
     */
    public boolean writeUser(String userName, String password)
    {
        if(userDoc!=null)
        {
            //we create the user only if he doesn't exist
            if(XMLTools.getNodeFromAttribute(this.userDoc, "User_Name", userName)==null)
            {

                Node userNode = XMLTools.addChild(this.userDoc,this.userDoc.getFirstChild(), "User", "");
                XMLTools.addChild(this.userDoc, userNode, "User_Name", userName);
                XMLTools.addChild(this.userDoc, userNode, "User_Password", password);
                XMLTools.updateXMLFile(xmlTotalPath, this.userDoc);
                this.existingUsers.put(userName, password);
                return true;

            }
        }
        return false;

    }

    /**
     * Return the map of connected users
     * @return
     */
    public Map<AID, Contact> getConnectedUsers()
    {
        return this.connectedUsers;
    }

    /**
     * Return the Contact associated with the conversation whose id is given as argument
     *
     * @param user
     * @return null if no user registered with this conversation id
     */
    public Contact getContact(AID user) {
        return this.connectedUsers.get(user);
    }

    /**
     * close a converstation
     * @param userID
     * @param conversationId
     */
    public void closeConversation(AID userID, String conversationId) {
        this.writeLogMessage("Job finished", Priority.INFO, conversationId, this.getName(), null);
        //this.users.remove(userID);
        this.getParadisePlatform().unregisterConversation(this.getLocalName(), conversationId);
    }

    /**
     *
     * The method login an user if he is not already logged on and if some criteria are accepted
     *
     * @param userName
     * @param userPass
     * @param anonymous : "True" if he tried to connect in anonymous
     * @param requester : the AID of the user that asked to login
     * @return boolean : true if the user can log on, false either
     *
     */
    public boolean logIn(String userName, String userPass, String anonymous, AID requester)
    {
        if((userPass!=null || anonymous!=null) && userName!=null && !this.connectedUsers.containsKey(requester))
        {
            String pass = this.existingUsers.get(userName);
            if(userPass.equals(pass) || anonymous.equals("true"))
            {
                Contact newContact;
                if(anonymous.equals("true"))
                {
                    newContact = new Contact(requester, userName, false);
                    this.successFullLogin(requester, userName, false);
                    this.connectedUsers.put(requester, newContact);
                }
                else
                {
                    newContact = new Contact(requester, userName,true);
                    this.successFullLogin(requester, userName,true);
                    this.connectedUsers.put(requester, newContact);
                }
                return true;
            }
            else
            {
                return
                        false;
            }
        }
        else
        {
            return false;
        }
    }


    /**
     *
     * when someone is successfully logged in we send the Contact to all other connected users
     * @param requester
     * @param userName
     * @param registered
     */
    public void successFullLogin(AID requester, String userName, boolean registered)
    {

        Contact newUser = new Contact(requester, userName, registered);

        ACLMessage cfp = new ACLMessage(ACLMessage.CFP);
        cfp.setConversationId(CONTACT_CONNECTED);
        try {
            cfp.setContentObject(newUser);
        } catch (IOException e) {
            e.printStackTrace();
        }

        Iterator it = this.getConnectedUsers().entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry pairs = (Map.Entry)it.next();
            cfp.addReceiver((AID)pairs.getKey());
        }

        send(cfp);

    }

    /**
     * Verify that this user is connected
     * @param name
     * @return
     */
    public boolean isNameConnected(String name)
    {

        Map.Entry pairs = null;
        Iterator it = this.connectedUsers.entrySet().iterator();
        while (it.hasNext()) {
            pairs = (Map.Entry)it.next();
            if(((Contact)pairs.getValue()).name.equals(name))
            {
                return true;
            }
        }
        return false;

    }

    /**
     * Verify that this user is registered
     * @param name
     * @return
     */
    public boolean isNameRegistered(String name)
    {
        return this.existingUsers.containsKey(name);
    }


    /**
     *
     * Remove the AID user from the connected users list
     * @param requester
     *
     */
    public String unLog(AID requester)
    {

        String name = "";
        if(this.connectedUsers.containsKey(requester))
        {
            name = getContact(requester).name;
            this.connectedUsers.remove(requester);

            // we have to alert everyone that the user is disconnected
            this.disconnectedUser(requester);
        }
        return name;
    }

    /**
     * 	Alert every connected users that this user is disconnected
     *
     * @param requester
     *
     */
    public void disconnectedUser(AID requester)
    {

        ACLMessage cfp = new ACLMessage(ACLMessage.CFP);
        cfp.setConversationId(CONTACT_DISCONNECTED);
        try {
            cfp.setContentObject(requester);
        } catch (IOException e) {
            e.printStackTrace();
        }

        Iterator it = this.getConnectedUsers().entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry pairs = (Map.Entry)it.next();
            cfp.addReceiver((AID)pairs.getKey());
        }

        send(cfp);
    }

    /**
     * Initialize the behaviours of the ParadiseCommunityAgent
     */
    protected void setup() {
        super.setup();
        this.addBehaviour(new SubscribeCommunityBehaviour());
        this.addBehaviour(new CancelCommunityBehaviour());
        this.addBehaviour(new RequestCommunityBehaviour());
    }

}
