package fr.unistra.ibmc.paradise.tools;

import jade.gui.GuiEvent;
import jade.wrapper.StaleProxyException;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.lang.acl.UnreadableException;
import fr.unistra.ibmc.paradise.core.analysis.Analysis;
import fr.unistra.ibmc.paradise.core.analysis.Parameter;
import fr.unistra.ibmc.paradise.core.analysis.Parameters;
import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.ParadiseFeature;
import fr.unistra.ibmc.paradise.core.ParadiseID;
import fr.unistra.ibmc.paradise.core.Source;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;

import java.util.*;


public class MathematicaAgent extends AbstractParadiseToolAgent {

    public MathematicaAgent(final ParadiseWorkingSession workingSession) throws StaleProxyException {
        super(new MathematicaTool(workingSession));
    }

    public void annotateTertiaryStructure(TertiaryStructure ts) {
        ArrayList<Molecule> molecules = new ArrayList<Molecule>();
        for (Molecule m:ts.getMolecules()) {
            m.removeAllSelectedFeatures();
            for (Residue3D r : ts.getSubFeatures(Residue3D.class, m))
                m.addSelectedFeature(r);
            molecules.add(m);
        }
        this.addBehaviour(new ListAgentsBehaviour(Analysis.TERTIARY_STRUCTURE_ANNOTATION, new DefaultAnswerBehaviour(Analysis.TERTIARY_STRUCTURE_ANNOTATION,molecules, new Parameters()), paradiseTool.getProgressMonitor()));
    }

    public void predictTertiaryStructure(List<StructuralDomain> structuralDomains) {
        Set<Molecule> molecules = new HashSet<Molecule>();
        for (StructuralDomain sd:structuralDomains) {
            Collection<Molecule> _molecules = sd.getMolecules();
            molecules.addAll(_molecules);
            Molecule m = _molecules.iterator().next();
            m.addSelectedFeature(sd); //the selectedFeature is only for one Molecule even for an inter-Helix (otherwise the interHelix will be generated twice)
        }
        this.addBehaviour(new ListAgentsBehaviour(Analysis.TERTIARY_STRUCTURE_PREDICTION, new DefaultAnswerBehaviour(Analysis.TERTIARY_STRUCTURE_ANNOTATION, new ArrayList<Molecule>(molecules), new Parameters()), paradiseTool.getProgressMonitor()));
    }

    public void predictSecondaryStructure(Molecule m) {
        this.addBehaviour(new ListAgentsBehaviour(Analysis.SECONDARY_STRUCTURE_PREDICTION, new DefaultAnswerBehaviour(Analysis.SECONDARY_STRUCTURE_PREDICTION, new ArrayList<Molecule>(Arrays.asList(new Molecule[]{m})), new Parameters()), paradiseTool.getProgressMonitor()));
    }

    /**
     * Do nothing for this agent
     * @param guiEvent
     */
    protected void onGuiEvent(GuiEvent guiEvent) {
    }

    /**
     * This behaviour is really simple. It copies the selected features from the molecules found in the message received to the original ones.
     */
    private class DefaultAnswerBehaviour extends AnswerBehaviour {

        private String analysisType;

        private DefaultAnswerBehaviour(String analysisType, ArrayList<Molecule> selectedMolecules, Parameters selectedParameters) {
            super(selectedMolecules, selectedParameters);
            this.analysisType = analysisType;
        }

        public void action() {
            ACLMessage message = myAgent.receive(MessageTemplate.MatchConversationId("" + this.conversationId));
            if (message != null && message.getPerformative() == ACLMessage.PROPOSE) {
                try {
                    ArrayList<Molecule> remoteMolecules = ((ArrayList<Molecule>) message.getContentObject());
                    if (this.analysisType.equals(Analysis.TERTIARY_STRUCTURE_ANNOTATION)) {
                        SecondaryStructure ss = null;
                        for (Molecule remoteMolecule:remoteMolecules)
                            for (ParadiseFeature f:remoteMolecule.getSelectedFeatures()) {
                                if (ss == null)
                                    ss = (SecondaryStructure)f;
                                ss.getFactory().copySecondaryStructure(selectedMolecules,ss,false);
                            }
                        selectedMolecules.get(0).getSelectedFeatures().get(0).getParentFeatures(TertiaryStructure.class).get(0).linkSecondaryStructure(ss);
                    }
                    else if (this.analysisType.equals(Analysis.SECONDARY_STRUCTURE_PREDICTION)) {
                        SecondaryStructure ss = remoteMolecules.get(0).getSelectedFeatures(SecondaryStructure.class).get(0);
                        ss.getFactory().copySecondaryStructure(selectedMolecules, ss, false);    
                    }
                    else if (this.analysisType.equals(Analysis.TERTIARY_STRUCTURE_PREDICTION)) {
                        List<TertiaryStructure> tss = remoteMolecules.get(0).getSelectedFeatures(TertiaryStructure.class);
                        if (tss.size() > 0) {
                            StructuralFeatureFactory factory = paradiseTool.createStructuralFeatureFactory();
                            TertiaryStructure remote3D = tss.get(0), local3D = factory.createTertiaryStructure(new ParadiseID(),"tertiary Structure", Source.getGraphicalToolAsSource(paradiseTool.getDescription()));
                            List<Residue3D> residues3D = new ArrayList<Residue3D>();
                            for (Residue3D r : remote3D.getResidues3D()) {
                                Molecule localMolecule = null, remoteMolecule = r.getMolecules().iterator().next();
                                for (Molecule _m:selectedMolecules) {
                                    if (remoteMolecule.equals(_m)) {
                                        localMolecule = _m;
                                        break;
                                    }
                                }
                                try {
                                    Residue3D r3d = factory.createResidue3D(local3D, localMolecule , r.getFullLocation(remoteMolecule).getStart());
                                    for (Residue3D.Atom ra : r.getAtoms()) {
                                        if (ra.hasCoordinatesFilled()) {
                                            Residue3D.Atom la = r3d.getAtom(ra.getName());
                                            la.setCoordinates(ra.getX(), ra.getY(), ra.getZ());
                                        }
                                    }
                                    residues3D.add(r3d);
                                }
                                catch (BiologicalSymbolException e) {
                                    monitor.printException(e);
                                    return;
                                }
                            }

                        }
                    }
                }
                catch (Exception e) {
                    monitor.printException(e);
                }
                monitor.stopAnimation();
                this.done = true;
            } else if (message != null && message.getPerformative() == ACLMessage.INFORM) {
                monitor.printMessage(message.getContent());
            } else if (message != null && message.getPerformative() == ACLMessage.FAILURE) {
                try {
                    monitor.printException((Exception) message.getContentObject());
                } catch (UnreadableException e) {
                    e.printStackTrace();
                }
                monitor.stopAnimation();
                this.done = true;
            }
        }

    }
}
