package fr.unistra.ibmc.paradise.tools;

import fr.unistra.ibmc.paradise.BiologicalSymbolException;
import fr.unistra.ibmc.paradise.Paradise;
import fr.unistra.ibmc.paradise.core.*;
import jade.core.Profile;
import jade.core.ProfileImpl;
import jade.wrapper.AgentContainer;
import jade.wrapper.ControllerException;

import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

public final class ParadiseWorkingSession extends AbstractMoleculeFactory {

    private List<Molecule> molecules;
    //the current agent container attached to this working session
    private AgentContainer container;
    private static String clientAddress, platformAddress;
    private List<Transformation> molecularTransformations;
    transient private List<ParadiseTool> paradiseTools;
    private File currentParadiseProject;

    public ParadiseWorkingSession() throws ControllerException, UnknownHostException, ConnectionException, IncompatibilityException {
        this(null);
    }

    /**
     * Create a PARADISE working session connecting a PARADISE platform defined by the IP address given as argument
     *
     * @param platformAddress the IP address of the machine hosting a PARADISE platform. Can be equal to localhost if the PARADISE platform is on the same machine than the working session. Can be null.
     * @throws UnknownHostException
     */
    public ParadiseWorkingSession(String platformAddress) throws UnknownHostException, ControllerException, ConnectionException, IncompatibilityException {
        super("ParadiseWorkingSession[at]" + InetAddress.getLocalHost().getHostAddress());
        this.molecules = new ArrayList<Molecule>();
        this.paradiseTools = new ArrayList<ParadiseTool>();
        this.molecularTransformations = new ArrayList<Transformation>();
        if (platformAddress != null) {
            this.clientAddress = InetAddress.getLocalHost().getHostAddress();            
            this.connectToPlatform(platformAddress);
        }
    }

    public void connectToPlatform(String platformAddress) throws UnknownHostException, ControllerException, ConnectionException, IncompatibilityException {
        if (platformAddress == null)
            return;
        this.platformAddress = platformAddress;
        jade.core.Runtime r = jade.core.Runtime.instance();
        Profile profile = new ProfileImpl();
        if (platformAddress.equals("localhost")) // local work
            profile.setParameter(Profile.MAIN_HOST, this.clientAddress);
        else {  //distant work
            profile.setParameter(Profile.MAIN_HOST, platformAddress);
            //####### a tip to get the public ip address of the user #################
            //otherwise, if the user is in a LAN, the agent will give the local IP address for the LOCALHOST (for example 192.168.1.10)
            String publicIP = null;
            /*
            try {
                URL tempURL = new URL("http://www.whatismyip.org/");
                HttpURLConnection tempConn = (HttpURLConnection)tempURL.openConnection();
                InputStream tempInStream = tempConn.getInputStream();
                InputStreamReader tempIsr = new InputStreamReader(tempInStream);
                BufferedReader tempBr = new BufferedReader(tempIsr);
                publicIP = tempBr.readLine();
                tempBr.close();
                tempInStream.close();
            } catch (Exception ex) {
                publicIP = this.clientAddress;
            }*/
            publicIP = this.clientAddress;
            profile.setParameter(Profile.LOCAL_HOST,publicIP);
        }
        profile.setParameter(Profile.CONTAINER_NAME, this.clientAddress);
        this.container = r.createAgentContainer(profile);
        if (this.container == null)
            throw new ConnectionException();
        this.container.start();
        //the compatibility test between the PARADISE engine versions of the client and the platform is done here !!!!
        if (!Paradise.releaseName.equals(this.container.getPlatformName()))
            throw new IncompatibilityException();
    }

    public void restorePreviousWorkingSession() {
        final File backup = new File(new File(Paradise.getInstallPath()),"paradise.backup");
        if (backup.exists()) {
            this.molecules.clear();
            ObjectInputStream stream = null;
            try {
                stream = new ObjectInputStream(new FileInputStream(backup));
                //the first object serialized is the file of the paradise project
                this.currentParadiseProject = (File)stream.readObject();
                for (Molecule m: ((List<Molecule>)stream.readObject())) {
                    this.molecules.add(m);
                    m.recoverBackupAnnotations();
                }
            } catch (IOException e1) {
                e1.printStackTrace();
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
            }
            finally {
                if (stream != null)
                    try {
                        stream.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
            }
        }
    }


    public File getCurrentParadiseProject() {
        return this.currentParadiseProject;
    }

    public void setCurrentParadiseProject(File paradiseProject) {
        this.currentParadiseProject = paradiseProject;
    }

    public void addMolecularAffiliation(Molecule precursor, int precursorStart, int precursorEnd, Molecule product, int productStart, int productEnd) {
        this.molecularTransformations.add(new Transformation(precursor, precursorStart, precursorEnd, product, productStart, productEnd));
    }

    public List<Transformation> getMolecularAffiliations() {
        return new ArrayList<Transformation>(molecularTransformations);
    }

    public RNA createRNA(ParadiseID paradiseID, String name, Source source, String sequence) throws BiologicalSymbolException {
        RNA rna = super.createRNA(paradiseID, name, source, sequence);
        this.molecules.add(rna);
        return rna;
    }

    public Protein createProtein(ParadiseID paradiseID, String name, Source source, String sequence) throws BiologicalSymbolException {
        Protein protein = super.createProtein(paradiseID, name, source, sequence);
        this.molecules.add(protein);
        return protein;
    }

    public LinearGenomicDNA createLinearGenomicDNA(ParadiseID paradiseID, String name, Source source, String sequence) throws BiologicalSymbolException {
        LinearGenomicDNA dna = super.createLinearGenomicDNA(paradiseID, name, source, sequence);
        this.molecules.add(dna);
        return dna;
    }

    public CircularGenomicDNA createCircularGenomicDNA(ParadiseID paradiseID, String name, Source source, String sequence) throws BiologicalSymbolException {
        CircularGenomicDNA dna = super.createCircularGenomicDNA(paradiseID, name, source, sequence);
        this.molecules.add(dna);
        return dna;
    }

    public void removeMolecule(Molecule molecule) {
        this.molecules.remove(molecule);
    }

    public void removeAllMolecules() {
        for (Molecule m : this.getMolecules())
            this.removeMolecule(m);
    }

    public void addAllMolecules(Collection<Molecule> molecules) {
        this.molecules.addAll(molecules);
    }

    public List<Molecule> getMolecules() {
        return new ArrayList<Molecule>(this.molecules);
    }

    public List<RNA> getRNAs() {
        List<RNA> rnas = new ArrayList<RNA>();
        for (Molecule m:this.getMolecules())
            if (RNA.class.isInstance(m))
                rnas.add((RNA)m);
        return rnas;
    }

    public List<Protein> getProteins() {
        List<Protein> proteins = new ArrayList<Protein>();
        for (Molecule m:this.getMolecules())
            if (Protein.class.isInstance(m))
                proteins.add((Protein)m);
        return proteins;
    }

    /**
     * Return all the MainParadiseFeature stored in the working session and which are instances of the class given as argument
     * @param c the class to which the MainParadiseFeature objects returned have to be instance of.
     * @return
     */
    public <T extends MainParadiseFeature> List<T> getMainParadiseFeatures(Class<T> c) {
        Set<T> features = new HashSet<T>();
        for (Molecule m:this.getMolecules())
            for (Molecule.Annotation a:m.getAnnotations(c))
                features.add((T) a.getFeature());
        return new ArrayList<T>(features);
    }


    /**
     * Return the molecule stored in the working session with the paradise ID equals to the paradise ID given as argument
     *
     * @param paradiseID
     * @return null if no molecule with such an paradise ID
     */
    public Molecule getMolecule(ParadiseID paradiseID) {
        for (Molecule m : this.molecules)
            if (m.getParadiseID().equals(paradiseID))
                return m;
        return null;
    }

    public Molecule getMolecule(String source) {
        for (Molecule m : this.molecules)
            if (m.getSource().equals(source))
                return m;
        return null;
    }

    /**
     * @return null is the paradise working session is not connected to a paradise platform
     */
    public AgentContainer getAgentContainer() {
        return this.container;
    }

    public String getAddress() {
        return clientAddress;
    }

    public String getPlatformAddress() {
        return platformAddress;
    }

    public List<ParadiseTool> getParadiseTools() {
        return new ArrayList<ParadiseTool>(this.paradiseTools);
    }

    public static String getWikiURL() {
        return "http://paradise-ibmc.u-strasbg.fr/wiki/en/";
    }
}
