package fr.unistra.ibmc.paradise.utils;

import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;


public class DrawingUtils {

    /**
     * Return the distance between two points
     *
     * @param p1
     * @param p2
     * @return the distance
     */
    public static double getDistance(final Point2D p1, final Point2D p2) {
        return getDistance(p1.getX(), p1.getY(), p2.getX(), p2.getY());
    }

    /**
     * Return the distance between two points
     *
     * @param x1
     * @param y1
     * @param x2
     * @param y2
     * @return the distance
     */

    public static double getDistance(final double x1, final double y1, final double x2, final double y2) {
        final double horizontal = x1 - x2;
        final double vertical = y1 - y2;
        return Math.sqrt(horizontal * horizontal + vertical * vertical);
    }

    /**
     * Calculate two points placed on the same line defined by the points p1 and p2
     * coordinates. The first point returned is separated from the p1 point by the distance value.
     * The same for the second point returned and the p2 point.
     * If the distance value is greater than 0, the first point returned is calculated in direction of the
     * the p2 point and the second point returned is calculated in direction of the p1 point.
     * The opposite directions if the distance value is lesser than 0.
     *
     * @param p1       the first original point
     * @param p2       the second original point
     * @param distance the distance between the original points and the corresponding new ones.
     * @return the two new points
     */
    public static Point2D[] fit(final Point2D p1, final Point2D p2, final double distance) {
        return fit(p1.getX(), p1.getY(), p2.getX(), p2.getY(), distance);
    }

    /**
     * Calculate two points placed on the same line defined by the points having the (x1,y1) and (x2,y2)
     * coordinates. The first point returned is separated from the point defined by the (x1,y1) coordinates
     * by the distance value. The same for the second point returned and the point defined by the (x2,y2)
     * coordinates.
     * If the distance value is greater than 0, the first point returned is calculated in direction of the
     * the point defined by the (x2,y2) coordinates and the second point returned is calculated in direction
     * of the point defined by the (x1,y1). The opposite directions if the distance value is lesser than 0.
     *
     * @param x1       the x coordinate of the first original point
     * @param y1       the y coordinate of the first original point
     * @param x2       the x coordinate of the second original point
     * @param y2       the y coordinate of the second original point
     * @param distance the distance between the original points and the corresponding new ones.
     * @return the two new points
     */
    public static Point2D[] fit(final double x1, final double y1, final double x2, final double y2, final double distance) {
        final Point2D[] drawingPoints = new Point2D[2];
        final double angle = getAngle(getDistance(x1, y1, x1, y2), getDistance(x1, y2, x2, y2));
        final double newx1;
        final double newy1;
        final double newx2;
        final double newy2;

        if (x1 >= x2) {
            newx2 = x2 + DrawingUtils.getAdjacentSide(angle, distance);
            newx1 = x1 - DrawingUtils.getAdjacentSide(angle, distance);
        } else {
            newx2 = x2 - DrawingUtils.getAdjacentSide(angle, distance);
            newx1 = x1 + DrawingUtils.getAdjacentSide(angle, distance);
        }
        if (y1 >= y2) {
            newy2 = y2 + DrawingUtils.getOppositeSide(angle, distance);
            newy1 = y1 - DrawingUtils.getOppositeSide(angle, distance);
        } else {
            newy2 = y2 - DrawingUtils.getOppositeSide(angle, distance);
            newy1 = y1 + DrawingUtils.getOppositeSide(angle, distance);
        }

        drawingPoints[0] = new Point2D.Double(newx1, newy1);
        drawingPoints[1] = new Point2D.Double(newx2, newy2);

        return drawingPoints;
    }

    /**
     * Calculate a new point from an original one after a rotation step
     * defined by a specific angle value and a center point.
     *
     * @param p      the original point
     * @param angle  the angle of rotation in radians
     * @param center the center of rotation
     * @return the new point after the rotation step
     */
    public static Point2D getRotatedPoint(final Point2D p, final double angle, final Point2D center) {
        if (angle != 0.0) {
            //we set the rotation
            final AffineTransform rot = new AffineTransform();
            rot.setToRotation(angle, center.getX(), center.getY());
            //we get the rotated point with this transformation
            final Point2D pointRot = rot.transform(p, null);
            return new Point2D.Double(pointRot.getX(), pointRot.getY());
        }
        return p;
    }

    /**
     * Calculate the two points defining a line perpendicular to the one passing through p1 and p2.
     * This perpendicular line cut the (p1,p2) line on the p0 point.
     *
     * @param p0       the location on the line defined by p1 and p2 where we want to place the perpendicular points
     * @param p1
     * @param p2
     * @param distance the distance between each perpendicular point and pO
     * @return the perpendicular points
     */

    public static Point2D[] getPerpendicular(final Point2D p0, final Point2D p1, final Point2D p2, final double distance) {
        final Point2D[] pp = new Point2D[2];
        pp[0] = (Point2D) p0.clone();
        pp[1] = (Point2D) p0.clone();
        final double angle = getAngle(p1.getY() - p2.getY(), p1.getX() - p2.getX());
        pp[0].setLocation(pp[0].getX() + DrawingUtils.getOppositeSide(angle, distance), pp[0].getY() - DrawingUtils.getAdjacentSide(angle, distance));
        pp[1].setLocation(pp[1].getX() - DrawingUtils.getOppositeSide(angle, distance), pp[1].getY() + DrawingUtils.getAdjacentSide(angle, distance));
        return pp;
    }

    public static double getAngle(final double oppositeSide, final double adjacentSide) {
        return Math.atan(oppositeSide / adjacentSide);
    }

    public static double getOppositeSide(final double angle, final double hypothenuse) {
        return Math.sin(angle) * hypothenuse;
    }

    public static double getAdjacentSide(final double angle, final double hypothenuse) {
        return Math.cos(angle) * hypothenuse;
    }

    public static Point2D getCenter(final Point2D p1, final Point2D p2) {
        return new Point2D.Double((p1.getX() + p2.getX()) / 2, (p1.getY() + p2.getY()) / 2);
    }

    public static double crossProduct(final Point2D p1, final Point2D p2, final Point2D p3) {
        return crossProduct(p1.getX(), p1.getY(), p2.getX(), p2.getY(), p3.getX(), p3.getY());
    }


    /**
     * Gives the cross product for three points.
     * p1 is the common point for the two vectors (p1p2 and p1p3)
     *
     * @param p1x
     * @param p1y
     * @param p2x
     * @param p2y
     * @param p3x
     * @param p3y
     * @return the cross product
     */
    public static double crossProduct(final double p1x, final double p1y, final double p2x, final double p2y, final double p3x, final double p3y) {
        final double a1 = p2x - p1x;
        final double a2 = p2y - p1y;
        final double b1 = p3x - p1x;
        final double b2 = p3y - p1y;
        return a1 * b2 - a2 * b1;
    }

    /**
     * Return angle between lines p1p2 and p1p3
     *
     * @param p1
     * @param p2
     * @param p3
     * @return
     */
    public static double getAngle(Point2D p1, Point2D p2, Point2D p3) {
        double distp1p2 = getDistance(p1, p2),
                distp1p3 = getDistance(p1, p3),
                distp2p3 = getDistance(p2, p3);
        return Math.acos((distp1p2 * distp1p2 + distp1p3 * distp1p3 - distp2p3 * distp2p3) / (2 * distp1p2 * distp1p3));

    }
}
