package fr.unistra.ibmc.paradise.utils;

import fr.unistra.ibmc.paradise.core.features.SecondaryStructure;
import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.features.StructuralAlignment;
import fr.unistra.ibmc.paradise.core.features.TertiaryStructure;

import java.io.*;
import java.util.Timer;
import java.util.TimerTask;
import java.net.URL;


public class IOUtils {

    /**
     * Writes an InputStream into a given file.
     */
    public static void writeToFile(String fileName, InputStream in) {
        File f = new File(fileName);
        BufferedOutputStream fOut = null;
        try {
            fOut = new BufferedOutputStream(new FileOutputStream(f));
            byte[] buffer = new byte[32 * 1024];
            int bytesRead = 0;
            while ((bytesRead = in.read(buffer)) != -1) {
                fOut.write(buffer, 0, bytesRead);
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        finally {
            try {
                in.close();
                fOut.close();
            }
            catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    public static boolean isParadiseProject(File file) {
        return file.isDirectory() && file.listFiles(new FilenameFilter() {
            public boolean accept(File file, String s) {
                return (Molecule.class.getSimpleName()+"s").equals(s)
                        || (TertiaryStructure.class.getSimpleName()+"s").equals(s)
                        || (SecondaryStructure.class.getSimpleName()+"s").equals(s)
                        || (StructuralAlignment.class.getSimpleName()+"s").equals(s);
            }
        }).length == 4;
    }

    public static boolean containsStructuralAlignmentData(File paradiseProject) {
        if (!isParadiseProject(paradiseProject))
            return false;
        else
            return new File(paradiseProject, StructuralAlignment.class.getSimpleName()+"s").listFiles(new FileFilter() {

                public boolean accept(File file) {
                    return file.getName().endsWith(".rnaml");
                }
            }).length > 0;
    }

    public static boolean containsTertiaryStructureData(File paradiseProject) {
        if (!isParadiseProject(paradiseProject))
            return false;
        else
            return new File(paradiseProject, TertiaryStructure.class.getSimpleName()+"s").listFiles(new FileFilter() {

                public boolean accept(File file) {
                    return file.getName().endsWith(".rnaml");
                }
            }).length > 0;
    }

    public static boolean containsSecondaryStructureData(File paradiseProject) {
        if (!isParadiseProject(paradiseProject))
            return false;
        else
            return new File(paradiseProject, SecondaryStructure.class.getSimpleName()+"s").listFiles(new FileFilter() {

                public boolean accept(File file) {
                    return file.getName().endsWith(".rnaml");
                }
            }).length > 0;
    }

    public static void deleteDirectory(File directory) {
        clearDirectory(directory);
        directory.delete();
    }

    public static void clearDirectory(File directory) {
        for(File f:directory.listFiles())
            if(f.isDirectory())
                deleteDirectory(f);
            else
                f.delete();
    }

    public static void copyDirectory(File sourceLocation, File targetLocation) throws IOException {
        if (sourceLocation.isDirectory()) {
            if (!targetLocation.exists()) {
                targetLocation.mkdir();
            }
            String[] children = sourceLocation.list();
            for (int i=0; i<children.length; i++) {
                if (!children[i].equals(".svn"))
                    copyDirectory(new File(sourceLocation, children[i]), new File(targetLocation, children[i]));
            }
        } else {
            InputStream in = new FileInputStream(sourceLocation);
            OutputStream out = new FileOutputStream(targetLocation);

            // Copy the bits from instream to outstream
            byte[] buf = new byte[1024];
            int len;
            while ((len = in.read(buf)) > 0) {
                out.write(buf, 0, len);
            }
            in.close();
            out.close();
        }
    }

    public static File createTemporaryFile(String fileName) throws IOException {
        File f = new File(fr.unistra.ibmc.paradise.Paradise.getTmpDirectory(),fileName+System.nanoTime());
        f.deleteOnExit();
        return f;
    }


    static private class CommandLineProgressBar {
        private Timer timer;

        private CommandLineProgressBar() {
        }

        private void start() {
            this.timer = new Timer();
            this.timer.schedule(new ProgressionTask(), 0, 2000);
        }

        private void stop() {
            this.timer.cancel();
        }

        private class ProgressionTask extends TimerTask {
            public void run() {
                System.out.print(".");
            }
        }
    }

    /**
     * Extract a Jar or Zip file
     * @param destDir where to extract the file
     * @param archivedFile the file to extract (zip or jar files are supported)
     * @throws IOException
     */
    public static void extractArchivedFile(String destDir,File archivedFile) throws IOException {

        java.util.jar.JarFile jar = new java.util.jar.JarFile(archivedFile);
        java.util.Enumeration entries = jar.entries();
        while (entries.hasMoreElements()) {
            java.util.jar.JarEntry file = (java.util.jar.JarEntry) entries.nextElement();
            java.io.File f = new java.io.File(destDir + java.io.File.separator + file.getName());
            if (file.isDirectory()) { // if its a directory, create it
                f.mkdir();
                continue;
            }
            java.io.InputStream is = jar.getInputStream(file); // get the input stream
            java.io.FileOutputStream fos = new java.io.FileOutputStream(f);
            while (is.available() > 0) {  // write contents of 'is' to 'fos'
                fos.write(is.read());
            }
            fos.close();
            is.close();
        }
    }


}
