package fr.unistra.ibmc.s2s.rnalign.features;

import fr.unistra.ibmc.paradise.core.Location;
import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.ParadiseID;
import fr.unistra.ibmc.paradise.core.Source;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;
import fr.unistra.ibmc.s2s.rnalign.graphics.GraphicContext;
import fr.unistra.ibmc.s2s.Mediator;
import fr.unistra.ibmc.s2s.S2S;

import java.awt.*;
import java.text.DecimalFormat;
import java.util.*;
import java.util.List;

import org.jdesktop.swingworker.SwingWorker;

/**
 * An Rnalign Alignment is composed with several LetterSequence objects :
 * <ul>
 * <li> a SequenceMeter object (not editable)
 * <li> a Structure2D3D object
 * <li> one or several aligned BiologicalSequence objects (the real alignment)(editable)
 * </ul>
 * Among the BiologicalSequence objects, one is defined as a ReferenceSequence
 */
public class Alignment {

    public static final byte ISOSTERICITY_MATRICES_CONSERVATION = 0, LW_FAMILIES_CONSERVATION = 1;
    private SequenceMeter sequenceMeter;
    private Structure2D3D structure2D3D;
    private List<BiologicalSequence> biologicalSequences;
    private StructuralAlignment alignment;
    private List<Step> steps;
    private static int MAXIMUM_STEPS = 300;
    private int steps_cursor = -1; //from -1 to this.steps.size()-1
    private Mediator mediator;

    public Alignment(final S2S s2s, final StructuralAlignment alignment, Molecule referenceMolecule, SecondaryStructure referenceStructure) {
        this.alignment = alignment;
        this.mediator = s2s.getMediator();
        this.mediator.getSequencesList().clearList();
        this.steps =  new ArrayList<Step>();
        final List<StructuralIdentity> identities = this.alignment.getStructuralIdentities(true);
        final List<Molecule> molecules = alignment.getSortedMolecules();
        this.sequenceMeter = new SequenceMeter(alignment.getLength());
        this.biologicalSequences = new ArrayList<BiologicalSequence>(molecules.size());
        this.biologicalSequences.add(new BiologicalSequence(identities,referenceMolecule));
        mediator.getSequencesList().addRow(referenceMolecule);
        mediator.getSequencesList().revalidate();
        molecules.remove(referenceMolecule);
        this.structure2D3D = new Structure2D3D(this.biologicalSequences.get(0),referenceStructure);
        new SwingWorker() {
            protected Object doInBackground() {
                try {
                    for (Molecule m: molecules) {
                        Alignment.this.biologicalSequences.add(new BiologicalSequence(identities,m));
                        mediator.getSequencesList().addRow(m);
                        s2s.getAlignmentCanvas().repaint();
                    }
                }
                catch (Exception e) {
                    e.printStackTrace();
                }
                return null;
            }
        }.execute();
    }

    /**
     * Modify the reference molecule and structure in the current alignment.
     * The molecule should be annotated by the structure given as argument (not checked by this method)
     * @param referenceMolecule
     * @param referenceSecondaryStructure
     */
    public void setReferenceMolecule(BiologicalSequence referenceMolecule, SecondaryStructure referenceSecondaryStructure) {
        this.biologicalSequences.remove(referenceMolecule);
        this.biologicalSequences.add(0,referenceMolecule);
        this.structure2D3D = new Structure2D3D(referenceMolecule,referenceSecondaryStructure);
    }

    public boolean hasMoreUndo() {
        return this.steps_cursor >= 0;
    }

    public boolean hasMoreRedo() {
        return this.steps_cursor < this.steps.size()-1;
    }

    public void addStep(Step step) {
        if (this.steps_cursor == Alignment.MAXIMUM_STEPS)
            this.steps.remove(0);
        else if (this.steps_cursor == -1)
            this.steps.clear();
        else
            for (int i=this.steps_cursor+1; i <= this.steps.size()-1;i++)
                this.steps.remove(i);
        this.steps.add(step);
        this.steps_cursor = this.steps.size()-1;
        this.mediator.setRedoUndoButtonsState();
    }

    public void undo(GraphicContext gc) {
        if (this.steps_cursor >= 0)  {
            this.steps.get(this.steps_cursor).undo(this, gc);
            this.steps_cursor--;
        }
    }

    public void redo(GraphicContext gc) {
        if (this.steps_cursor < this.steps.size()-1)  {
            this.steps_cursor++;
            this.steps.get(this.steps_cursor).redo(this, gc);
        }
    }

    public SequenceMeter getSequenceMeter() {
        return this.sequenceMeter;
    }

    public Structure2D3D getStructure2D3D() {
        return this.structure2D3D;
    }

    public BiologicalSequence getBiologicalSequenceAt(final int index) {
        return this.biologicalSequences.get(index);
    }

    /**
     * Create and return a new SecondaryStructure object for the biological sequence given as argument.
     * @param targetSequence
     */
    public SecondaryStructure deduceNewSecondaryStructure(BiologicalSequence targetSequence) throws BiologicalSymbolException {
        StructuralFeatureFactory factory = mediator.getS2s().createStructuralFeatureFactory();
        SecondaryStructure ss = factory.createRNASecondaryStructure(new ParadiseID(), "Secondary Structure", Source.getGraphicalToolAsSource(mediator.getS2s().getDescription()));
        Helix newHelix =null;
        int positionInTargetSequence = -1, previousPositionInTargetSequence = -1, pairedPositionInTargetSequence = -1, previousPairedPositionInTargetSequence = -1;
        char previousEdge1 = '?', previousEdge2= '?', previousOrientation= '?';
        int previousDominant = -1;
        for (int i=0;i <this.structure2D3D.size();i++) {
            StructuralSymbol structuralSymbol = (StructuralSymbol)this.structure2D3D.getSymbol(i);
            positionInTargetSequence = targetSequence.getSymbol(i).getPositionInSequence();
            for (BaseBaseInteraction bbi:structuralSymbol.getReferenceBaseBaseInteractions()) {
                StructuralSymbol pairedStructuralSymbol = structuralSymbol.getPairedSymbol(bbi);
                pairedPositionInTargetSequence = targetSequence.getSymbol(this.structure2D3D.getIndex(pairedStructuralSymbol)).getPositionInSequence();
                if (positionInTargetSequence < pairedPositionInTargetSequence /*to avoid to do twice the same stuff*/) {
                    if (bbi.isSecondaryInteraction()) {
                        if (previousPositionInTargetSequence == positionInTargetSequence-1 && previousPairedPositionInTargetSequence == pairedPositionInTargetSequence+1) {
                            if (newHelix != null) //we add the new interaction to the currentHelix
                                factory.createBaseBaseInteraction(bbi.getEdge1(), bbi.getEdge2(), bbi.getOrientation(), bbi.getDominant(), newHelix, targetSequence.getMolecule(),new Location(new Location(positionInTargetSequence), new Location(pairedPositionInTargetSequence)));
                            else {//we create a new one with the same name than the original one
                                newHelix = factory.createHelix(bbi.getParentFeatures(Helix.class).get(0).getName(),ss,targetSequence.getMolecule(), new Location(new Location(previousPositionInTargetSequence,positionInTargetSequence), new Location(pairedPositionInTargetSequence,previousPairedPositionInTargetSequence)));
                                factory.createBaseBaseInteraction(previousEdge1, previousEdge2, previousOrientation, previousDominant, newHelix, targetSequence.getMolecule(),new Location(new Location(previousPositionInTargetSequence), new Location(previousPairedPositionInTargetSequence)));
                                factory.createBaseBaseInteraction(bbi.getEdge1(), bbi.getEdge2(), bbi.getOrientation(), bbi.getDominant(), newHelix, targetSequence.getMolecule(),new Location(new Location(positionInTargetSequence), new Location(pairedPositionInTargetSequence)));
                            }
                        }
                        else {
                            if (newHelix == null && previousDominant != -1 && previousPositionInTargetSequence != -1 && previousPairedPositionInTargetSequence != -1)
                                // this means that there was a previous secondary interaction that was not contiguous with the previous one (since newHelix == null)
                                //and since we're here, the current secondary interaction is also not contiguous, so we have a secondary interaction that become a tertiary one (if they're no gaps at the corresponding positions in the target sequence)
                                factory.createBaseBaseInteraction(previousEdge1, previousEdge2, previousOrientation, previousDominant, ss, targetSequence.getMolecule(),new Location(new Location(previousPositionInTargetSequence), new Location(previousPairedPositionInTargetSequence)));
                            newHelix = null;
                        }
                        previousEdge1 = bbi.getEdge1();
                        previousEdge2= bbi.getEdge2();
                        previousOrientation = bbi.getOrientation();
                        previousDominant = bbi.getDominant();
                        previousPositionInTargetSequence = positionInTargetSequence;
                        previousPairedPositionInTargetSequence = pairedPositionInTargetSequence;
                    }
                    else if (positionInTargetSequence != -1 && pairedPositionInTargetSequence != -1)//it is a tertiary structure as in the reference sequence (if they're no gaps at the corresponding positions in the target sequence)
                        factory.createBaseBaseInteraction(bbi.getEdge1(), bbi.getEdge2(), bbi.getOrientation(), bbi.getDominant(), ss, targetSequence.getMolecule(),new Location(new Location(positionInTargetSequence), new Location(pairedPositionInTargetSequence)));
                }
            }
        }
        //now with the helices just constructed, we deduce the locations of the single-strands
        Location singleStrandsLocation = new Location(1,targetSequence.getMolecule().getLength());
        for (Helix h:ss.getHelicesWithoutPKnots())
            singleStrandsLocation = singleStrandsLocation.differenceOf(h.getFullLocation(targetSequence.getMolecule()));
        int i=0;
        int[] boundaries = singleStrandsLocation.getBoundaries();
        for (int j=0;j<boundaries.length-1;j+=2)
            factory.createSingleStrand("SS"+(i++),ss,targetSequence.getMolecule(), new Location(boundaries[j],boundaries[j+1]));
        return ss;
    }

    /**
     * Return the number of LetterSequence objects in this RnalignAlignment object
     *
     * @return
     */
    public int getSymbolSequenceCount() {
        return 2 + this.getBiologicalSequenceCount();
    }

    /**
     * Return the number of aligned symbols
     *
     * @return
     */
    public int getBiologicalSequenceCount() {
        return this.biologicalSequences.size();
    }

    /**
     * Return the alignment's length
     *
     * @return
     */
    public int getLength() {
        return this.getSequenceMeter().size();
    }

    public void addBiologicalSequence(final BiologicalSequence seq) {
        int more = 0;
        SymbolSequence sequence = null;
        for (int j = 0; j < this.getSymbolSequenceCount(); j++) {
            sequence = this.getLetterSequenceAt(j);
            more = seq.size() - sequence.size();
            while (more != 0) {
                if (more > 0) {
                    sequence.increaseSize();
                    more--;
                } else {
                    seq.increaseSize();
                    more++;
                }
            }
        }
        this.biologicalSequences.add(this.biologicalSequences.size(),seq);
    }

    public SymbolSequence getLetterSequenceAt(final int index) {
        switch (index) {
            case 0:
                return this.sequenceMeter;
            case 1:
                return this.structure2D3D;
            default :
                return this.biologicalSequences.get(index - 2);
        }
    }

    public int getMaxLabelSize() {
        return 1+Integer.toString(this.biologicalSequences.size()).length()+1; /* the first 1 constant is for the S starting each sequence's label, the second 1 is to have a space between the label and the sequence itself*/
    }

    public BiologicalSequence getBiologicalReferenceSequence() {
        return this.biologicalSequences.get(0);
    }

    public SecondaryStructure getReferenceStructure() {
        return this.structure2D3D.getReferenceSecondaryStructure();
    }

    public StructuralAlignment getStructuralAlignment() {
        return this.alignment;
    }

    /**
     * Modify a Biological symbols in the alignment
     *
     * @param seq      the sequence to modify
     * @param name     the new name for this sequence
     * @param sequence the new sequence with gap symbols allowed
     */
    public void modifyBiologicalSequence(final BiologicalSequence seq, final String name, final String sequence) {
        List<Symbol> newSeq = new ArrayList<Symbol>();
        final char[] letters = this.getAlignedSequence(sequence);
        int positionInSequence = 0;
        for (int i = 0; i < letters.length; i++) {
            Symbol s = Symbol.getSymbol(letters[i]);
            if ( letters[i] != '-')
                s.setPositionInSequence(++positionInSequence);
            newSeq.add(s);
        }
        int more = 0;
        SymbolSequence ls = null;
        for (int j = 0; j < this.getSymbolSequenceCount(); j++) {
            ls = this.getLetterSequenceAt(j);
            more = newSeq.size() - ls.size();
            while (more != 0) {
                if (more > 0) {
                    ls.increaseSize();
                    more--;
                } else {
                    newSeq.add(Symbol.createGap());
                    more++;
                }
            }
        }
        seq.modifySequence(name, newSeq);
    }

    private char[] getAlignedSequence(String alignedSequence) {
        alignedSequence.trim();
        alignedSequence = alignedSequence.replace('.', '-');
        alignedSequence = alignedSequence.toUpperCase();
        final StringBuffer alSequence = new StringBuffer();
        for (int i = 0; i < alignedSequence.length(); i++) {
            char base = alignedSequence.charAt(i);
            if (Character.isWhitespace(base)) base = '-';
            if (base >= 'A' && base <= 'z' || base == '-')
                alSequence.append(base);
        }
        return alSequence.toString().toCharArray();
    }

    public double calculateMutualInformation(int pos1InAlignment, int pos2InAlignment) {
        BiologicalSequence seq = null;
        double A1 = 1, A2 = 1, U1 = 1, U2 = 1, G1 = 1, G2 = 1, C1 = 1, C2 = 1, gap1 = 1, gap2 = 1,
                AA = 1, AU = 1, AG = 1, AC = 1,
                UA = 1, UU = 1, UG = 1, UC = 1,
                GA = 1, GU = 1, GG = 1, GC = 1,
                CA = 1, CU = 1, CG = 1, CC = 1,
                gA = 1, gU = 1, gG = 1, gC = 1,
                Ag = 1, Ug = 1, Gg = 1, Cg = 1,
                gg = 1;
        for (int k = 0; k < this.getBiologicalSequenceCount(); k++) {
            seq = this.getBiologicalSequenceAt(k);
            switch (seq.getSymbol(pos1InAlignment).getSymbol()) {
                case 'A' :
                    switch (seq.getSymbol(pos2InAlignment).getSymbol()) {
                        case 'A' :
                            A1++;
                            A2++;
                            AA++;
                            break;
                        case 'U' :
                            A1++;
                            U2++;
                            AU++;
                            break;
                        case 'G' :
                            A1++;
                            G2++;
                            AG++;
                            break;
                        case 'C' :
                            A1++;
                            C2++;
                            AC++;
                            break;
                        case '-' :
                            A1++;
                            gap2++;
                            Ag++;
                            break;
                    }
                    ;
                    break;
                case 'U' :
                    switch (seq.getSymbol(pos2InAlignment).getSymbol()) {
                        case 'A' :
                            U1++;
                            A2++;
                            UA++;
                            break;
                        case 'U' :
                            U1++;
                            U2++;
                            UU++;
                            break;
                        case 'G' :
                            U1++;
                            G2++;
                            UG++;
                            break;
                        case 'C' :
                            U1++;
                            C2++;
                            UC++;
                            break;
                        case '-' :
                            U1++;
                            gap2++;
                            Ug++;
                            break;
                    }
                    ;
                    break;
                case 'G' :
                    switch (seq.getSymbol(pos2InAlignment).getSymbol()) {
                        case 'A' :
                            G1++;
                            A2++;
                            GA++;
                            break;
                        case 'U' :
                            G1++;
                            U2++;
                            GU++;
                            break;
                        case 'G' :
                            G1++;
                            G2++;
                            GG++;
                            break;
                        case 'C' :
                            G1++;
                            C2++;
                            GC++;
                            break;
                        case '-' :
                            G1++;
                            gap2++;
                            Gg++;
                            break;
                    }
                    ;
                    break;
                case 'C' :
                    switch (seq.getSymbol(pos2InAlignment).getSymbol()) {
                        case 'A' :
                            C1++;
                            A2++;
                            CA++;
                            ;
                            break;
                        case 'U' :
                            C1++;
                            U2++;
                            CU++;
                            break;
                        case 'G' :
                            C1++;
                            G2++;
                            CG++;
                            break;
                        case 'C' :
                            C1++;
                            C2++;
                            CC++;
                            break;
                        case '-' :
                            C1++;
                            gap2++;
                            Cg++;
                            break;
                    }
                    ;
                    break;
                case '-' :
                    switch (seq.getSymbol(pos2InAlignment).getSymbol()) {
                        case 'A' :
                            gap1++;
                            A2++;
                            gA++;
                            ;
                            break;
                        case 'U' :
                            gap1++;
                            U2++;
                            gU++;
                            break;
                        case 'G' :
                            gap1++;
                            G2++;
                            gG++;
                            break;
                        case 'C' :
                            gap1++;
                            C2++;
                            gC++;
                            break;
                        case '-' :
                            gap1++;
                            gap2++;
                            gg++;
                            break;
                    }
                    ;
                    break;
            }
        }
        int t = this.getBiologicalSequenceCount();
        double result = Math.log10((AA/t)/((A1/t)*(A2/t)))
                +Math.log10((AU/t)/((A1/t)*(U2/t)))
                +Math.log10((AG/t)/((A1/t)*(G2/t)))
                +Math.log10((AC/t)/((A1/t)*(C2/t)))
                +Math.log10((GG/t)/((G1/t)*(G2/t)))
                +Math.log10((GU/t)/((G1/t)*(U2/t)))
                +Math.log10((GA/t)/((G1/t)*(A2/t)))
                +Math.log10((GC/t)/((G1/t)*(C2/t)))
                +Math.log10((UU/t)/((U1/t)*(U2/t)))
                +Math.log10((UA/t)/((U1/t)*(A2/t)))
                +Math.log10((UG/t)/((U1/t)*(G2/t)))
                +Math.log10((UC/t)/((U1/t)*(C2/t)))
                +Math.log10((CC/t)/((C1/t)*(C2/t)))
                +Math.log10((CU/t)/((C1/t)*(U2/t)))
                +Math.log10((CG/t)/((C1/t)*(G2/t)))
                +Math.log10((CA/t)/((C1/t)*(A2/t)))
                ;
        return result;

    }

    /**
     * Calculate and print the covariations between two positions within the alignment
     *
     * @param pos1
     * @param pos2
     * @return
     */
    public int[][] calculateCovariation(final int pos1, final int pos2) {
        int[][] m = null;
        if (pos1 != -1 && pos2 != -1) {
            BiologicalSequence seq = null;
            m = new int[6][6];
            for (int k = 0; k < this.getBiologicalSequenceCount(); k++) {
                seq = this.getBiologicalSequenceAt(k);
                //System.out.println(seq.getSymbol(this.pos1).getSymbol()+" "+seq.getSymbol(this.pos2).getSymbol());
                final char c2 = seq.getSymbol(pos2).getSymbol();
                switch (seq.getSymbol(pos1).getSymbol()) {
                    case 'A' :
                        switch (c2) {
                            case 'A' :
                                m[0][0]++;
                                break;
                            case 'U' :
                                m[0][1]++;
                                break;
                            case 'G' :
                                m[0][2]++;
                                break;
                            case 'C' :
                                m[0][3]++;
                                break;
                            case '-' :
                                m[0][4]++;
                                break;
                        }
                        ;
                        break;
                    case 'U' :
                        switch (c2) {
                            case 'A' :
                                m[1][0]++;
                                break;
                            case 'U' :
                                m[1][1]++;
                                break;
                            case 'G' :
                                m[1][2]++;
                                break;
                            case 'C' :
                                m[1][3]++;
                                break;
                            case '-' :
                                m[1][4]++;
                                break;
                        }
                        ;
                        break;
                    case 'G' :
                        switch (c2) {
                            case 'A' :
                                m[2][0]++;
                                break;
                            case 'U' :
                                m[2][1]++;
                                break;
                            case 'G' :
                                m[2][2]++;
                                break;
                            case 'C' :
                                m[2][3]++;
                                break;
                            case '-' :
                                m[2][4]++;
                                break;
                        }
                        ;
                        break;
                    case 'C' :
                        switch (c2) {
                            case 'A' :
                                m[3][0]++;
                                break;
                            case 'U' :
                                m[3][1]++;
                                break;
                            case 'G' :
                                m[3][2]++;
                                break;
                            case 'C' :
                                m[3][3]++;
                                break;
                            case '-' :
                                m[3][4]++;
                                break;
                        }
                        ;
                        break;
                    case '-' :
                        switch (c2) {
                            case 'A' :
                                m[4][0]++;
                                break;
                            case 'U' :
                                m[4][1]++;
                                break;
                            case 'G' :
                                m[4][2]++;
                                break;
                            case 'C' :
                                m[4][3]++;
                                break;
                            case '-' :
                                m[4][4]++;
                                break;
                        }
                        ;
                        break;
                }
            }
            for (int i = 0; i < 5; i++)
                for (int j = 0; j < 5; j++)
                    m[i][5] += m[i][j];
            for (int i = 0; i < 5; i++)
                for (int j = 0; j < 5; j++)
                    m[5][i] += m[j][i];
            for (int i = 0; i < 5; i++)
                m[5][5] += m[i][5];
            //we print it
            final DecimalFormat df = new DecimalFormat();
            df.setMaximumFractionDigits(2);
            System.out.println("Statistics between positions " + (pos1 + 1) + "-" + (pos2 + 1) + ":\n");
            System.out.println("\t\tPosition " + (pos2 + 1) + "\n");
            System.out.println("\tA\tU\tG\tC\t-\n");
            for (int i = 0; i < 6; i++) {
                switch (i) {
                    case 0 :
                        System.out.print("A\t");
                        break;
                    case 1 :
                        System.out.print("U\t");
                        break;
                    case 2 :
                        System.out.print("G\t");
                        break;
                    case 3 :
                        System.out.print("C\t");
                        break;
                    case 4 :
                        System.out.print("-\t");
                        break;
                    case 5 :
                        System.out.print(" \t");
                        break;
                }
                for (int j = 0; j < 6; j++)
                    System.out.print(m[i][j] + "\t");
                if (i == 2)
                    System.out.print("Position " + (pos1 + 1));
                if (i != 5) {
                    System.out.println();
                    System.out.print("\t");
                    for (int j = 0; j < 5; j++)
                        System.out.print("(" + df.format((float) m[5][j] * (float) m[i][5] / (float) m[5][5]) + ")\t");
                }
                System.out.println();
                System.out.println();
            }

        }
        return m;
    }

    /**
     * Change a sequence position in the multiple alignment
     */
    public void moveSequence(int newPos, BiologicalSequence seq) {
        this.biologicalSequences.remove(seq);
        this.biologicalSequences.add(newPos, seq);
    }

    public void increaseSize() {
        for (int i = 0; i < this.getSymbolSequenceCount(); i++)
            this.getLetterSequenceAt(i).increaseSize();
    }


    public void insertGaps(int index) {
        for (int i = 0; i < this.getSymbolSequenceCount(); i++)
            this.getLetterSequenceAt(i).insertGap(index);
    }

    public void removeBiologicalSequenceAt(int index) {
        this.biologicalSequences.remove(index);
    }

    public List<BiologicalSequence> getBiologicalSequences() {
        return new ArrayList<BiologicalSequence>(this.biologicalSequences);
    }

    /**
     * Calculate the conservation scores for all the secondary interactions stored in the SecondaryStructure
     * */
    public void updateConservationScores() {
        this.updateConservationScores(0,this.getLength()-1);
    }

    public void updateConservationScores(int start, int end) {
    }

    public void  updateStructuralAlignment() {
        //TODO update also the reference secondarystructure when the ability to change the reference structure and sequence will be implemented in the RNAlign GUI
        for (StructuralIdentity si:this.alignment.getStructuralIdentities(false))
            this.alignment.removeStructuralIdentity(si);
        for (Molecule m:this.alignment.getMolecules())
            m.removeFeature(this.alignment);
        //now the alignment annotates nothing, so it has been automatically removed by the PARADISE engine from its parent features (so from the secondary structure)
        //we we will re-add it 
        IdentityFeatureFactory factory = mediator.getS2s().createIdentityFeatureFactory();
        if (false) {
            //this version is very heavy. One Structural Identity for each Structural domain defined by the Structure2D3D object
            int previousEnd = -1;
            for (Structure2D3D.Domain d:this.structure2D3D.domains) {
                StructuralIdentity si = null;
                int start = this.structure2D3D.symbols.indexOf(d.getStart()),
                        end =  this.structure2D3D.symbols.indexOf(d.getEnd());
                if (start-1 != previousEnd) {
                    //additional domain before
                    for (BiologicalSequence s:this.biologicalSequences) {
                        Location molecularLocation = new Location(), gapsLocation = new Location();
                        int localPosition = 1;
                        for (int i=previousEnd+1; i<=start-1;i++) {
                            Symbol symbol = s.getSymbol(i);
                            if (symbol.isGap())
                                gapsLocation.add(localPosition);
                            else
                                molecularLocation.add(symbol.getPositionInSequence());
                            localPosition++;
                        }
                        if (!molecularLocation.isEmpty()) {
                            if (si == null) {
                                si = factory.createStructuralIdentity(this.alignment,"additional domain",s.getMolecule(),molecularLocation);
                            }
                            else
                                s.getMolecule().addFeature(si,molecularLocation);
                        }
                        if (si!= null && !gapsLocation.isEmpty())
                            si.addGapsLocation(s.getMolecule(),gapsLocation);
                    }
                }
                si = null;
                for (BiologicalSequence s:this.biologicalSequences) {
                    Location molecularLocation = new Location(), gapsLocation = new Location();
                    int localPosition = 1;
                    for (int i=start; i<=end;i++) {
                        Symbol symbol = s.getSymbol(i);
                        if (symbol.isGap())
                            gapsLocation.add(localPosition);
                        else
                            molecularLocation.add(symbol.getPositionInSequence());
                        localPosition++;
                    }
                    if (!molecularLocation.isEmpty()) {
                        if (si == null) {
                            si = factory.createStructuralIdentity(this.alignment,d.getStructuralDomain().getName(),s.getMolecule(),molecularLocation);
                        }
                        else
                            s.getMolecule().addFeature(si,molecularLocation);
                    }
                    if (si!= null && !gapsLocation.isEmpty())
                        si.addGapsLocation(s.getMolecule(),gapsLocation);
                }
                previousEnd = end;
            }
            if (previousEnd < this.getLength()) {
                StructuralIdentity si = null;
                for (BiologicalSequence s:this.biologicalSequences) {
                    Location molecularLocation = new Location(), gapsLocation = new Location();
                    int localPosition = 1;
                    for (int i=previousEnd+1; i<this.getLength();i++) {
                        Symbol symbol = s.getSymbol(i);
                        if (symbol.isGap())
                            gapsLocation.add(localPosition);
                        else
                            molecularLocation.add(symbol.getPositionInSequence());
                        localPosition++;
                    }
                    if (!molecularLocation.isEmpty()) {
                        if (si == null) {
                            si = factory.createStructuralIdentity(this.alignment,"additional domain",s.getMolecule(),molecularLocation);
                        }
                        else
                            s.getMolecule().addFeature(si,molecularLocation);
                    }
                    if (si!= null && !gapsLocation.isEmpty())
                        si.addGapsLocation(s.getMolecule(),gapsLocation);
                }
            }
        }
        //in this version, only one StructuralIdentity => the complete alignment
        StructuralIdentity si = null;
        for (BiologicalSequence s:this.biologicalSequences) {
            Location molecularLocation = new Location(), gapsLocation = new Location();
            for (int i=0; i<s.size();i++) {
                Symbol symbol = s.getSymbol(i);
                if (symbol.isGap()) {
                    gapsLocation.add(i+1);
                }
                else
                    molecularLocation.add(symbol.getPositionInSequence());
            }
            if (si == null)
                si = factory.createStructuralIdentity(this.alignment,"",s.getMolecule(),molecularLocation);
            else
                s.getMolecule().addFeature(si,molecularLocation);
            if (!gapsLocation.isEmpty())
                si.addGapsLocation(s.getMolecule(),gapsLocation);
            this.alignment.setPositionInAlignment(s.getMolecule(),this.biologicalSequences.indexOf(s));
        }
        //we re-make the link alignment-2D
        this.structure2D3D.referenceSecondaryStructure.linkStructuralAlignment(this.alignment);

    }

    public int indexOf(Molecule m) {
        return this.biologicalSequences.indexOf(m);
    }

    public class SequenceMeter extends SymbolSequence {

        private MeterSize size;

        public String getSequence() {
            return null;
        }

        private SequenceMeter(int length) {
            super();
            this.color = Color.DARK_GRAY;
            this.size = new MeterSize(length);
        }

        public Symbol getSymbol(final int index) {
            return null;
        }

        public void drawSymbol(final Alignment alignment, Graphics2D g, int currentX, int viewX, int currentY, int viewY, GraphicContext gc, int index) {
            g.setColor(Color.BLACK);
            if (index == 0 || (index + 1) % gc.getRatio() == 0) {
                String position = new StringBuffer().append(index + 1).toString();
                g.drawString(position, (float)currentX - (float)viewX-(float)(gc.getLetterWidth()*position.trim().length())/4f, (float)(currentY - gc.getLetterHeight()));
                g.drawString(new StringBuffer().append('|').toString(), currentX - viewX, currentY);
            } else
                g.drawString(new StringBuffer().append('.').toString(), (float) currentX - (float) viewX, (float) currentY);
            g.setColor(Color.BLACK);
        }


        public int size() {
            return this.size.getSize();
        }

        public void insertGap(final int index) {
            this.size.increaseSize();
        }

        public boolean removeGap(final int index) {
            if (this.isGap(index)) {
                this.size.decreaseSize();
                return true;
            }
            return false;
        }

        public void increaseSize() {
            this.size.increaseSize();
        }

        public boolean isGap(final int index) {
            return true;
        }

        public void setSelected(final boolean selected) {
            this.selected = false;
        }

        public void setSymbolAt(Symbol s, int i) {
            //does nothing
        }

        private class MeterSize {

            private int size;

            private MeterSize(final int size) {
                this.size = size;
            }

            private void setSize(final int size) {
                this.size = size;
            }

            private void increaseSize() {
                this.size++;
            }

            private void decreaseSize() {
                this.size--;
            }

            private int getSize() {
                return size;
            }
        }

    }
}
