package fr.unistra.ibmc.s2s.ssviewer.features;

import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.paradise.core.ParadiseID;
import fr.unistra.ibmc.paradise.core.RNA;
import fr.unistra.ibmc.paradise.core.Location;
import fr.unistra.ibmc.paradise.core.features.*;
import fr.unistra.ibmc.paradise.BiologicalSymbolException;
import fr.unistra.ibmc.paradise.ArchitectureException;
import fr.unistra.ibmc.s2s.ssviewer.graphics.GraphicContext;
import fr.unistra.ibmc.s2s.Mediator;

import java.awt.*;
import java.awt.geom.Area;
import java.awt.geom.Rectangle2D;
import java.util.*;
import java.util.List;

public class Model2D extends Rna2DElement {

    private List<Rna2D> rnas;
    private List<BaseBaseInteraction2D> interMolecularTertiaryInteractions;
    private List<Helix2D> interMolecularHelices;
    private SecondaryStructure secondaryStructure;

    public Model2D(final SecondaryStructure secondaryStructure,final SecondaryStructureDisplay display, final Mediator mediator, final GraphicContext gc) {
        //first step, construction of the central repository for all the Residue2D objects corresonding to this SecondaryStructureDisplay
        super(new Residues2DRepository(display, mediator,gc),mediator,gc);
        this.secondaryStructure = secondaryStructure;
        final Collection<Molecule> molecules = display.getMolecules();
        this.rnas = new ArrayList<Rna2D>(molecules.size());
        this.interMolecularTertiaryInteractions = new ArrayList<BaseBaseInteraction2D>();
        this.interMolecularHelices = new ArrayList<Helix2D>();
        Rna2D rna2D;
        for (Molecule m:secondaryStructure.getMolecules()) {
            rna2D = new Rna2D((RNA)m, secondaryStructure, this.repository, mediator, gc);
            this.rnas.add(rna2D);
        }
        for (BaseBaseInteraction inter:secondaryStructure.getTertiaryInteractions()) {
            if (!inter.isSecondaryInteraction() && inter.getMolecules().size() > 1)
                this.interMolecularTertiaryInteractions.add(BaseBaseInteraction2D.getInteraction(inter,this.repository, mediator,gc));
        }
        for (Helix h:secondaryStructure.getHelices()) {
            if (h.getMolecules().size() == 2) {
                this.interMolecularHelices.add(new Helix2D(h, this.repository, mediator,gc));
            }
        }
    }

    public void extendWithModel2D(Model2D m) throws BiologicalSymbolException{
        for (Helix2D h2D:m.getInterMolecularHelices()) {
            this.interMolecularHelices.add(h2D);
            this.gc.intializeHelixCurrentContext(h2D);
        }
        for (Rna2D r:m.getRnas()) {
            for (Helix2D h2D:r.getHelices())
                this.gc.intializeHelixCurrentContext(h2D);
            this.rnas.add(r);
        }
    }

    public SecondaryStructure getSecondaryStructure() {
        return this.secondaryStructure;
    }

    public List<Residue2D> getResidues() {
        List<Residue2D> residues = new ArrayList<Residue2D>();
        for (Rna2D r:this.rnas)
            residues.addAll(r.getResidues());
        return residues;
    }

    public void draw(final java.awt.Graphics2D g2) {
        g2.setColor(Color.BLACK);
        for (Rna2D r:this.rnas)
            r.draw(g2);
        for (BaseBaseInteraction2D i:this.interMolecularTertiaryInteractions)
            i.draw(g2);
        for (Helix2D h:this.interMolecularHelices)
            h.draw(g2);
    }

    public boolean contains(final double x, final double y) {
        return this.getArea().contains(x, y);
    }

    public Area getArea() {
        Rectangle2D surf = null;
        for (Helix2D h:this.interMolecularHelices) {
            if (!h.isDisplayed())
                continue;
            if (surf == null)
                surf = h.getArea().getBounds2D();
            else
                surf.add(h.getArea().getBounds2D());
        }
        Shape s = null;
        for (Rna2D r:this.rnas) {
            s = r.getArea();
            //an Rna can have a null surface if no intraHelix and no sstrands (only bases establishing inter-helix base pairs)
            if (s != null)
                if (surf == null)
                    surf = s.getBounds2D();
                else
                    surf.add(s.getBounds2D());
        }
        if(surf != null)
            return new Area(surf);
        else
            return new Area();
    }

    public String toSVG() {
        final StringBuffer svg = new StringBuffer("<title>");
        svg.append("Model n");
        svg.append("???");
        svg.append("</title>\n");
        for (Rna2D r: this.rnas)
            svg.append(r.toSVG());
        for (BaseBaseInteraction2D i:this.interMolecularTertiaryInteractions)
            svg.append(i.toSVG());
        for (Helix2D h:this.interMolecularHelices)
            svg.append(h.toSVG());
        return svg.toString();
    }

    public List<Helix2D> getInterMolecularHelices() {
        return new ArrayList<Helix2D>(this.interMolecularHelices);
    }

    public List<Rna2D> getRnas() {
        return new ArrayList<Rna2D>(this.rnas);
    }

    public List<BaseBaseInteraction2D> getInterMolecularTertiaryInteractions() {
        return new ArrayList<BaseBaseInteraction2D>(this.interMolecularTertiaryInteractions);
    }

    public List<BaseBaseInteraction2D> getAllTertiaryInteractions() {
        List<BaseBaseInteraction2D> interactions = new ArrayList<BaseBaseInteraction2D>();
        interactions.addAll(this.interMolecularTertiaryInteractions);
        for (Helix2D h:this.interMolecularHelices)
            for (BaseBaseInteraction2D inter : h.getSecondaryInteractions())
                interactions.add(inter);
        for (Rna2D r:this.rnas) {
            for (BaseBaseInteraction2D inter : r.getTertiaryInteractions())
                interactions.add(inter);
        }
        return interactions;
    }

    /**
     * return all the interaction established by two residues
     * @return
     */
    public List<Interaction2D> getInteractions(ParadiseID moleculeResidueId, int residueAbsolutePosition,  ParadiseID moleculePartnerResidueId, int partnerResidueAbsolutePosition) {
        List<Interaction2D> hits = new ArrayList<Interaction2D>();
        for (Interaction2D inter:this.getBaseBaseInteractions())
            if (inter.getResidue().getAbsolutePosition() == residueAbsolutePosition && inter.getResidue().getMoleculeId().equals(moleculeResidueId) && inter.getPartnerResidue().getAbsolutePosition() == partnerResidueAbsolutePosition && inter.getPartnerResidue().getMoleculeId() == moleculePartnerResidueId)
                hits.add(inter);
        return hits;
    }

    /**
     * Return a Residue2D object given the indexes for the residue wrapped and the RNA molecule containing
     * this residue
     *
     * @param moleculeID  the rna paradise ID
     * @param absolutePosition the base's position in the rna
     * @return
     */
    public Residue2D getResidue(final ParadiseID moleculeID, final int absolutePosition) {
        return this.repository.getResidue(moleculeID, absolutePosition);
    }

    public List<Helix2D> getHelices() {
        List<Helix2D> helices = new ArrayList<Helix2D>();
        helices.addAll(this.interMolecularHelices);
        for (Rna2D r:this.rnas)
            helices.addAll(r.getHelices());
        return helices;
    }

    public List<Sstrand2D> getSstrands() {
        List<Sstrand2D> sstrands = new ArrayList<Sstrand2D>();
        for (Rna2D rna:this.getRnas())
            sstrands.addAll(rna.getStrands());
        return sstrands;
    }

    /**
     * return all the interactions established by one residue
     *
     * @param moleculeId
     * @param absolutePosition
     * @return
     */
    public List<Interaction2D> getInteractions(final ParadiseID moleculeId, final int absolutePosition) {
        List<Interaction2D> hits = new ArrayList<Interaction2D>();
        for(Interaction2D inter:this.getBaseBaseInteractions())
            if ((inter.getResidue().getAbsolutePosition() == absolutePosition && inter.getResidue().getMoleculeId() == moleculeId) ||
                    (inter.getPartnerResidue().getAbsolutePosition() == absolutePosition && inter.getPartnerResidue().getMoleculeId() == moleculeId))
                hits.add(inter);
        return hits;
    }

    public List<BaseBaseInteraction2D> getBaseBaseInteractions() {
        List<BaseBaseInteraction2D> interactions = new ArrayList<BaseBaseInteraction2D>();
        interactions.addAll(this.interMolecularTertiaryInteractions);
        for (Helix2D h:this.interMolecularHelices)
            for (BaseBaseInteraction2D inter : h.getSecondaryInteractions())
                interactions.add(inter);
        for (Rna2D r:this.rnas) {
            for (BaseBaseInteraction2D inter : r.getTertiaryInteractions())
                interactions.add(inter);
            for (Helix2D h : r.getHelices())
                for (BaseBaseInteraction2D inter : h.getSecondaryInteractions())
                    interactions.add(inter);
        }
        return interactions;
    }

    private boolean alreadyExist(Residue2D r1, Residue2D r2) {
        for (BaseBaseInteraction2D interaction:this.getBaseBaseInteractions()) {
            if (interaction.getResidue() == r1 && interaction.getPartnerResidue() == r2 ||
                    interaction.getResidue() == r2 && interaction.getPartnerResidue() == r1)
                return true;
        }
        return false;
    }

    public void modifyInteractionType (BaseBaseInteraction2D interactionToModify, Residue2D residue1, char edge1, Residue2D residue2, char edge2, char orientation, int dominant) {
        try {
            BaseBaseInteraction newInteraction;
            BaseBaseInteraction2D newInteraction2D;
            Iterator<Molecule> iterator = interactionToModify.getBaseBaseInteractionFeature().getMolecules().iterator();
            if (interactionToModify.getBaseBaseInteractionFeature().isSecondaryInteraction()) {
                Molecule m1 =  iterator.next();
                if (iterator.hasNext())  {
                    Molecule m2 = iterator.next();
                    INTERHELIX:for (Helix2D helix:this.interMolecularHelices) {
                        for (BaseBaseInteraction2D _interaction:new ArrayList<BaseBaseInteraction2D>(helix.secondaryInteractions)) {
                            if (_interaction == interactionToModify) {
                                helix.secondaryInteractions.remove(interactionToModify);
                                this.mediator.removeNodeFromExplorer(helix.getHelixFeature(),interactionToModify.getBaseBaseInteractionFeature());
                                //with the creation of the new interaction, the old one is automatically deleted (see the fr.unistra.ibmc.paradise.core.features.Helix class). That's why it was necessary to remove the old one before to create the new one
                                mediator.getAlignmentCanvas().removeBaseBaseInteraction(interactionToModify.getBaseBaseInteractionFeature());
                                newInteraction = interactionToModify.getBaseBaseInteractionFeature().getFactory().createBaseBaseInteraction(edge1,edge2,orientation,dominant, interactionToModify.getBaseBaseInteractionFeature().getParentFeatures(Helix.class).get(0), m1, interactionToModify.getBaseBaseInteractionFeature().getFullLocation(m1), m2, interactionToModify.getBaseBaseInteractionFeature().getFullLocation(m2));
                                newInteraction.generateAtomAtomInteractions();
                                mediator.getAlignmentCanvas().addBaseBaseInteraction(newInteraction);
                                newInteraction2D = BaseBaseInteraction2D.getInteraction(newInteraction,this.repository, this.mediator,this.gc);
                                helix.secondaryInteractions.add(newInteraction2D);
                                this.mediator.insertNodeIntoExplorer(helix.getHelixFeature(),newInteraction);
                                break INTERHELIX;
                            }
                        }
                    }
                }
                else {
                    RNAS:for (Rna2D rna:this.rnas) {
                        for (Helix2D helix:rna.helices)
                            for (BaseBaseInteraction2D _interaction:new ArrayList<BaseBaseInteraction2D>(helix.secondaryInteractions)) {
                                if (_interaction == interactionToModify) {
                                    helix.secondaryInteractions.remove(interactionToModify);
                                    this.mediator.removeNodeFromExplorer(helix.getHelixFeature(),interactionToModify.getBaseBaseInteractionFeature());
                                    //with the creation of the new interaction, the old one is automatically deleted (see the fr.unistra.ibmc.paradise.core.features.Helix class). That's why it was necessary to remove the old one before to create the new one
                                    mediator.getAlignmentCanvas().removeBaseBaseInteraction(interactionToModify.getBaseBaseInteractionFeature());
                                    newInteraction = interactionToModify.getBaseBaseInteractionFeature().getFactory().createBaseBaseInteraction(edge1,edge2,orientation,dominant, interactionToModify.getBaseBaseInteractionFeature().getParentFeatures(Helix.class).get(0), m1, interactionToModify.getBaseBaseInteractionFeature().getFullLocation(m1));
                                    newInteraction.generateAtomAtomInteractions();
                                    mediator.getAlignmentCanvas().addBaseBaseInteraction(newInteraction);
                                    newInteraction2D = BaseBaseInteraction2D.getInteraction(newInteraction,this.repository, this.mediator,this.gc);
                                    helix.secondaryInteractions.add(newInteraction2D);
                                    this.mediator.insertNodeIntoExplorer(helix.getHelixFeature(),newInteraction);
                                    break RNAS;
                                }
                            }
                    }
                }

            }
            else { //tertiary interaction
                Molecule m1 =  iterator.next();
                if (iterator.hasNext())  {
                    Molecule m2 = iterator.next();

                    for (BaseBaseInteraction2D interaction:new ArrayList<BaseBaseInteraction2D>(this.interMolecularTertiaryInteractions)) {
                        if (interaction == interactionToModify) {
                            this.interMolecularTertiaryInteractions.remove(interactionToModify);
                            this.mediator.removeNodeFromExplorer(this.getSecondaryStructure(),interactionToModify.getBaseBaseInteractionFeature());
                            //with the creation of the new interaction, the old one is automatically deleted (see the fr.unistra.ibmc.paradise.core.features.Helix class). That's why it was necessary to remove the old one before to create the new one
                            mediator.getAlignmentCanvas().removeBaseBaseInteraction(interactionToModify.getBaseBaseInteractionFeature());
                            newInteraction = interactionToModify.getBaseBaseInteractionFeature().getFactory().createBaseBaseInteraction(edge1,edge2,orientation,dominant, interactionToModify.getBaseBaseInteractionFeature().getParentFeatures(SecondaryStructure.class).get(0), m1, interactionToModify.getBaseBaseInteractionFeature().getFullLocation(m1), m2, interactionToModify.getBaseBaseInteractionFeature().getFullLocation(m2));
                            newInteraction.generateAtomAtomInteractions();
                            mediator.getAlignmentCanvas().addBaseBaseInteraction(newInteraction);
                            newInteraction2D = BaseBaseInteraction2D.getInteraction(newInteraction,this.repository, this.mediator,this.gc);
                            this.interMolecularTertiaryInteractions.add(newInteraction2D);
                            this.mediator.insertNodeIntoExplorer(this.getSecondaryStructure(),newInteraction);
                            break;
                        }
                    }
                }
                else {
                    RNAS:for (Rna2D rna:this.rnas)
                        for (BaseBaseInteraction2D interaction:new ArrayList<BaseBaseInteraction2D>(rna.tertiaryInteractions))
                            if (interaction == interactionToModify)  {
                                rna.tertiaryInteractions.remove(interactionToModify);
                                this.mediator.removeNodeFromExplorer(rna.getRnaMolecule(),interactionToModify.getBaseBaseInteractionFeature());
                                //with the creation of the new interaction, the old one is automatically deleted (see the fr.unistra.ibmc.paradise.core.features.Helix class). That's why it was necessary to remove the old one before to create the new one
                                mediator.getAlignmentCanvas().removeBaseBaseInteraction(interactionToModify.getBaseBaseInteractionFeature());
                                newInteraction = interactionToModify.getBaseBaseInteractionFeature().getFactory().createBaseBaseInteraction(edge1,edge2,orientation,dominant, interactionToModify.getBaseBaseInteractionFeature().getParentFeatures(SecondaryStructure.class).get(0), m1, interactionToModify.getBaseBaseInteractionFeature().getFullLocation(m1));
                                newInteraction.generateAtomAtomInteractions();
                                mediator.getAlignmentCanvas().addBaseBaseInteraction(newInteraction);
                                newInteraction2D = BaseBaseInteraction2D.getInteraction(newInteraction,this.repository, this.mediator,this.gc);
                                rna.tertiaryInteractions.add(newInteraction2D);
                                this.mediator.insertNodeIntoExplorer(rna.getRnaMolecule(),newInteraction);
                                break RNAS;
                            }
                }
            }
        }
        catch(fr.unistra.ibmc.paradise.BiologicalSymbolException bioe) {
            bioe.printStackTrace();
        }
    }

    public void createHelix(Residue2D firstFivePrime, Residue2D firstThreePrime, Residue2D secondFivePrime, Residue2D secondThreePrime) throws fr.unistra.ibmc.paradise.ArchitectureException, BiologicalSymbolException {

        Sstrand2D firstSstrandToRemove,  secondSstrandToRemove;
        Helix2D h2D = null;
        Set<Helix2D> helicesToRemove = new HashSet<Helix2D>();
        Residue2D nextResidue = firstFivePrime;

        //test is the helix to be created is ok
        Location newL1  = new Location(firstFivePrime.getAbsolutePosition(), firstThreePrime.getAbsolutePosition()),
                newL2  = new Location(secondFivePrime.getAbsolutePosition(), secondThreePrime.getAbsolutePosition());

        if (newL1.getLength() != newL2.getLength())
            throw new ArchitectureException("The two strands need to have the same length");

        if (newL1.getLength() < 2 )
            throw new ArchitectureException("You cannot construct an helix with strands shorter than two residues");

        //if this point is reached, everything ok

        //first step, we check if some helices overlap the upcoming one. If so, we store them to delete them
        while (!firstThreePrime.equals(nextResidue)) {
            if (Helix2D.class.isInstance(nextResidue.getStructuralDomain()))
                helicesToRemove.add((Helix2D)nextResidue.getStructuralDomain());
            nextResidue = nextResidue.getNextResidue();
        }

        nextResidue = secondFivePrime;

        while (!secondThreePrime.equals(nextResidue)) {
            if (Helix2D.class.isInstance(nextResidue.getStructuralDomain()))
                helicesToRemove.add((Helix2D)nextResidue.getStructuralDomain());
            nextResidue = nextResidue.getNextResidue();
        }

        //we remove the overlapping helices and we keep the secondary interactions as tertiary ones.
        for (Helix2D h:helicesToRemove)
            this.removeHelix(h,true);

        firstSstrandToRemove = (Sstrand2D)firstFivePrime.getStructuralDomain();
        secondSstrandToRemove = (Sstrand2D)secondFivePrime.getStructuralDomain();

        Location    previousL1 = firstSstrandToRemove.getSingleStrandFeature().getFullLocation(firstSstrandToRemove.getRna().getRnaMolecule()),
                previousL2 = secondSstrandToRemove.getSingleStrandFeature().getFullLocation(secondSstrandToRemove.getRna().getRnaMolecule());

        //the single-strands are deleted
        firstSstrandToRemove.getRna().removeSstrand(firstSstrandToRemove);
        this.mediator.removeNodeFromExplorer(firstSstrandToRemove.getRna().getRnaMolecule(),firstSstrandToRemove.getSingleStrandFeature());
        this.secondaryStructure.removeSingleStrand(firstSstrandToRemove.getSingleStrandFeature());
        //next step only if two strands of the new helix don't come from the same sstrand...
        if (firstSstrandToRemove != secondSstrandToRemove)  {
            secondSstrandToRemove.getRna().removeSstrand(secondSstrandToRemove);
            this.mediator.removeNodeFromExplorer(secondSstrandToRemove.getRna().getRnaMolecule(),secondSstrandToRemove.getSingleStrandFeature());
            this.secondaryStructure.removeSingleStrand(secondSstrandToRemove.getSingleStrandFeature());
        }

        //the new helix is created
        Helix h = null;
        if (firstSstrandToRemove.getRna() == secondSstrandToRemove.getRna()) { //intra-helix
            h = this.secondaryStructure.getFactory().createHelix("Helix", this.secondaryStructure,firstSstrandToRemove.getRna().getRnaMolecule(), new Location(newL1,newL2));
            h2D = new Helix2D(h,repository,mediator, gc);
            firstSstrandToRemove.getRna().helices.add(h2D);
            this.mediator.insertNodeIntoExplorer(firstSstrandToRemove.getRna().getRnaMolecule(),h);
        }
        else { //inter-helix
            h = this.secondaryStructure.getFactory().createHelix("Helix", this.secondaryStructure,firstSstrandToRemove.getRna().getRnaMolecule(), newL1, secondSstrandToRemove.getRna().getRnaMolecule(), newL2);
            h2D = new Helix2D(h,repository,mediator, gc);
            this.interMolecularHelices.add(h2D);
            this.mediator.insertNodeIntoExplorer(this.secondaryStructure,h);
        }
        for (BaseBaseInteraction interaction:h.getBaseBaseInteractions()) {
            this.mediator.insertNodeIntoExplorer(h,interaction);
            this.mediator.getAlignmentCanvas().addBaseBaseInteraction(interaction);
        }
        for (BaseBaseInteraction2D interaction2D:h2D.getSecondaryInteractions()) {
            //if one of the tertiary interactions equals the secondary, the default type of this secondary is updated
            for (BaseBaseInteraction2D tertiary:this.getAllTertiaryInteractions()) {
                if (tertiary.getResidue().equals(interaction2D.getResidue()) && tertiary.getPartnerResidue().equals(interaction2D.getPartnerResidue())) {
                    this.modifyInteractionType(interaction2D,
                            interaction2D.getResidue(),
                            tertiary.getBaseBaseInteractionFeature().getEdge1(),
                            interaction2D.getPartnerResidue(),
                            tertiary.getBaseBaseInteractionFeature().getEdge2(),
                            tertiary.getBaseBaseInteractionFeature().getOrientation(),
                            tertiary.getBaseBaseInteractionFeature().getDominant());
                    //the tertiary is deleted
                    this.removeTertiaryInteraction(tertiary);
                    break;
                }
                else if (tertiary.getPartnerResidue().equals(interaction2D.getResidue()) && tertiary.getResidue().equals(interaction2D.getPartnerResidue())) {
                    this.modifyInteractionType(interaction2D,
                            interaction2D.getResidue(),
                            tertiary.getBaseBaseInteractionFeature().getEdge2(),
                            interaction2D.getPartnerResidue(),
                            tertiary.getBaseBaseInteractionFeature().getEdge1(),
                            tertiary.getBaseBaseInteractionFeature().getOrientation(),
                            tertiary.getBaseBaseInteractionFeature().getDominant());
                    //the tertiary is deleted
                    this.removeTertiaryInteraction(tertiary);
                    break;
                }

            }
        }
        //the new sstrands (if any) are created
        if (firstSstrandToRemove != secondSstrandToRemove)
            previousL1.remove(newL1);
        else {
            //if the two strands of the new helix come from the same sstrand...
            previousL1.remove(newL1);
            previousL1.remove(newL2);
        }
        if (!previousL1.isEmpty()) {
            int[] boundaries = previousL1.getBoundaries();
            for (int i=0; i<boundaries.length-1;i+=2) {
                SingleStrand ss = this.secondaryStructure.getFactory().createSingleStrand("Single-Strand",this.secondaryStructure, firstSstrandToRemove.getRna().getRnaMolecule(), new Location(boundaries[i], boundaries[i+1]));
                Sstrand2D ss2D = new Sstrand2D(firstSstrandToRemove.getRna(),this.secondaryStructure,ss,repository,mediator,gc);
                ss2D.setCoordinates();
                firstSstrandToRemove.getRna().strands.add(ss2D);
                this.mediator.insertNodeIntoExplorer(firstSstrandToRemove.getRna().getRnaMolecule(),ss);
            }
        }
        //next step only if two strands of the new helix don't come from the same sstrand...
        if (firstSstrandToRemove != secondSstrandToRemove) {
            previousL2.remove(newL2);
            if (!previousL2.isEmpty()) {
                int[] boundaries = previousL2.getBoundaries();
                for (int i=0; i<boundaries.length-1;i+=2) {
                    SingleStrand ss = this.secondaryStructure.getFactory().createSingleStrand("Single-Strand",this.secondaryStructure,  secondSstrandToRemove.getRna().getRnaMolecule(), new Location(boundaries[i], boundaries[i+1]));
                    Sstrand2D ss2D = new Sstrand2D( secondSstrandToRemove.getRna(),this.secondaryStructure,ss,repository,mediator,gc);
                    ss2D.setCoordinates();
                    secondSstrandToRemove.getRna().strands.add(ss2D);
                    this.mediator.insertNodeIntoExplorer( secondSstrandToRemove.getRna().getRnaMolecule(),ss);
                }
            }
        }

        //the four residues of the new helix define, by default, a rectangular shape.
        //the point to place the new helix is the center between the current location of the 4 points
        int     x = (int)(firstFivePrime.getRealX()+firstThreePrime.getRealX()+secondFivePrime.getRealX()+secondThreePrime.getRealX())/4,
                y = (int)(firstFivePrime.getRealY()+firstThreePrime.getRealY()+secondFivePrime.getRealY()+secondThreePrime.getRealY())/4;
        firstFivePrime.setRealCoordinates(x,y);
        firstThreePrime.setRealCoordinates(x,y-gc.getHeight()*(newL1.getLength()-1));
        secondThreePrime.setRealCoordinates(x+gc.getWidth() *4,y);
        secondFivePrime.setRealCoordinates(x+gc.getWidth() *4,y-gc.getHeight()*(newL1.getLength()-1));
        h2D.setCoordinates();
    }

    public void removeHelix(Helix2D h,boolean keepInteractionsAsTertiary) throws BiologicalSymbolException {
        Rna2D rna = null;
        Molecule m = null;
        //extend or create single strands with the residues of this helix
        for (Residue2D r:h.get5PrimeEnds()) {
            m = r.getResidue2DFeature().getResidues().iterator().next().getMolecule();
            for (Rna2D _rna:this.getRnas())
                if (_rna.getRnaMolecule().equals(m)) {
                    rna = _rna;
                    break;
                }
            Residue2D previousResidue = r.getPreviousResidue(), nextResidue = h.get3PrimeEnd(r).getNextResidue();
            if (previousResidue != null) {
                if (nextResidue == null) {
                    if (Helix2D.class.isInstance(previousResidue.getStructuralDomain())) {
                        //its a new terminal single-strand
                        SingleStrand ss = this.secondaryStructure.getFactory().createSingleStrand("Single-Strand",this.secondaryStructure, m, new Location(r.getAbsolutePosition(), m.getLength()));
                        Sstrand2D ss2D = new Sstrand2D(rna,this.secondaryStructure,ss,repository,mediator,gc);
                        ss2D.setCoordinates();
                        rna.strands.add(ss2D);
                        this.mediator.insertNodeIntoExplorer(rna.getRnaMolecule(),ss);
                    }
                    else {
                        //the single-strand is extended
                        ((Sstrand2D)previousResidue.getStructuralDomain()).extend(new Location(r.getAbsolutePosition(),m.getLength()));
                        previousResidue.getStructuralDomain().setCoordinates();
                    }
                }
                else {
                    if (Helix2D.class.isInstance(previousResidue.getStructuralDomain()) && Helix2D.class.isInstance(nextResidue.getStructuralDomain())) {
                        //System.out.println("TEST1");
                        //its a new single-strand
                        SingleStrand ss = this.secondaryStructure.getFactory().createSingleStrand("Single-Strand",this.secondaryStructure, m, new Location(r.getAbsolutePosition(), nextResidue.getPreviousResidue().getAbsolutePosition()));
                        Sstrand2D ss2D = new Sstrand2D(rna,this.secondaryStructure,ss,repository,mediator,gc);
                        ss2D.setCoordinates();
                        rna.strands.add(ss2D);
                        this.mediator.insertNodeIntoExplorer(rna.getRnaMolecule(),ss);
                    }
                    else if (Helix2D.class.isInstance(previousResidue.getStructuralDomain()) && Sstrand2D.class.isInstance(nextResidue.getStructuralDomain())) {
                        //System.out.println("TEST2");

                        //the single-strand is extended
                        ((Sstrand2D)nextResidue.getStructuralDomain()).extend(new Location(r.getAbsolutePosition(),nextResidue.getPreviousResidue().getAbsolutePosition()));
                        nextResidue.getStructuralDomain().setCoordinates();
                    }
                    else if (Sstrand2D.class.isInstance(previousResidue.getStructuralDomain()) && Helix2D.class.isInstance(nextResidue.getStructuralDomain())) {
                        //System.out.println("TEST3");

                        //the single-strand is extended
                        ((Sstrand2D)previousResidue.getStructuralDomain()).extend(new Location(r.getAbsolutePosition(),nextResidue.getPreviousResidue().getAbsolutePosition()));
                        previousResidue.getStructuralDomain().setCoordinates();
                    }
                    else {
                        //System.out.println("TEST4");

                        Residue2D lastSingleStrandPosition = ((Sstrand2D)nextResidue.getStructuralDomain()).getBase3();
                        Sstrand2D ssToRemove = (Sstrand2D)nextResidue.getStructuralDomain();

                        //one single-strand is extended, the other one is deleted
                        ((Sstrand2D)previousResidue.getStructuralDomain()).extend(new Location(r.getAbsolutePosition(),lastSingleStrandPosition.getAbsolutePosition()));
                        previousResidue.getStructuralDomain().setCoordinates();

                        rna.removeSstrand(ssToRemove);
                        this.mediator.removeNodeFromExplorer(rna.getRnaMolecule(),ssToRemove.getSingleStrandFeature());
                        this.secondaryStructure.removeSingleStrand(ssToRemove.getSingleStrandFeature());
                    }
                }
            }
            else {
                if (nextResidue == null) {
                    //its a new single-strand
                    SingleStrand ss = this.secondaryStructure.getFactory().createSingleStrand("Single-Strand",this.secondaryStructure, m, new Location(r.getAbsolutePosition(), m.getLength()));
                    Sstrand2D ss2D = new Sstrand2D(rna,this.secondaryStructure,ss,repository,mediator,gc);
                    ss2D.setCoordinates();
                    rna.strands.add(ss2D);
                    this.mediator.insertNodeIntoExplorer(rna.getRnaMolecule(),ss);

                }
                else {
                    if (Helix2D.class.isInstance(nextResidue.getStructuralDomain())) {
                        //its a new terminal single-strand
                        SingleStrand ss = this.secondaryStructure.getFactory().createSingleStrand("Single-Strand",this.secondaryStructure, m, new Location(r.getAbsolutePosition(), nextResidue.getPreviousResidue().getAbsolutePosition()));
                        Sstrand2D ss2D = new Sstrand2D(rna,this.secondaryStructure,ss,repository,mediator,gc);
                        ss2D.setCoordinates();
                        rna.strands.add(ss2D);
                        this.mediator.insertNodeIntoExplorer(rna.getRnaMolecule(),ss);
                    }
                    else {
                        //the single-strand is extended
                        ((Sstrand2D)nextResidue.getStructuralDomain()).extend(new Location(r.getAbsolutePosition(),nextResidue.getAbsolutePosition()-1));
                        nextResidue.getStructuralDomain().setCoordinates();
                    }
                }
            }
        }

        for (BaseBaseInteraction2D _interaction:h.getSecondaryInteractions()) {
            this.mediator.getAlignmentCanvas().removeBaseBaseInteraction(_interaction.getBaseBaseInteractionFeature());
            _interaction.getResidue().removeInteraction(_interaction);
            _interaction.getPartnerResidue().removeInteraction(_interaction);
            m.removeFeature(_interaction.getBaseBaseInteractionFeature());
            if (keepInteractionsAsTertiary)
                this.createTertiaryInteraction(_interaction.getResidue(), _interaction.getBaseBaseInteractionFeature().getEdge1(),
                        _interaction.getPartnerResidue(),_interaction.getBaseBaseInteractionFeature().getEdge2(),_interaction.getBaseBaseInteractionFeature().getOrientation(),_interaction.getBaseBaseInteractionFeature().getDominant());
            //for each residue the new structural domain has been registered above, with the creation/extension of sstrands
        }
        boolean isInterHelix = (h.getHelixFeature().getMolecules().size() == 2);
        if (isInterHelix) {
            this.interMolecularHelices.remove(h);
            this.mediator.removeNodeFromExplorer(this.secondaryStructure,h.getHelixFeature());
        }
        else {
            rna.helices.remove(h);
            this.mediator.removeNodeFromExplorer(rna.getRnaMolecule(),h.getHelixFeature());
        }
        this.secondaryStructure.removeHelix(h.getHelixFeature());
    }

    public void createTertiaryInteraction(Residue2D residue1, char edge1, Residue2D residue2, char edge2, char orientation, int dominant) {
        Molecule m1 = residue1.getResidue2DFeature().getMolecules().iterator().next(),
                m2 = residue2.getResidue2DFeature().getMolecules().iterator().next();
        int pos1 = residue1.getAbsolutePosition(), pos2 = residue2.getAbsolutePosition();
        try {
            BaseBaseInteraction2D interaction = null;
            BaseBaseInteraction bbi;
            if(m1 == m2) {
                if (pos1 > pos2) {
                    int tmp = pos2;
                    pos2 = pos1;
                    pos1 = tmp;
                    char _tmp = edge2;
                    edge2 = edge1;
                    edge1 = _tmp;
                }
                bbi = this.getSecondaryStructure().getFactory().createBaseBaseInteraction(
                        edge1,
                        edge2,
                        orientation,
                        dominant,
                        this.getSecondaryStructure(),
                        m1,
                        new Location(new Location(pos1), new Location(pos2)));

                for (Rna2D rna:this.getRnas())
                    if (rna.getRnaMolecule().getParadiseID().equals(m1.getParadiseID())) {
                        interaction = BaseBaseInteraction2D.getInteraction(bbi,this.repository, this.mediator,this.gc);
                        rna.addTertiaryInteraction(interaction);
                        this.mediator.insertNodeIntoExplorer(rna.getRnaMolecule(),bbi);
                        this.mediator.getAlignmentCanvas().addBaseBaseInteraction(bbi);
                        break;
                    }
            }
            else {
                bbi = this.getSecondaryStructure().getFactory().createBaseBaseInteraction(
                        edge1,
                        edge2,
                        orientation,
                        dominant,
                        this.getSecondaryStructure(),
                        m1,
                        new Location(pos1),
                        m2,
                        new Location(pos2));
                interaction = BaseBaseInteraction2D.getInteraction(bbi,this.repository, this.mediator,this.gc);
                this.mediator.insertNodeIntoExplorer(this.getSecondaryStructure(),bbi);
                this.mediator.getAlignmentCanvas().addBaseBaseInteraction(bbi);
                this.interMolecularTertiaryInteractions.add(interaction);
            }
            bbi.generateAtomAtomInteractions();
        }
        catch(BiologicalSymbolException bioe) {
            bioe.printStackTrace();
        }
    }

    public void removeTertiaryInteraction(BaseBaseInteraction2D interaction) {
        if (!interaction.getBaseBaseInteractionFeature().isSecondaryInteraction()) {
            for (BaseBaseInteraction2D _interaction:this.getInterMolecularTertiaryInteractions())
                if (_interaction == interaction) {
                    this.mediator.getAlignmentCanvas().removeBaseBaseInteraction(_interaction.getBaseBaseInteractionFeature());
                    this.interMolecularTertiaryInteractions.remove(interaction);
                    this.mediator.removeNodeFromExplorer(this.getSecondaryStructure(),interaction.getBaseBaseInteractionFeature());
                    break;
                }
            for (Rna2D rna:this.getRnas())
                for (BaseBaseInteraction2D _interaction:rna.getTertiaryInteractions())
                    if (_interaction == interaction) {
                        this.mediator.getAlignmentCanvas().removeBaseBaseInteraction(_interaction.getBaseBaseInteractionFeature());
                        rna.removeTertiaryInteraction(interaction);
                        this.mediator.removeNodeFromExplorer(rna.getRnaMolecule(),interaction.getBaseBaseInteractionFeature());
                        break;
                    }
            interaction.getResidue().removeInteraction(interaction);
            interaction.getPartnerResidue().removeInteraction(interaction);
            this.secondaryStructure.removeBaseBaseInteraction(interaction.getBaseBaseInteractionFeature());
            for (Molecule m:interaction.getBaseBaseInteractionFeature().getMolecules())
                m.removeFeature(interaction.getBaseBaseInteractionFeature());
        }
    }

}
