package fr.unistra.ibmc.s2s.ssviewer.features;

import fr.unistra.ibmc.paradise.core.ParadiseID;
import fr.unistra.ibmc.paradise.core.Molecule;
import fr.unistra.ibmc.s2s.Mediator;
import fr.unistra.ibmc.s2s.rnalign.graphics.AlignmentCanvas;
import fr.unistra.ibmc.s2s.ssviewer.graphics.GraphicContext;
import fr.unistra.ibmc.s2s.ssviewer.utils.DrawingUtils;

import java.awt.*;
import java.awt.geom.Area;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * A Residue2D objet registers only real coordinates but can return real ones or the coordinates
 * used in the current display (whose properties are defined within the GraphicContext object)
 *
 * @author fjossinet
 */
public class Residue2D extends Rna2DElement implements Comparable {

    private boolean updated, isDisplayed = true, selected;
    private List<Interaction2D> interactions;
    private PhosphodiesterBond2D phosphodiesterBond2D;
    private StructuralDomain2D structuralDomain;
    private fr.unistra.ibmc.paradise.core.features.Residue2D residue2D;

    Residue2D(final fr.unistra.ibmc.paradise.core.features.Residue2D residue2D, final Residues2DRepository repository, final Mediator mediator, final GraphicContext gc) {
        super(repository, mediator, gc);
        this.residue2D = residue2D;
        this.updated = true;
        this.interactions = new ArrayList<Interaction2D>();
    }

    public void isSelected(final boolean selected) {
        this.selected = selected;
    }

    public boolean isSelected() {
        return this.selected;
    }

    public fr.unistra.ibmc.paradise.core.features.Residue2D getResidue2DFeature() {
        return residue2D;
    }

    public char getSymbol() {
        return this.residue2D.getResidues().iterator().next().getSymbol();
    }

    public int getAbsolutePosition() {
        return this.residue2D.getFullLocation(this.residue2D.getMolecules().iterator().next()).getStart();
    }

    public double getRealX() {
        return this.residue2D.getX();
    }

    public double getRealY() {
        return this.residue2D.getY();
    }

    public double getCurrentX() {
        return this.residue2D.getX() * this.gc.getFinalZoomLevel() + this.gc.getViewX();
    }

    public double getCurrentY() {
        return this.residue2D.getY() * this.gc.getFinalZoomLevel() + this.gc.getViewY();
    }

    public Point2D getRealCoordinates() {
        return new Point2D.Double(this.getRealX(), this.getRealY());
    }

    public Point2D getCurrentCoordinates() {
        return new Point2D.Double(this.getCurrentX(), this.getCurrentY());
    }

    void setRealCoordinates(final java.awt.geom.Point2D coord) {
        this.setRealCoordinates(coord.getX(), coord.getY());
    }

    public void setRealCoordinates(final double x, final double y) {
        this.setX(x);
        this.setY(y);
    }

    void setX(final double x) {
        this.residue2D.setX((float)x);
        this.isUpdated(true);
    }

    void setY(final double y) {
        this.residue2D.setY((float)y);
        this.isUpdated(true);
    }

    public Residue2D getNextResidue() {
        return this.repository.getResidue(this.residue2D.getResidues().iterator().next().getNextResidue());
    }

    public Residue2D getPreviousResidue() {
        return this.repository.getResidue(this.residue2D.getResidues().iterator().next().getPreviousResidue());
    }

    public boolean isInsideDrawingArea() {
        Rectangle r = this.gc.getDrawingArea();
        return r.contains(this.getCurrentX(), this.getCurrentY())
                || r.contains(this.getCurrentX() + this.gc.getCurrentWidth(), this.getCurrentY())
                || r.contains(this.getCurrentX(), this.getCurrentY() - this.gc.getCurrentHeight())
                || r.contains(this.getCurrentX() + this.gc.getCurrentWidth(), this.getCurrentY() - this.gc.getCurrentHeight());
    }

    public boolean isStrictlyInsideDrawingArea() {
        Rectangle r = this.gc.getDrawingArea();
        return r.contains(this.getCurrentX(), this.getCurrentY())
                && r.contains(this.getCurrentX() + this.gc.getCurrentWidth(), this.getCurrentY())
                && r.contains(this.getCurrentX(), this.getCurrentY() - this.gc.getCurrentHeight())
                && r.contains(this.getCurrentX() + this.gc.getCurrentWidth(), this.getCurrentY() - this.gc.getCurrentHeight());
    }

    public boolean isInsideHelix() {
        return Helix2D.class.isInstance(this.structuralDomain);
    }

    public StructuralDomain2D getStructuralDomain() {
        return structuralDomain;
    }

    public void setStructuralDomain(StructuralDomain2D structure) {
        this.structuralDomain = structure;
        if (Helix2D.class.isInstance(structure))
            this.setColor(Helix2D.DEFAULT_COLOR);
        else
            this.setColor(Sstrand2D.DEFAULT_COLOR);
    }

    public void setPhosphodiesterBond2D(PhosphodiesterBond2D phosphodiesterBond2D) {
        this.phosphodiesterBond2D = phosphodiesterBond2D;
    }

    public void addInteraction(Interaction2D interaction) {
        this.interactions.add(interaction);
    }

    public void removeInteraction(BaseBaseInteraction2D interaction) {
        this.interactions.remove(interaction);
    }

    /**
     * return all the interactions established by this residue
     *
     * @return return an empty list if no interactions for this residue
     */
    public List<Interaction2D> getInteractions() {
        return this.interactions;
    }

    public boolean isUpdated() {
        return this.updated;
    }

    public void isUpdated(final boolean updated) {
        this.updated = updated;
        if (updated) {
            for (Iterator i = this.interactions.iterator(); i.hasNext();)
                ((Interaction2D) i.next()).isUpdated(updated);
            if (this.phosphodiesterBond2D != null) {
                this.phosphodiesterBond2D.isUpdated(updated);
                if (this.getPreviousResidue() != null)
                    this.getPreviousResidue().phosphodiesterBond2D.isUpdated(updated);
            }
        }
    }

    public boolean contains(final double x, final double y) {
        return this.getArea().contains(x, y);
    }

    public List<Residue2D> getResidues() {
        List<Residue2D> residues = new ArrayList<Residue2D>();
        residues.add(this);
        return residues;
    }

    private boolean isToBeLinked() {
        fr.unistra.ibmc.paradise.core.features.Residue2D r = this.getResidue2DFeature();
        Molecule m = r.getMolecules().iterator().next();
        int p = r.getFullLocation(m).getStart();
        return this.repository.mediator.isToBeLinked(m, p);
    }

    public void draw(final java.awt.Graphics2D g2) {
        if (this.isDisplayed() && this.isInsideDrawingArea()) {
            if (this.isSelected())
                g2.setColor(java.awt.Color.red);
            else if (Helix2D.class.isInstance(this.structuralDomain))
                g2.setColor(AlignmentCanvas.SecondaryInteraction_Color);

            else {
                switch (this.getSymbol()) {
                    case 'A': g2.setColor(AlignmentCanvas.A_Color); break;
                    case 'U': g2.setColor(AlignmentCanvas.U_Color); break;
                    case 'G': g2.setColor(AlignmentCanvas.G_Color); break;
                    case 'C': g2.setColor(AlignmentCanvas.C_Color); break;
                    default:  g2.setColor(Color.BLACK);
                }
            }
            if (isToBeLinked())
                g2.setColor(Color.GREEN.darker());
            drawStringCentered(g2,""+this.getSymbol(),this.getCurrentCenterX(),this.getCurrentCenterY());
            g2.setColor(Color.BLACK);
            Residue2D nextR = this.getNextResidue(), previousR = this.getPreviousResidue();
            if (nextR != null && (previousR == null || !previousR.isDisplayed()))
                g2.drawString("5'", (float) (this.getCurrentX() - (nextR.getCurrentX() - this.getCurrentX())), (float) (this.getCurrentY() - (nextR.getCurrentY() - this.getCurrentY())));
            else if (previousR != null && (nextR == null || !nextR.isDisplayed()))
                g2.drawString("3'", (float) (this.getCurrentX() + this.getCurrentX() - previousR.getCurrentX()), (float) (this.getCurrentY() + this.getCurrentY() - previousR.getCurrentY()));
            this.isUpdated(false);
        }
        if (this.phosphodiesterBond2D!= null)
            if (this.isDisplayed() && this.isInsideDrawingArea() || this.getNextResidue().isDisplayed && this.getNextResidue().isInsideDrawingArea())
                this.phosphodiesterBond2D.draw(g2);
    }

    private void drawStringCentered(Graphics2D g2, String res,double x, double y) {
        Dimension d = getStringDimension(g2, res);
        x -= (double)d.width/2.0;
        y += (double)d.height/2.0;
        g2.drawString(res, (int)x, (int)y);
    }

    private Dimension getStringDimension(Graphics2D g, String s) {
        FontMetrics fm = g.getFontMetrics();
        Rectangle2D r = fm.getStringBounds(s, g);
        return (new Dimension((int)r.getWidth(),(int)fm.getAscent()-fm.getDescent()));
    }

    public Area getArea() {
        return new Area(new java.awt.geom.Rectangle2D.Double(this.getCurrentX(), this.getCurrentY() - gc.getCurrentHeight(), gc.getCurrentWidth(), gc.getCurrentHeight()));
    }

    public double getCurrentCenterX() {
        return this.getCurrentX() + this.gc.getCurrentWidth() / 2;
    }

    public double getRealCenterX() {
        return this.getRealX() + this.gc.getRealWidth() / 2;
    }

    public double getCurrentCenterY() {
        return this.getCurrentY() - this.gc.getCurrentHeight() / 2;
    }

    public double getRealCenterY() {
        return this.getRealY() - this.gc.getRealHeight() / 2;
    }

    public java.awt.geom.Point2D getCurrentCenter() {
        return new java.awt.geom.Point2D.Double(this.getCurrentCenterX(), this.getCurrentCenterY());
    }

    public java.awt.geom.Point2D getRealCenter() {
        return new java.awt.geom.Point2D.Double(this.getRealCenterX(), this.getRealCenterY());
    }

    public double getMaxX() {
        return this.getArea().getBounds2D().getMaxX();
    }

    public double getMaxY() {
        return this.getArea().getBounds2D().getMaxY();
    }

    public double getMinX() {
        return this.getArea().getBounds2D().getMinX();
    }

    public double getMinY() {
        return this.getArea().getBounds2D().getMinY();
    }

    public boolean isDisplayed() {
        return this.isDisplayed;
    }

    public void isDisplayed(boolean displayed) {
        this.isDisplayed = displayed;
    }

    public String toSVG() {
        final StringBuffer svg = new StringBuffer();
        svg.append("<text fill=\"" + DrawingUtils.colorToString(this.getSVGColor()) + "\" font-moleculesCount=\"" + (int) (gc.getFont().getSize() / gc.getFinalZoomLevel()) +
                "\" x=\"" + this.getRealX() + "\" y=\"" + this.getRealY() + "\">" + this.residue2D.getResidues().iterator().next().getSymbol() + "</text>\n");
        if (this.phosphodiesterBond2D!= null)
            svg.append(this.phosphodiesterBond2D.toSVG());
        return svg.toString();
    }

    private Color getSVGColor() {
        return Color.BLACK;
    }

    public void translate(double transX, double transY) {
        transX /= this.gc.getFinalZoomLevel();
        transY /= this.gc.getFinalZoomLevel();
        this.setX(this.getRealX() + transX);
        this.setY(this.getRealY() + transY);
    }

    public void rotate(java.awt.geom.Point2D centerPoint, double angle) {
        Point2D p = DrawingUtils.getRotatedPoint(this.getCurrentCoordinates(), angle, centerPoint);
        this.setRealCoordinates((p.getX() - this.gc.getViewX()) / this.gc.getFinalZoomLevel(), (p.getY() - this.gc.getViewY()) / this.gc.getFinalZoomLevel());
    }

    public void flip(final Point2D center) {
        this.translate(-2.0 * (this.getCurrentX() - center.getX()), 0.0);
    }

    public ParadiseID getMoleculeId() {
        return this.residue2D.getResidues().iterator().next().getMoleculeId();
    }

    public boolean equals(Object obj) {
        if (!this.getClass().isInstance(obj))
            return false;
        return this.residue2D.getResidues().iterator().next().equals(((Residue2D)obj).residue2D.getResidues().iterator().next());
    }

    public int compareTo(Object o) {
        if (!Residue2D.class.isInstance(o))
            return -1;
        else if (!this.getMoleculeId().equals(((Residue2D)o).getMoleculeId()))
            return this.getMoleculeId().toString().compareTo(((Residue2D)o).getMoleculeId().toString());
        else
            return new Integer(this.getAbsolutePosition()).compareTo(new Integer(((Residue2D)o).getAbsolutePosition()));
    }

}
