package fr.unistra.ibmc.s2s.ssviewer.graphics;

import fr.unistra.ibmc.s2s.ssviewer.features.Interaction2D;
import fr.unistra.ibmc.s2s.ssviewer.features.Residue2D;
import fr.unistra.ibmc.s2s.ssviewer.features.Helix2D;
import fr.unistra.ibmc.s2s.ssviewer.utils.DrawingUtils;

import java.awt.*;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.*;
import java.util.List;

/**
 * This class manages the drawing properties to display a secondary structure within a SecondaryCanvas
 *
 * @author fjossinet
 */
public class GraphicContext {

    static final byte BASECOLOR = 0, STRUCTURECOLOR = 1, SEQUENCECOLOR = 2, INTERACTIONCOLOR = 3;
    private byte colorMode = STRUCTURECOLOR;
    private Font f;
    //the upperleft corner of the current view
    double viewX, viewY;
    double finalZoomLevel = 1;
    private java.util.List<String> families2show ;
    private boolean showFamily, displayFiveToThreeOrientation = true, displayLabels = true;
    private Point2D gravityPoint;
    private Rectangle drawingArea;
    private double width,
            height,
            half = Math.sqrt(width * width + height * height) / 2,
            symbolsize = 4* width - 2* half;
    private SecondaryCanvas canvas;

    GraphicContext(final SecondaryCanvas canvas) {
        this(6,9,canvas);
    }

    GraphicContext(final double width, final double height, final SecondaryCanvas canvas) {
        this.width = width;
        this.height = height;
        this.half = Math.sqrt(width * width + height * height) / 2;
        this.symbolsize = 4* width - 2* half;
        this.canvas = canvas;
    }

    public void initialize() {
        for (Helix2D h2D:canvas.getModel2D().getHelices())
            this.intializeHelixCurrentContext(h2D);
        this.families2show = new ArrayList<String>();
        this.addFamily("CWW");
        this.addFamily("TWW");
        this.addFamily("CHH");
        this.addFamily("THH");
        this.addFamily("CSS");
        this.addFamily("TSS");
        this.addFamily("CHW");
        this.addFamily("CWH");
        this.addFamily("THW");
        this.addFamily("TWH");
        this.addFamily("CHS");
        this.addFamily("CSH");
        this.addFamily("THS");
        this.addFamily("TSH");
        this.addFamily("CSW");
        this.addFamily("CWS");
        this.addFamily("TSW");
        this.addFamily("TWS");
    }

    public double getHeight() {
        return height;
    }

    public double getWidth() {
        return width;
    }

    public void intializeHelixCurrentContext(Helix2D h2D) {
        Residue2D[] _5PrimEnds = h2D.get5PrimeEnds();
        Residue2D _5PrimeEnd1 = _5PrimEnds[0], _5PrimeEnd2 = _5PrimEnds[1];
        double  length = DrawingUtils.getDistance(_5PrimeEnd1.getRealCoordinates(),h2D.get3PrimeEnd(_5PrimeEnd1).getRealCoordinates());
        Point2D[]   _5 = DrawingUtils.fit(_5PrimeEnd1.getRealCoordinates(),h2D.get3PrimeEnd(_5PrimeEnd1).getRealCoordinates(), (length-this.height *(h2D.getLength()-1))/2),
                _3 = DrawingUtils.fit(_5PrimeEnd2.getRealCoordinates(),h2D.get3PrimeEnd(_5PrimeEnd2).getRealCoordinates(), (length-this.height *(h2D.getLength()-1))/2);
        h2D.setEnds(_5PrimeEnd1,_5[0], h2D.get3PrimeEnd(_5PrimeEnd1), _5[1], _5PrimeEnd2, _3[0], h2D.get3PrimeEnd(_5PrimeEnd2), _3[1]);
        double _width = DrawingUtils.getDistance(_5PrimeEnd1.getRealCoordinates(),h2D.getPairedResidue(_5PrimeEnd1).getRealCoordinates());
        _5 = DrawingUtils.fit(_5PrimeEnd1.getRealCoordinates(),h2D.getPairedResidue(_5PrimeEnd1).getRealCoordinates(), (_width-this.width *4)/2);
        _3 = DrawingUtils.fit(_5PrimeEnd2.getRealCoordinates(),h2D.getPairedResidue(_5PrimeEnd2).getRealCoordinates(), (_width-this.width *4)/2);
        h2D.setEnds(_5PrimeEnd1, _5[0], h2D.getPairedResidue(_5PrimeEnd1), _5[1], _5PrimeEnd2, _3[0], h2D.getPairedResidue(_5PrimeEnd2), _3[1]);
    }

    void moveView(final double transX, final double transY) {
        this.viewX += transX;
        this.viewY += transY;
    }

    void zoomParameters(final double zoomFactor) {
        this.finalZoomLevel *= zoomFactor;
    }

    public double getCurrentWidth() {
        return width * this.finalZoomLevel;
    }

    public double getRealWidth() {
        return width;
    }

    public double getCurrentHeight() {
        return height * this.finalZoomLevel;
    }

    public double getRealHeight() {
        return height;
    }

    public double getCurrentSymbolSize() {
        return symbolsize * this.finalZoomLevel;
    }

    public double getRealSymbolSize() {
        return symbolsize;
    }

    public double getCurrentHalf() {
        return half * this.finalZoomLevel;
    }

    public double getRealHalf() {
        return half;
    }

    public Font getFont() {
        return new java.awt.Font("Helvetica", java.awt.Font.PLAIN, (int) this.getCurrentHeight());
    }

    public double getViewX() {
        return viewX;
    }

    public double getViewY() {
        return viewY;
    }

    Point2D getGravityPoint() {
        return gravityPoint;
    }

    public double getFinalZoomLevel() {
        return finalZoomLevel;
    }

    public boolean containFamily(final String family) {
        for (String f:this.families2show)
            if (f.equals(family))
                return true;
        return false;
    }

    void addFamily(final String family) {
        if (this.containFamily(family))
            this.families2show.remove(family);
        else
            this.families2show.add(family);
    }

    public void increaseFontSize() {
        this.width *= 1.5;
        this.height *= 1.5;
        this.half = Math.sqrt(this.width * this.width + this.height * this.height) / 2;
        this.symbolsize = 4* this.width - 2* this.half;
        for (Helix2D h2D:canvas.getModel2D().getHelices())
            this.intializeHelixCurrentContext(h2D);
    }

    public void decreaseFontSize() {
        this.width /= 1.5;
        this.height /= 1.5;
        this.half = Math.sqrt(this.width * this.width + this.height * this.height) / 2;
        this.symbolsize = 4* this.width - 2* this.half;
        for (Helix2D h2D:canvas.getModel2D().getHelices())
            this.intializeHelixCurrentContext(h2D);
    }

    private void clearFamilies() {
        this.families2show.clear();
    }

    void setFiveToThreeOrientationDisplay(boolean display) {
        this.displayFiveToThreeOrientation = display;
    }

    public boolean isFiveToThreeOrientationDisplayed() {
        return this.displayFiveToThreeOrientation;
    }

    boolean isShowFamily() {
        return this.showFamily;
    }

    void setShowFamily(final boolean showFamily) {
        this.showFamily = showFamily;
    }

    void setGravityPoint(final Point2D gravityPoint) {
        this.gravityPoint = gravityPoint;
    }

    void translateGravityPoint(final double transX, final double transY) {
        if (this.gravityPoint != null)
            this.gravityPoint.setLocation(this.gravityPoint.getX() + transX, this.gravityPoint.getY() + transY);
    }

    void zoomGravityPoint(final double zoomLevel) {
        if (this.gravityPoint != null)
            this.gravityPoint.setLocation(this.gravityPoint.getX() * zoomLevel, this.gravityPoint.getY() * zoomLevel);
    }

    void setColorMode(final byte colorMode) {
        this.colorMode = colorMode;
    }

    void setDrawingArea(final Rectangle drawingArea) {
        this.drawingArea = drawingArea;
    }

    public Rectangle getDrawingArea() {
        return drawingArea;
    }

    public boolean isLabelsDisplayed() {
        return displayLabels;
    }

    public void setDisplayLabels(boolean displayLabels) {
        this.displayLabels = displayLabels;
    }

    public boolean isTertiaryInteractionsDisplayed() {
        return true;
    }
}
