package fr.unistra.ibmc.s2s.ssviewer.graphics;

import fr.unistra.ibmc.paradise.core.utils.Residue;
import fr.unistra.ibmc.paradise.core.ParadiseFeature;
import fr.unistra.ibmc.paradise.core.features.SecondaryStructure;
import fr.unistra.ibmc.paradise.core.features.SecondaryStructureDisplay;
import fr.unistra.ibmc.paradise.tools.ParadiseWorkingSession;
import fr.unistra.ibmc.s2s.ssviewer.features.*;
import fr.unistra.ibmc.s2s.Mediator;
import fr.unistra.ibmc.s2s.S2S;

import javax.swing.*;
import javax.swing.Timer;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.util.*;
import java.util.List;

import org.jdesktop.swingworker.SwingWorker;

/**
 * This class represents the drawing area for the secondary representation of the model
 *
 * @author fjossinet
 */

public class SecondaryCanvas extends JPanel implements MouseWheelListener, java.awt.event.MouseListener, java.awt.event.MouseMotionListener {

    protected static final byte UPDATE = 99;
    protected static final byte TRANSLATION = 100;
    protected static final byte ROTATION = 101;
    protected static final byte ZOOM = 102;
    protected static final byte NEW_STATE = 103;
    protected static final byte INTERACTION_DISPLAY = 106;
    protected static final byte FLIP = 107;
    protected static final byte CLOSE = 108;
    protected GraphicContext gc;
    protected State state;
    protected Model2D model;
    protected java.awt.Image offScreenBuffer;
    private List<StructuralDomain2D> currentStructures;
    private List<Residue2D> currentResidues;
    private List<Interaction2D> currentInteractions;
    protected Point2D gravityPoint;
    protected Mediator mediator;
    protected Rectangle2D selectionRectangle;
    private java.util.List<TextAnnotation> textAnnotations;
    private ParadiseWorkingSession workingSession;

    SecondaryCanvas(final ParadiseWorkingSession workingSession, final Mediator mediator) {
        this.workingSession = workingSession;
        this.currentResidues = new ArrayList<Residue2D>();
        this.currentInteractions = new ArrayList<Interaction2D>();
        this.currentStructures = new ArrayList<StructuralDomain2D>();
        this.addMouseListener(this);
        this.addMouseWheelListener(this);
        this.addMouseMotionListener(this);
        this.setBackground(java.awt.Color.white);
        this.textAnnotations = new ArrayList<TextAnnotation>();
        this.textAnnotations.add(new TextAnnotation());
        this.setCursor(Rna2DViewer.selectCursor);
        this.mediator = mediator;
        this.setState(mediator.getSelectionMode());
        //new LoadNewModel2DTask(model2D).execute();
    }

    public void setState(byte mode) {
        switch (mode) {
            case S2S.SELECT_STRUCTURAL_DOMAIN : this.state = new StructSelect(mediator); break;
            case S2S.SELECT_INTERACTION : this.state = new InteractionSelect(mediator); break;
            case S2S.SELECT_RESIDUE : this.state = new ResidueSelect(mediator); break;
        }
    }

    public void clearDisplay() {
        this.model = null;
        this.gc = null;
        this.repaint();
    }

    public void paintComponent(final java.awt.Graphics g) {
        super.paintComponent(g);
        if (this.model != null && this.gc != null) {
            final java.awt.Graphics2D g2 = (java.awt.Graphics2D) g;
            g2.setRenderingHint(java.awt.RenderingHints.KEY_ANTIALIASING, java.awt.RenderingHints.VALUE_ANTIALIAS_ON);
            g2.setBackground(java.awt.Color.white);
            g2.setFont(this.gc.getFont());
            //we start with the BLACK color
            g2.setColor(Color.BLACK);
            //we draw the (0,0) origin point
            g2.fillOval((int) (this.gc.getViewX() - 5 * this.gc.getFinalZoomLevel()), (int) (this.gc.getViewY() - 5 * this.gc.getFinalZoomLevel()), (int) (10 * this.gc.getFinalZoomLevel()), (int) (10 * this.gc.getFinalZoomLevel()));
            g2.drawLine((int) (this.gc.getViewX() - 10 * this.gc.getFinalZoomLevel()), (int) this.gc.getViewY(), (int) (this.gc.getViewX() + 20 * this.gc.getFinalZoomLevel()), (int) this.gc.getViewY());
            g2.drawLine((int) this.gc.getViewX(), (int) (this.gc.getViewY() - 10 * this.gc.getFinalZoomLevel()), (int) this.gc.getViewX(), (int) (this.gc.getViewY() + 20 * this.gc.getFinalZoomLevel()));
            if (this.selectionRectangle != null) {
                g2.setColor(Color.LIGHT_GRAY);
                g2.fill(this.selectionRectangle);
                g2.setColor(Color.BLACK);
            }
            this.gc.setDrawingArea(new Rectangle(0, 0, this.getWidth(), this.getHeight()));
            this.model.draw(g2);
        }

    }

    public Model2D getModel2D() {
        return this.model;
    }

    public GraphicContext getGraphicContext() {
        return this.gc;
    }

    public void addFamily(final String family) {
        this.gc.addFamily(family);
        this.repaint();
    }

    public boolean displayFamily(final String family) {
        return this.gc.containFamily(family);
    }

    public void setFiveToThreeOrientationDisplay(boolean display) {
        this.gc.setFiveToThreeOrientationDisplay(display);
        this.repaint();
    }

    public boolean isFiveToThreeOrientationDisplayed() {
        return this.gc.isFiveToThreeOrientationDisplayed();
    }

    void setColorMode(final byte colorMode) {
        this.gc.setColorMode(colorMode);
    }

    void setColor(final Color color) {
        Iterator i = this.getCurrentResidues().iterator();
        while (i.hasNext())
            ((Rna2DElement) i.next()).setColor(color);
        i = this.getCurrentStructures().iterator();
        while (i.hasNext())
            ((Rna2DElement) i.next()).setColor(color);
        i = this.getCurrentInteractions().iterator();
        while (i.hasNext())
            ((Rna2DElement) i.next()).setColor(color);
        List<Residue> residues = new ArrayList<Residue>(this.getCurrentResidues().size());
        for (Residue2D r2D : this.getCurrentResidues())
            residues.add(r2D.getResidue2DFeature().getResidues().iterator().next());
        this.repaint();
    }

    public void update(final java.awt.Graphics g) {
        final java.awt.Graphics2D gr;
        if (this.offScreenBuffer == null ||
                (!(offScreenBuffer.getWidth(this) == this.getSize().width
                        && offScreenBuffer.getHeight(this) == this.getSize().height))) {
            this.offScreenBuffer = this.createImage(this.getSize().width, this.getSize().height);
        }
        // We need to use our buffer Image as a Graphics object:
        gr = (java.awt.Graphics2D) this.offScreenBuffer.getGraphics();
        paintComponent(gr);
        g.drawImage(this.offScreenBuffer, 0, 0, this);
    }

    /**
     * Show interactions established by one Residue
     */
    void showInteractionsFrom(Residue r) {
        java.util.List interactions = this.model.getInteractions(r.getMoleculeId(), r.getAbsolutePosition());
        Interaction2D inter = null;
        for (Iterator i = interactions.iterator(); i.hasNext();) {
            inter = (Interaction2D) i.next();
            inter.drawRealSymbol(true);
            inter.isDisplayed(true);
        }
        if (interactions.size() > 0)
            this.repaint();
    }

    /**
     * Return all the selected Residue2D features within this canvas.
     *
     * @return all the selected Residue2D features in a <b>sorted</b> List
     */
    public List<Residue2D> getCurrentResidues() {
        List<Residue2D> residues = new ArrayList<Residue2D>(this.currentResidues);
        for (StructuralDomain2D s2D : this.currentStructures)
            residues.addAll(s2D.getResidues());
        for (Interaction2D interaction : this.currentInteractions)
            residues.addAll(interaction.getResidues());
        Collections.sort(residues);
        return residues;
    }

    /**
     * Return all the selected StructuralDomain2D features within this canvas
     *
     * @return all the selected StructuralDomain2D nodesin a List
     */
    public List<StructuralDomain2D> getCurrentStructures() {
        return this.currentStructures;
    }

    /**
     * Return all the selected Interaction2D features within this canvas
     *
     * @return all the selected Interaction2D features in a List
     */
    public List<Interaction2D> getCurrentInteractions() {
        List<Interaction2D> interactions = new ArrayList<Interaction2D>(this.currentInteractions);
        for (StructuralDomain2D s2D : this.currentStructures)
            if (Helix2D.class.isInstance(s2D))
                interactions.addAll(((Helix2D) s2D).getSecondaryInteractions());
        return interactions;
    }

    /**
     * give the current State for this canvas2D
     *
     * @return the current State
     */
    State getState() {
        return this.state;
    }

    public double getViewX() {
        return this.gc.getViewX();
    }

    public double getViewY() {
        return this.gc.getViewY();
    }

    public double getFinalZoomLevel() {
        return this.gc.getFinalZoomLevel();
    }

    Point2D getGravityPoint() {
        return this.gc.getGravityPoint();
    }


    public void mouseClicked(final MouseEvent e) {
        if (this.model != null) {
            this.state.mouseClicked(e);
        }
        if (e.getClickCount() == 2 && this.currentStructures.size() > 0) {
            final String label = JOptionPane.showInputDialog(this, "Set Label", this.currentStructures.iterator().next().getLabel());
            if (label != null && label.length() > 0) {
                this.currentStructures.iterator().next().setLabel(label);
                this.repaint();
            }
        }
        /*
        for (TextAnnotation a:this.textAnnotations)
            a.click(0,0);*/
    }

    public void mouseEntered(final MouseEvent e) {
    }

    public void mouseExited(final MouseEvent e) {
    }

    public void mousePressed(final MouseEvent e) {
        if (this.model != null) {
            this.state.mousePressed(e);
        }
    }

    public void mouseReleased(final MouseEvent e) {
        if (this.model != null) {
            this.state.mouseReleased(e);
        }
    }

    public void mouseDragged(final MouseEvent e) {
        if (this.model != null) {
            this.state.mouseDragged(e);
        }
    }

    public void mouseMoved(final MouseEvent e) {
        if (this.model != null) {
            this.state.mouseMoved(e);
        }
    }

    public void mouseWheelMoved(MouseWheelEvent e) {
        int notches = e.getWheelRotation();
        java.awt.geom.Point2D zoomCenter = new java.awt.geom.Point2D.Double(e.getX() * this.getFinalZoomLevel() + this.getViewX(), e.getY() * this.getFinalZoomLevel() + this.getViewY()),
                newDisplayZoomCenter = null;
        if (notches < 0)
            this.zoomView(1.5);
        else
            this.zoomView(1 / 1.5);
        newDisplayZoomCenter = new java.awt.geom.Point2D.Double(e.getX() * this.getFinalZoomLevel() + this.getViewX(), e.getY() * this.getFinalZoomLevel() + this.getViewY());
        this.translateView(zoomCenter.getX() - newDisplayZoomCenter.getX(), zoomCenter.getY() - newDisplayZoomCenter.getY());
    }

    void updateGravityPoint() {
        Rectangle2D r = null;
        for (Residue2D b : this.getCurrentResidues()) {
            if (b.isInsideHelix()) {
                if (r == null)
                    r = b.getArea().getBounds2D();
                else
                    r.add(b.getArea().getBounds2D());
            }
        }
        if (r != null)
            this.gc.setGravityPoint(new Point2D.Double(r.getCenterX(), r.getCenterY()));
        else
            this.gc.setGravityPoint(null);
    }

    /**
     * Move the view to center the model's center on canvas'center
     */
    void centerView() {
        if (this.model != null) {
            double centerX = 0, centerY = 0;
            List<Residue2D> residues = this.model.getResidues();
            for (Residue2D r:residues) {
                centerX+=r.getCurrentCenterX();
                centerY+=r.getCurrentCenterY();
            }
            centerX = centerX /residues.size();
            centerY = centerY /residues.size();
            this.translateView(this.getWidth() / 2
                    - centerX, this.getHeight()/ 2 - centerY);
            this.repaint();
        }
    }

    public Residue2D seekResidue2D(Residue r) {
        for (Residue2D r2D:this.model.getResidues()) {
            if (r2D.getAbsolutePosition() == r.getAbsolutePosition() && r2D.getMoleculeId().equals(r.getMoleculeId()))
                return r2D;
        }
        return null;
    }

    /**
     * Seek the Rna2DElement wrapping the ParadiseFeature given as argument
     *
     * @param feature
     * @return
     */
    public Rna2DElement seekRna2DNode(final ParadiseFeature feature) {
        for (BaseBaseInteraction2D interaction : this.model.getInterMolecularTertiaryInteractions())
            if (feature == interaction.getBaseBaseInteractionFeature())
                return interaction;
        for (Helix2D h : this.model.getInterMolecularHelices()) {
            if (feature == h.getHelixFeature())
                return h;
            for (BaseBaseInteraction2D interaction : h.getSecondaryInteractions())
                if (feature == interaction.getBaseBaseInteractionFeature())
                    return interaction;
            for (Residue2D r : h.getResidues())
                if (r.getResidue2DFeature().getResidues().iterator().next() == feature)
                    return r;
        }
        for (Rna2D rna : this.model.getRnas()) {
            for (BaseBaseInteraction2D interaction : rna.getTertiaryInteractions())
                if (feature == interaction.getBaseBaseInteractionFeature())
                    return interaction;
            for (Helix2D h : rna.getHelices()) {
                if (feature == h.getHelixFeature())
                    return h;
                for (BaseBaseInteraction2D interaction : h.getSecondaryInteractions())
                    if (feature == interaction.getBaseBaseInteractionFeature())
                        return interaction;
                for (Residue2D r : h.getResidues())
                    if (r.getResidue2DFeature().getResidues().iterator().next() == feature)
                        return r;
            }
            for (Sstrand2D ss : rna.getStrands()) {
                if (feature == ss.getSingleStrandFeature())
                    return ss;
                for (Residue2D r : ss.getResidues())
                    if (r.getResidue2DFeature().getResidues().iterator().next() == feature)
                        return r;
            }
        }
        return null;
    }

    /**
     * Return all the SecondaryElement associated with the label given as argument
     *
     * @param label
     * @return
     */
    //TODO used only for Helix or Sstrand ?? => if so, not necessary to parse Interaction objects => to fix
    public List<Rna2DElement> getSecondaryElementByLabel(String label) {
        List<Rna2DElement> elements = new ArrayList<Rna2DElement>();
        for (Interaction2D interaction : this.model.getInterMolecularTertiaryInteractions())
            if (label.equals(interaction.getLabel()))
                elements.add(interaction);
        for (Helix2D h : this.model.getInterMolecularHelices()) {
            if (label.equals(h.getLabel()))
                elements.add(h);
            for (Interaction2D interaction : h.getSecondaryInteractions())
                if (label.equals(interaction.getLabel()))
                    elements.add(interaction);
            for (Residue2D r : h.getResidues())
                if (label.equals(r.getLabel()))
                    elements.add(r);
        }
        for (Rna2D rna : this.model.getRnas()) {
            for (Interaction2D interaction : rna.getTertiaryInteractions())
                if (label.equals(interaction.getLabel()))
                    elements.add(interaction);
            for (Helix2D h : rna.getHelices()) {
                if (label.equals(h.getLabel()))
                    elements.add(h);

                for (Interaction2D interaction : h.getSecondaryInteractions())
                    if (label.equals(interaction.getLabel()))
                        elements.add(interaction);
                for (Residue2D r : h.getResidues())
                    if (label.equals(r.getLabel()))
                        elements.add(r);

            }
            for (Sstrand2D ss : rna.getStrands()) {
                if (label.equals(ss.getLabel()))
                    elements.add(ss);
                for (Residue2D r : ss.getResidues())
                    if (label.equals(r.getLabel()))
                        elements.add(r);
            }
        }
        return elements;
    }

    public void setState(final State state) {
        this.state = state;
    }

    /**
     * Give a String describing the drawing with the SVG grammar
     *
     * @return a String containing the SVG details of the Model2D
     */
    public String toSVG() {
        final StringBuffer svg = new StringBuffer();
        svg.append("<?xml version=\"1.0\"?>\n");
        svg.append("<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.0//EN\" \"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd\">\n");
        svg.append("<svg width=\"");
        svg.append(this.gc.getCurrentWidth());
        svg.append("\" height=\"");
        svg.append(this.gc.getCurrentHeight());
        svg.append("\">\n");
        if (this.model != null)
            svg.append(this.model.toSVG());
        svg.append("</svg>");
        return svg.toString();
    }

    Interaction2D addCtrlCurrentInteraction(final Interaction2D inter) {
        if (this.currentInteractions.contains(inter)) {
            this.removeCurrentInteraction(inter);
            return inter;
        }
        this.addInteraction(inter);
        return null;
    }

    void addShiftCurrentInteraction(final Interaction2D inter) {
        if (this.currentInteractions.contains(inter))
            return;
        this.addInteraction(inter);
    }

    void addCurrentInteraction(final Interaction2D inter) {
        this.clearAllSelections();
        this.addInteraction(inter);
    }

    private void addInteraction(final Interaction2D inter) {
        this.currentInteractions.add(inter);
        inter.isSelected(true);
        inter.drawRealSymbol(true);
        inter.isDisplayed(true);
        List<Residue> selectedResidues = new ArrayList<Residue>();
        selectedResidues.add(inter.getResidue().getResidue2DFeature().getResidues().iterator().next());
        selectedResidues.add(inter.getPartnerResidue().getResidue2DFeature().getResidues().iterator().next());
        this.repaint();
    }

    private void removeCurrentInteraction(final Interaction2D inter) {
        inter.isSelected(false);
        this.currentInteractions.remove(inter);
        this.repaint();
    }

    Residue2D addCtrlCurrentResidue(final Residue2D residue2D) {
        if (this.currentResidues.contains(residue2D)) {
            this.removeCurrentResidue(residue2D);
            return residue2D;
        }
        this.addResidue(residue2D);
        return null;
    }

    void addShiftCurrentResidue(final Residue2D residue2D) {
        if (this.currentResidues.contains(residue2D))
            return;
        this.addResidue(residue2D);
    }

    void addCurrentResidue(final Residue2D residue) {
        this.clearAllSelections();
        this.addResidue(residue);
    }

    private void addResidue(final Residue2D residue) {
        this.currentResidues.add(residue);
        residue.isSelected(true);
        residue.isDisplayed(true);
        List<Residue> selectedResidues = new ArrayList<Residue>();
        selectedResidues.add(residue.getResidue2DFeature().getResidues().iterator().next());
        this.repaint();
    }

    private void removeCurrentResidue(final Residue2D base) {
        base.isSelected(false);
        this.currentResidues.remove(base);
        this.repaint();
    }

    void addCurrentStructure(final StructuralDomain2D structure) {
        this.clearAllSelections();
        this.addStructure(structure);
    }

    StructuralDomain2D addCtrlCurrentStructure(final StructuralDomain2D structure) {
        //we search if this fixedStructure is already in the array, if we find it we delete it
        if (this.currentStructures.contains(structure)) {
            this.removeCurrentStructure(structure);
            return structure;
        } else {
            this.addStructure(structure);
            return null;
        }
    }

    void addShiftCurrentStructure(final StructuralDomain2D structure) {
        //we search if this fixedStructure is already in the array, if we find it we do nothing
        if (this.currentStructures.contains(structure))
            return;
        this.addStructure(structure);
    }

    private void addStructure(final StructuralDomain2D structure) {
        structure.isSelected(true);
        structure.isDisplayed(true);
        this.currentStructures.add(structure);
        List<Residue> selectedResidues = new ArrayList<Residue>();
        for (Residue2D r : structure.getResidues())
            selectedResidues.add(r.getResidue2DFeature().getResidues().iterator().next());
        this.repaint();
    }

    public void removeCurrentStructure(final StructuralDomain2D structure) {
        structure.isSelected(false);
        this.currentStructures.remove(structure);
        this.repaint();
    }

    public void clearAllSelections() {
        for (Residue2D r : this.currentResidues)
            r.isSelected(false);
        this.currentResidues.clear();
        for (StructuralDomain2D sd : this.currentStructures)
            sd.isSelected(false);
        this.currentStructures.clear();
        for (Interaction2D i : this.currentInteractions)
            i.isSelected(false);
        this.currentInteractions.clear();
        this.repaint();
    }

    /**
     * Hide all the tertiary interactions displayed in the SecondaryCanvas.
     */
    public void clearTertiaries() {
        Rna2D r = null;
        for (Interaction2D interaction : this.model.getInterMolecularTertiaryInteractions()) {
            interaction.isDisplayed(false);
            interaction.drawRealSymbol(false);
        }
        for (Rna2D rna : this.model.getRnas()) {
            for (Interaction2D interaction : rna.getTertiaryInteractions()) {
                interaction.isDisplayed(false);
                interaction.drawRealSymbol(false);
            }
        }
        this.repaint();
    }

    public void translateView(final double transX, final double transY) {
        this.gc.moveView(transX, transY);
        this.repaint();
    }

    public void zoomView(final double zoomLevel) {
        this.gc.zoomParameters(zoomLevel);
        this.repaint();
    }

    void translateSelection(final double transX, final double transY) {
        this.updateGravityPoint();
        Residue2D r = null;
        for (Iterator sc = this.getCurrentResidues().iterator(); sc.hasNext();) {
            r = (Residue2D) sc.next();
            if (!r.isInsideHelix() && !((Sstrand2D) r.getStructuralDomain()).isAutomaticCoordinates())
                r.translate(transX, transY);
        }
        for (Iterator sc = this.getCurrentStructures().iterator(); sc.hasNext();)
            ((StructuralDomain2D) sc.next()).translateEnds(transX, transY);
        this.repaint();
    }

    void rotateSelection(final double angle, final Point2D centerOfRotation) {
        for (Iterator sc = this.getCurrentStructures().iterator(); sc.hasNext();)
            ((StructuralDomain2D) sc.next()).rotateEnds(centerOfRotation, Math.toRadians(angle));
        this.repaint();
    }

    void flipSelection() {
        this.updateGravityPoint();
        Residue2D r = null;
        for (Iterator sc = this.getCurrentResidues().iterator(); sc.hasNext();) {
            r = (Residue2D) sc.next();
            if (!r.isInsideHelix() && !((Sstrand2D) r.getStructuralDomain()).isAutomaticCoordinates())
                r.flip(this.getGravityPoint());
        }
        for (Iterator sc = this.getCurrentStructures().iterator(); sc.hasNext();)
            ((StructuralDomain2D) sc.next()).flipEnds(this.getGravityPoint());
        this.repaint();
    }

    public void showAll(boolean show) {
        List sstrands = this.model.getSstrands(),
                helices = this.model.getHelices();
        Sstrand2D ss = null;
        Helix2D h = null;
        for (Iterator sc = sstrands.iterator(); sc.hasNext();)
            ((Sstrand2D) sc.next()).isDisplayed(show);
        for (Iterator sc = helices.iterator(); sc.hasNext();)
            ((Helix2D) sc.next()).isDisplayed(show);
        this.repaint();
    }

    void showFamily(final boolean show) {
        this.gc.setShowFamily(show);
    }

    /**
     *
     */
    public void showPeers() {
        for (Residue2D r : this.getCurrentResidues())
            for (Iterator j = r.getInteractions().iterator(); j.hasNext();) {
                Interaction2D inter = (Interaction2D) j.next();
                if (!SingleHBond.class.isInstance(inter) && !PhosphodiesterBond2D.class.isInstance(inter)) {
                    inter.isDisplayed(true);
                    inter.getResidue().getStructuralDomain().isDisplayed(true);
                    inter.getPartnerResidue().getStructuralDomain().isDisplayed(true);
                }
            }
        this.repaint();
    }

    public BufferedImage getImage() {
        BufferedImage bufferedImage = new BufferedImage(this.getSize().width,
                this.getSize().height, BufferedImage.TYPE_BYTE_INDEXED);
        Graphics2D g2 = bufferedImage.createGraphics();
        g2.setColor(Color.WHITE);
        g2.fill(new Rectangle(0, 0, this.getSize().width,
                this.getSize().height));
        g2.setRenderingHint(java.awt.RenderingHints.KEY_ANTIALIASING,
                java.awt.RenderingHints.VALUE_ANTIALIAS_ON);
        g2.setBackground(java.awt.Color.white);
        g2.setFont(this.gc.getFont());
        //todo is the next line necessary ??
        this.setFont(this.gc.getFont());
        //we start with the BLACK color
        g2.setColor(Color.BLACK);
        if (this.model != null) {
            this.gc.setDrawingArea(new Rectangle(0, 0, this.getWidth(), this
                    .getHeight()));
            this.model.draw(g2);
        }
        g2.dispose();
        return bufferedImage;
    }

    public void setSelectionRectangle(Rectangle2D rectangle) {
        this.selectionRectangle = rectangle;
    }

    public void increaseFontSize() {
        this.gc.increaseFontSize();
        this.repaint();
    }

    public void decreaseFontSize() {
        this.gc.decreaseFontSize();
        this.repaint();
    }

    public boolean isLabelsDisplayed() {
        return this.gc.isLabelsDisplayed();
    }

    public void setDisplayLabels(boolean display) {
        this.gc.setDisplayLabels(display);
        this.repaint();
    }


    public void setModel2D(SecondaryStructure ss, SecondaryStructureDisplay ssd) {
        this.gc = new GraphicContext(this);
        this.model = new Model2D(ss, ssd, this.mediator, this.gc);
        this.gc.initialize();
        this.repaint();
    }

    private class ForceDirectedLayout extends SwingWorker implements ActionListener {
        protected Object doInBackground() throws Exception {
            class MyTimer extends Timer {
                public MyTimer(ActionListener actionListener) {
                    super(100, actionListener);
                }
            }
            MyTimer timer = new MyTimer(this);
            timer.start();
            return null;
        }

        public void actionPerformed(ActionEvent actionEvent) {
            Set<Helix2D> helicesToMove = new HashSet<Helix2D>();
            for (Helix2D h : model.getHelices()) {
                double dx = 0, dy = 0;
                Point2D c = h.getCenter();
                for (Residue2D r2D : h.get5PrimeEnds()) {
                    if (r2D.getPreviousResidue() != null && Helix2D.class.isInstance(r2D.getPreviousResidue().getStructuralDomain())) {
                        Helix2D _h = (Helix2D) r2D.getPreviousResidue().getStructuralDomain();
                        Point2D _c = _h.getCenter();
                        double vx = _c.getX() - c.getX();
                        double vy = _c.getY() - c.getY();
                        double len = Math.sqrt(vx * vx + vy * vy);
                        len = (len == 0) ? .0001 : len;
                        double f = (100 - len) / (len * 3);
                        dx = f * vx;
                        dy = f * vy;

                        _h.dx += dx;
                        _h.dy += dy;
                        h.dx += -dx;
                        h.dy += -dy;
                    }
                }
                for (Residue2D r2D : h.get3PrimeEnds()) {
                    if (r2D.getNextResidue() != null && Helix2D.class.isInstance(r2D.getNextResidue().getStructuralDomain())) {
                        Helix2D _h = (Helix2D) r2D.getNextResidue().getStructuralDomain();
                        Point2D _c = _h.getCenter();
                        double vx = _c.getX() - c.getX();
                        double vy = _c.getY() - c.getY();
                        double len = Math.sqrt(vx * vx + vy * vy);
                        len = (len == 0) ? .0001 : len;
                        double f = (50 - len) / (len * 3);
                        dx = f * vx;
                        dy = f * vy;

                        _h.dx += dx;
                        _h.dy += dy;
                        h.dx += -dx;
                        h.dy += -dy;
                    }
                }
                dx = 0;
                dy = 0;
                for (Helix2D _h : model.getHelices()) {
                    if (h == _h)
                        continue;
                    Point2D _c = _h.getCenter();
                    double vx = c.getX() - _c.getX(),
                            vy = c.getY() - _c.getY();
                    double len = vx * vx + vy * vy;
                    if (len == 0) {
                        dx += Math.random();
                        dy += Math.random();
                    } else
                    if (h.getArea().intersects(_h.getArea().getBounds2D()) /*this is the condition to stop the movement between h and _h*/) {
                        dx += vx / len;
                        dy += vy / len;
                    }
                }
                double dlen = dx * dx + dy * dy;
                if (dlen > 0) {
                    dlen = Math.sqrt(dlen) / 2;
                    h.dx += dx / dlen;
                    h.dx *= 2;
                    h.dy += dy / dlen;
                    h.dy *= 2;
                }
                if (!h.isSelected() && (h.dx != 0 || h.dy != 0))
                    helicesToMove.add(h);
                if (h.isSelected()) {
                    h.dx = 0;
                    h.dy = 0; //otherwise the selected helix accumulates dx and dy and once unselected, uses it to move
                }

            }
            boolean repaint = false;
            for (Helix2D h : helicesToMove) {
                double transX = Math.max(-20, Math.min(20, h.dx)), transY = Math.max(-20, Math.min(20, h.dy));
                //to round the dX and dY values to a number with two decimals (1.6666666 => 1.7)
                //otherwise the dX and dY values will never be equal to 0 (we can have 2.2E-23 instead)
                //and the repaint function will always be called (and consequently will interfer too much with the user work)
                transX *= 10.0;
                transX = Math.floor(transX + 0.5);
                transX /= 10.0;
                transY *= 10.0;
                transY = Math.floor(transY + 0.5);
                transY /= 10.0;
                if (transX != 0.0 && transY != 0.0) {
                    h.translateEnds(transX, transY);
                    h.dx /= 2;
                    h.dy /= 2;
                    //at least one helix is translated => repaint necessary
                    repaint = true;
                }
            }
            if (repaint)
                repaint();
        }

    }


}
